@@ -4,7 +4,6 @@ import (
 	"context"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"net/http"
 	"net/url"
 	"os"
@@ -14,6 +13,7 @@ import (
 
 	"github.com/sirupsen/logrus"
 	"github.com/urfave/cli"
+	"mvdan.cc/sh/v3/shell"
 
 	"gitlab.com/gitlab-org/gitlab-runner/commands/helpers/archive"
 	"gitlab.com/gitlab-org/gitlab-runner/commands/helpers/meter"
@@ -25,18 +25,18 @@ import (
 	_ "gocloud.dev/blob/azureblob" // Needed to register the Azure driver
 )
 
-//nolint:lll
 type CacheArchiverCommand struct {
 	fileArchiver
 	retryHelper
 	meter.TransferMeterCommand
 
-	File             string   `long:"file" description:"The path to file"`
-	URL              string   `long:"url" description:"URL of remote cache resource (pre-signed URL)"`
-	GoCloudURL       string   `long:"gocloud-url" description:"Go Cloud URL of remote cache resource (requires credentials)"`
-	Timeout          int      `long:"timeout" description:"Overall timeout for cache uploading request (in minutes)"`
-	Headers          []string `long:"header" description:"HTTP headers to send with PUT request (in form of 'key:value')"`
-	CompressionLevel string   `long:"compression-level" env:"CACHE_COMPRESSION_LEVEL" description:"Compression level (fastest, fast, default, slow, slowest)"`
+	File                   string   `long:"file" description:"The path to file"`
+	URL                    string   `long:"url" description:"URL of remote cache resource (pre-signed URL)"`
+	GoCloudURL             string   `long:"gocloud-url" description:"Go Cloud URL of remote cache resource (requires credentials)"`
+	Timeout                int      `long:"timeout" description:"Overall timeout for cache uploading request (in minutes)"`
+	Headers                []string `long:"header" description:"HTTP headers to send with PUT request (in form of 'key:value')"`
+	CompressionLevel       string   `long:"compression-level" env:"CACHE_COMPRESSION_LEVEL" description:"Compression level (fastest, fast, default, slow, slowest)"`
+	MaxUploadedArchiveSize int64    `long:"max-uploaded-archive-size" env:"CACHE_MAX_UPLOADED_ARCHIVE_SIZE" description:"Limit the size of the cache archive being uploaded to cloud storage, in bytes."`
 
 	client *CacheClient
 	mux    *blob.URLMux
@@ -142,15 +142,15 @@ func (c *CacheArchiverCommand) handleGoCloudURL(file io.Reader) error {
 	return nil
 }
 
-func (c *CacheArchiverCommand) createZipFile(filename string) error {
-	err := os.MkdirAll(filepath.Dir(filename), 0700)
+func (c *CacheArchiverCommand) createZipFile(filename string) (int64, error) {
+	err := os.MkdirAll(filepath.Dir(filename), 0o700)
 	if err != nil {
-		return err
+		return 0, err
 	}
 
-	f, err := ioutil.TempFile(filepath.Dir(filename), "archive_")
+	f, err := os.CreateTemp(filepath.Dir(filename), "archive_")
 	if err != nil {
-		return err
+		return 0, err
 	}
 	defer os.Remove(f.Name())
 	defer f.Close()
@@ -159,29 +159,32 @@ func (c *CacheArchiverCommand) createZipFile(filename string) error {
 
 	archiver, err := archive.NewArchiver(archive.Zip, f, c.wd, GetCompressionLevel(c.CompressionLevel))
 	if err != nil {
-		return err
+		return 0, err
 	}
 
 	// Create archive
 	err = archiver.Archive(context.Background(), c.files)
 	if err != nil {
-		return err
+		return 0, err
+	}
+
+	info, err := f.Stat()
+	if err != nil {
+		return 0, err
 	}
 
 	err = f.Close()
 	if err != nil {
-		return err
+		return 0, err
 	}
 
-	return os.Rename(f.Name(), filename)
+	return info.Size(), os.Rename(f.Name(), filename)
 }
 
 func (c *CacheArchiverCommand) Execute(*cli.Context) {
 	log.SetRunnerFormatter()
 
-	if c.File == "" {
-		logrus.Fatalln("Missing --file")
-	}
+	c.normalizeArgs()
 
 	// Enumerate files
 	err := c.enumerate()
@@ -197,21 +200,53 @@ func (c *CacheArchiverCommand) Execute(*cli.Context) {
 	}
 
 	// Create archive
-	err = c.createZipFile(c.File)
+	size, err := c.createZipFile(c.File)
 	if err != nil {
 		logrus.Fatalln(err)
 	}
 
-	// Upload archive if needed
-	if c.URL != "" || c.GoCloudURL != "" {
-		err := c.doRetry(c.upload)
-		if err != nil {
-			logrus.Fatalln(err)
+	c.uploadArchiveIfNeeded(size)
+}
+
+func (c *CacheArchiverCommand) normalizeArgs() {
+	if c.File == "" {
+		logrus.Fatalln("Missing --file")
+	}
+
+	for idx := range c.Paths {
+		if path, err := shell.Expand(c.Paths[idx], nil); err != nil {
+			logrus.Warnf("invalid path %q: %v", path, err)
+		} else {
+			c.Paths[idx] = path
 		}
-	} else {
+	}
+
+	for idx := range c.Exclude {
+		if path, err := shell.Expand(c.Exclude[idx], nil); err != nil {
+			logrus.Warnf("invalid path %q: %v", path, err)
+		} else {
+			c.Exclude[idx] = path
+		}
+	}
+}
+
+func (c *CacheArchiverCommand) uploadArchiveIfNeeded(size int64) {
+	if c.URL == "" && c.GoCloudURL == "" {
 		logrus.Infoln(
-			"No URL provided, cache will be not uploaded to shared cache server. " +
+			"No URL provided, cache will not be uploaded to shared cache server. " +
 				"Cache will be stored only locally.")
+		return
+	}
+
+	if c.MaxUploadedArchiveSize != 0 && size > c.MaxUploadedArchiveSize {
+		logrus.Infoln(fmt.Sprintf("Cache archive size (%d) is too big (Limit is set to %d). "+
+			"Cache will be stored only locally.", size, c.MaxUploadedArchiveSize))
+		return
+	}
+
+	err := c.doRetry(c.upload)
+	if err != nil {
+		logrus.Fatalln(err)
 	}
 }
 
@@ -226,12 +261,12 @@ func (c *CacheArchiverCommand) setHeaders(req *http.Request, fi os.FileInfo) {
 
 			req.Header.Set(strings.TrimSpace(parsed[0]), strings.TrimSpace(parsed[1]))
 		}
-
-		return
 	}
 
-	// Set default headers
-	req.Header.Set("Content-Type", "application/octet-stream")
+	// Set default headers. But don't override custom Content-Type.
+	if req.Header.Get("Content-Type") == "" {
+		req.Header.Set("Content-Type", "application/octet-stream")
+	}
 	req.Header.Set("Last-Modified", fi.ModTime().Format(http.TimeFormat))
 }
 
