@@ -1,4 +1,4 @@
-// +build integration
+//go:build integration
 
 package helpers_test
 
@@ -7,7 +7,6 @@ import (
 	"context"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"net/http"
 	"net/http/httptest"
 	"net/url"
@@ -19,6 +18,7 @@ import (
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/mock"
 	"github.com/stretchr/testify/require"
+	"github.com/urfave/cli"
 	"gocloud.dev/blob"
 	"gocloud.dev/blob/fileblob"
 
@@ -27,9 +27,33 @@ import (
 	testHelpers "gitlab.com/gitlab-org/gitlab-runner/helpers"
 )
 
-const cacheArchiverArchive = "archive.zip"
-const cacheArchiverTestArchivedFile = "archive_file"
-const cacheExtractorTestArchivedFile = "archive_file"
+const (
+	cacheArchiverArchive           = "archive.zip"
+	cacheArchiverTestArchivedFile  = "archive_file"
+	cacheExtractorTestArchivedFile = "archive_file"
+)
+
+func TestCacheArchiverUploadExpandArgs(t *testing.T) {
+	srv := httptest.NewServer(http.HandlerFunc(testCacheUploadHandler))
+	defer srv.Close()
+	defer os.Remove(cacheArchiverArchive)
+
+	os.Setenv("expand", "expanded")
+	defer os.Unsetenv("expand")
+
+	cmd := helpers.CacheArchiverCommand{
+		File:    cacheArchiverArchive,
+		URL:     srv.URL + "/cache.zip",
+		Timeout: 0,
+	}
+	cmd.Paths = []string{"unexpanded", "path/${expand}/${expand:1:3}"}
+	cmd.Exclude = []string{"unexpanded", "path/$expand/${foo:-bar}"}
+
+	cmd.Execute(&cli.Context{})
+
+	assert.Equal(t, []string{"unexpanded", "path/expanded/xpa"}, cmd.Paths)
+	assert.Equal(t, []string{"unexpanded", "path/expanded/bar"}, cmd.Exclude)
+}
 
 func TestCacheArchiverIsUpToDate(t *testing.T) {
 	helpers.OnEachZipArchiver(t, func(t *testing.T) {
@@ -100,8 +124,7 @@ func TestCacheArchiverRemoteServer(t *testing.T) {
 }
 
 func TestCacheArchiverGoCloudRemoteServer(t *testing.T) {
-	mux, bucketDir, cleanup := setupGoCloudFileBucket(t, "testblob")
-	defer cleanup()
+	mux, bucketDir := setupGoCloudFileBucket(t, "testblob")
 
 	objectName := "path/to/cache.zip"
 
@@ -189,7 +212,7 @@ func TestCacheArchiverCompressionLevel(t *testing.T) {
 			mockArchiver := new(archive.MockArchiver)
 			defer mockArchiver.AssertExpectations(t)
 
-			archive.Register(
+			prevArchiver, _ := archive.Register(
 				"zip",
 				func(w io.Writer, dir string, level archive.CompressionLevel) (archive.Archiver, error) {
 					assert.Equal(t, helpers.GetCompressionLevel(expectedLevel), level)
@@ -197,6 +220,11 @@ func TestCacheArchiverCompressionLevel(t *testing.T) {
 				},
 				nil,
 			)
+			defer archive.Register(
+				"zip",
+				prevArchiver,
+				nil,
+			)
 
 			mockArchiver.On("Archive", mock.Anything, mock.Anything).Return(nil)
 
@@ -216,18 +244,14 @@ func (o *dirOpener) OpenBucketURL(_ context.Context, u *url.URL) (*blob.Bucket,
 	return fileblob.OpenBucket(o.tmpDir, nil)
 }
 
-func setupGoCloudFileBucket(t *testing.T, scheme string) (m *blob.URLMux, bucketDir string, cleanup func()) {
-	tmpDir, err := ioutil.TempDir("", "test-bucket")
-	require.NoError(t, err)
+func setupGoCloudFileBucket(t *testing.T, scheme string) (m *blob.URLMux, bucketDir string) {
+	tmpDir := t.TempDir()
 
 	mux := new(blob.URLMux)
 	fake := &dirOpener{tmpDir: tmpDir}
 	mux.RegisterBucket(scheme, fake)
-	cleanup = func() {
-		os.RemoveAll(tmpDir)
-	}
 
-	return mux, tmpDir, cleanup
+	return mux, tmpDir
 }
 
 func goCloudObjectExists(t *testing.T, bucketDir string, objectName string) {
@@ -280,12 +304,54 @@ func testCacheUploadWithCustomHeaders(w http.ResponseWriter, r *http.Request) {
 		http.Error(w, "500 Wrong x-ms-blob-type header", http.StatusInternalServerError)
 	}
 
-	if r.Header.Get("Last-Modified") != "" {
-		http.Error(w, "500 Unexpected Last-Modified header included", http.StatusInternalServerError)
+	if r.Header.Get("Last-Modified") == "" {
+		http.Error(w, "500 Expected Last-Modified header included", http.StatusInternalServerError)
 	}
 }
 
 func writeTestFile(t *testing.T, fileName string) {
-	err := ioutil.WriteFile(fileName, nil, 0600)
+	err := os.WriteFile(fileName, nil, 0600)
 	require.NoError(t, err, "Writing file:", fileName)
 }
+
+func TestCacheArchiverUploadedSize(t *testing.T) {
+	tests := map[string]struct {
+		limit    int
+		exceeded bool
+	}{
+		"no-limit":    {limit: 0, exceeded: false},
+		"above-limit": {limit: 10, exceeded: true},
+		"equal-limit": {limit: 22, exceeded: false},
+		"below-limit": {limit: 25, exceeded: false},
+	}
+
+	for tn, tc := range tests {
+		t.Run(tn, func(t *testing.T) {
+			defer logrus.SetOutput(logrus.StandardLogger().Out)
+			defer testHelpers.MakeFatalToPanic()()
+
+			var buf bytes.Buffer
+			logrus.SetOutput(&buf)
+
+			ts := httptest.NewServer(http.HandlerFunc(testCacheBaseUploadHandler))
+			defer ts.Close()
+
+			defer os.Remove(cacheArchiverArchive)
+			cmd := helpers.CacheArchiverCommand{
+				File:                   cacheArchiverArchive,
+				MaxUploadedArchiveSize: int64(tc.limit),
+				URL:                    ts.URL + "/cache.zip",
+				Timeout:                0,
+			}
+			assert.NotPanics(t, func() {
+				cmd.Execute(nil)
+			})
+
+			if tc.exceeded {
+				require.Contains(t, buf.String(), "too big")
+			} else {
+				require.NotContains(t, buf.String(), "too big")
+			}
+		})
+	}
+}
