@@ -10,9 +10,12 @@ import (
 	"os"
 	"os/signal"
 	"runtime"
+	"sync"
 	"syscall"
 	"time"
 
+	"github.com/prometheus/client_golang/prometheus/collectors"
+
 	"github.com/kardianos/service"
 	"github.com/prometheus/client_golang/prometheus"
 	"github.com/prometheus/client_golang/prometheus/promhttp"
@@ -30,6 +33,16 @@ import (
 	"gitlab.com/gitlab-org/gitlab-runner/session"
 )
 
+const (
+	workerSlotOperationStarted = "started"
+	workerSlotOperationStopped = "stopped"
+)
+
+const (
+	workerProcessingFailureOther          = "other"
+	workerProcessingFailureNoFreeExecutor = "no_free_executor"
+)
+
 var (
 	concurrentDesc = prometheus.NewDesc(
 		"gitlab_runner_concurrent",
@@ -41,16 +54,36 @@ var (
 	limitDesc = prometheus.NewDesc(
 		"gitlab_runner_limit",
 		"The current value of concurrent setting",
-		[]string{"runner"},
+		[]string{"runner", "system_id"},
 		nil,
 	)
 )
 
+type runAtTask interface {
+	cancel()
+}
+
+type runAtTimerTask struct {
+	timer *time.Timer
+}
+
+func (t *runAtTimerTask) cancel() {
+	t.timer.Stop()
+}
+
+func runAt(t time.Time, f func()) runAtTask {
+	timer := time.AfterFunc(time.Until(t), f)
+	task := runAtTimerTask{
+		timer: timer,
+	}
+	return &task
+}
+
 type RunCommand struct {
 	configOptionsWithListenAddress
 	network common.Network
-	healthHelper
 
+	healthHelper healthHelper
 	buildsHelper buildsHelper
 
 	ServiceName      string `short:"n" long:"service" description:"Use different names for different services"`
@@ -61,8 +94,8 @@ type RunCommand struct {
 	sentryLogHook     sentry.LogHook
 	prometheusLogHook prometheus_helper.LogHook
 
-	failuresCollector               *prometheus_helper.FailuresCollector
-	networkRequestStatusesCollector prometheus.Collector
+	failuresCollector    *prometheus_helper.FailuresCollector
+	apiRequestsCollector prometheus.Collector
 
 	sessionServer *session.Server
 
@@ -83,10 +116,22 @@ type RunCommand struct {
 	// server.
 	stopSignal os.Signal
 
+	// configReloaded is used to notify that the config has been reloaded
+	configReloaded chan int
+
 	// runFinished is used to notify that run() did finish
 	runFinished chan bool
 
-	currentWorkers int
+	currentWorkers       int
+	reloadConfigInterval time.Duration
+
+	runAt func(time.Time, func()) runAtTask
+
+	runnerWorkerSlots             prometheus.Gauge
+	runnerWorkersFeeds            *prometheus.CounterVec
+	runnerWorkersFeedFailures     *prometheus.CounterVec
+	runnerWorkerSlotOperations    *prometheus.CounterVec
+	runnerWorkerProcessingFailure *prometheus.CounterVec
 }
 
 func (mr *RunCommand) log() *logrus.Entry {
@@ -101,11 +146,14 @@ func (mr *RunCommand) Start(_ service.Service) error {
 	mr.abortBuilds = make(chan os.Signal)
 	mr.runInterruptSignal = make(chan os.Signal, 1)
 	mr.reloadSignal = make(chan os.Signal, 1)
+	mr.configReloaded = make(chan int, 1)
 	mr.runFinished = make(chan bool, 1)
 	mr.stopSignals = make(chan os.Signal)
 
 	mr.log().Info("Starting multi-runner from ", mr.ConfigFile, "...")
 
+	mr.setupInternalMetrics()
+
 	userModeWarning(false)
 
 	if len(mr.WorkingDirectory) > 0 {
@@ -115,19 +163,145 @@ func (mr *RunCommand) Start(_ service.Service) error {
 		}
 	}
 
-	err := mr.loadConfig()
+	err := mr.reloadConfig()
 	if err != nil {
 		return err
 	}
 
+	config := mr.getConfig()
+	for _, runner := range config.Runners {
+		mr.runnerWorkersFeeds.WithLabelValues(runner.ShortDescription(), runner.Name, runner.GetSystemID()).Add(0)
+		mr.runnerWorkersFeedFailures.
+			WithLabelValues(runner.ShortDescription(), runner.Name, runner.GetSystemID()).
+			Add(0)
+		mr.runnerWorkerProcessingFailure.
+			WithLabelValues(
+				workerProcessingFailureOther,
+				runner.ShortDescription(), runner.Name, runner.GetSystemID(),
+			).
+			Add(0)
+		mr.runnerWorkerProcessingFailure.
+			WithLabelValues(
+				workerProcessingFailureNoFreeExecutor,
+				runner.ShortDescription(), runner.Name, runner.GetSystemID(),
+			).
+			Add(0)
+	}
+	mr.runnerWorkerSlots.Set(0)
+	mr.runnerWorkerSlotOperations.WithLabelValues(workerSlotOperationStarted).Add(0)
+	mr.runnerWorkerSlotOperations.WithLabelValues(workerSlotOperationStopped).Add(0)
+
 	// Start should not block. Do the actual work async.
 	go mr.run()
 
 	return nil
 }
 
-func (mr *RunCommand) loadConfig() error {
-	err := mr.configOptions.loadConfig()
+func (mr *RunCommand) setupInternalMetrics() {
+	mr.runnerWorkersFeeds = prometheus.NewCounterVec(
+		prometheus.CounterOpts{
+			Name: "gitlab_runner_worker_feeds_total",
+			Help: "Total number of times that runner worker is fed to the main loop",
+		},
+		[]string{"runner", "runner_name", "system_id"},
+	)
+
+	mr.runnerWorkersFeedFailures = prometheus.NewCounterVec(
+		prometheus.CounterOpts{
+			Name: "gitlab_runner_worker_feed_failures_total",
+			Help: "Total number of times that runner worker feeding have failed",
+		},
+		[]string{"runner", "runner_name", "system_id"},
+	)
+
+	mr.runnerWorkerSlots = prometheus.NewGauge(prometheus.GaugeOpts{
+		Name: "gitlab_runner_worker_slots_number",
+		Help: "Current number of runner worker slots",
+	})
+
+	mr.runnerWorkerSlotOperations = prometheus.NewCounterVec(
+		prometheus.CounterOpts{
+			Name: "gitlab_runner_worker_slot_operations_total",
+			Help: "Total number of runner workers slot operations (starting and stopping slots)",
+		},
+		[]string{"operation"},
+	)
+
+	mr.runnerWorkerProcessingFailure = prometheus.NewCounterVec(
+		prometheus.CounterOpts{
+			Name: "gitlab_runner_worker_processing_failures_total",
+			Help: "Total number of failures when processing runner worker",
+		},
+		[]string{"failure_type", "runner", "runner_name", "system_id"},
+	)
+}
+
+func nextRunnerToReset(config *common.Config) (*common.RunnerConfig, time.Time) {
+	var runnerToReset *common.RunnerConfig
+	var runnerResetTime time.Time
+
+	for _, runner := range config.Runners {
+		if runner.TokenExpiresAt.IsZero() {
+			continue
+		}
+
+		expirationInterval := runner.TokenExpiresAt.Sub(runner.TokenObtainedAt)
+		resetTime := runner.TokenObtainedAt.Add(
+			time.Duration(common.TokenResetIntervalFactor * float64(expirationInterval.Nanoseconds())),
+		)
+		if runnerToReset == nil || resetTime.Before(runnerResetTime) {
+			runnerToReset = runner
+			runnerResetTime = resetTime
+		}
+	}
+
+	return runnerToReset, runnerResetTime
+}
+
+func (mr *RunCommand) resetRunnerTokens() {
+	for mr.resetOneRunnerToken() {
+		// Handling runner tokens resetting - one by one - until mr.runFinished
+		// reports that mr.run() have been finished
+	}
+}
+
+func (mr *RunCommand) resetOneRunnerToken() bool {
+	var task runAtTask
+	runnerResetCh := make(chan *common.RunnerConfig)
+
+	config := mr.getConfig()
+	runnerToReset, runnerResetTime := nextRunnerToReset(config)
+	if runnerToReset != nil {
+		task = mr.runAt(runnerResetTime, func() {
+			runnerResetCh <- runnerToReset
+		})
+	}
+
+	select {
+	case runner := <-runnerResetCh:
+		if common.ResetToken(mr.network, &runner.RunnerCredentials, runner.GetSystemID(), "") {
+			err := mr.saveConfig()
+			if err != nil {
+				mr.log().WithError(err).Errorln("Failed to save config")
+			}
+		}
+
+	case <-mr.runFinished:
+		if task != nil {
+			task.cancel()
+		}
+		return false
+	case <-mr.configReloaded:
+		if task != nil {
+			task.cancel()
+		}
+	}
+
+	return true
+}
+
+func (mr *RunCommand) reloadConfig() error {
+	err := mr.loadConfig()
 	if err != nil {
 		return err
 	}
@@ -140,17 +314,20 @@ func (mr *RunCommand) loadConfig() error {
 
 	// pass user to execute scripts as specific user
 	if mr.User != "" {
+		mr.configMutex.Lock()
 		mr.config.User = mr.User
+		mr.configMutex.Unlock()
 	}
 
-	mr.healthy = nil
+	config := mr.getConfig()
+	mr.healthHelper.healthy = nil
 	mr.log().Println("Configuration loaded")
-	mr.log().Debugln(helpers.ToYAML(mr.config))
+	mr.log().Debugln(helpers.ToYAML(config))
 
 	// initialize sentry
-	if mr.config.SentryDSN != nil {
+	if config.SentryDSN != nil {
 		var err error
-		mr.sentryLogHook, err = sentry.NewLogHook(*mr.config.SentryDSN)
+		mr.sentryLogHook, err = sentry.NewLogHook(*config.SentryDSN)
 		if err != nil {
 			mr.log().WithError(err).Errorln("Sentry failure")
 		}
@@ -158,14 +335,17 @@ func (mr *RunCommand) loadConfig() error {
 		mr.sentryLogHook = sentry.LogHook{}
 	}
 
+	mr.configReloaded <- 1
+
 	return nil
 }
 
 func (mr *RunCommand) updateLoggingConfiguration() error {
 	reloadNeeded := false
 
-	if mr.config.LogLevel != nil && !log.Configuration().IsLevelSetWithCli() {
-		err := log.Configuration().SetLevel(*mr.config.LogLevel)
+	config := mr.getConfig()
+	if config.LogLevel != nil && !log.Configuration().IsLevelSetWithCli() {
+		err := log.Configuration().SetLevel(*config.LogLevel)
 		if err != nil {
 			return err
 		}
@@ -173,8 +353,8 @@ func (mr *RunCommand) updateLoggingConfiguration() error {
 		reloadNeeded = true
 	}
 
-	if mr.config.LogFormat != nil && !log.Configuration().IsFormatSetWithCli() {
-		err := log.Configuration().SetFormat(*mr.config.LogFormat)
+	if config.LogFormat != nil && !log.Configuration().IsFormatSetWithCli() {
+		err := log.Configuration().SetFormat(*config.LogFormat)
 		if err != nil {
 			return err
 		}
@@ -200,9 +380,13 @@ func (mr *RunCommand) run() {
 	mr.setupMetricsAndDebugServer()
 	mr.setupSessionServer()
 
+	go mr.resetRunnerTokens()
+
 	runners := make(chan *common.RunnerConfig)
 	go mr.feedRunners(runners)
 
+	mr.initUsedExecutorProviders()
+
 	signal.Notify(mr.stopSignals, syscall.SIGQUIT, syscall.SIGTERM, os.Interrupt)
 	signal.Notify(mr.reloadSignal, syscall.SIGHUP)
 
@@ -226,20 +410,59 @@ func (mr *RunCommand) run() {
 		}
 	}
 
-	// Wait for workers to shutdown
-	for mr.currentWorkers > 0 {
-		stopWorker <- true
-		mr.currentWorkers--
-	}
+	// Wait for workers to shut down
+	mr.stopWorkers(stopWorker)
+	mr.log().Info("All workers stopped.")
 
-	mr.log().Info("All workers stopped. Can exit now")
+	mr.shutdownUsedExecutorProviders()
+	mr.log().Info("All executor providers shut down.")
 
 	close(mr.runFinished)
+
+	mr.log().Info("Can exit now!")
+}
+
+func (mr *RunCommand) initUsedExecutorProviders() {
+	mr.log().Info("Initializing executor providers")
+
+	for _, provider := range common.GetExecutorProviders() {
+		managedProvider, ok := provider.(common.ManagedExecutorProvider)
+		if ok {
+			managedProvider.Init()
+		}
+	}
+}
+
+func (mr *RunCommand) shutdownUsedExecutorProviders() {
+	shutdownTimeout := mr.config.GetShutdownTimeout()
+
+	logger := mr.log().WithField("shutdown-timeout", shutdownTimeout)
+	logger.Info("Shutting down executor providers")
+
+	ctx, cancelFn := context.WithTimeout(context.Background(), shutdownTimeout)
+	defer cancelFn()
+
+	wg := new(sync.WaitGroup)
+	for _, provider := range common.GetExecutorProviders() {
+		managedProvider, ok := provider.(common.ManagedExecutorProvider)
+		if ok {
+			wg.Add(1)
+			go func(_ common.ManagedExecutorProvider) {
+				defer wg.Done()
+				managedProvider.Shutdown(ctx)
+			}(managedProvider)
+		}
+	}
+
+	wg.Wait()
+
+	if ctx.Err() != nil {
+		logger.Warn("Executor providers shutdown timeout exceeded")
+	}
 }
 
 func (mr *RunCommand) setupMetricsAndDebugServer() {
 	listenAddress, err := mr.listenAddress()
-
 	if err != nil {
 		mr.log().Errorf("invalid listen address: %s", err.Error())
 		return
@@ -279,9 +502,13 @@ func (mr *RunCommand) serveMetrics(mux *http.ServeMux) {
 	registry := prometheus.NewRegistry()
 	// Metrics about the runner's business logic.
 	registry.MustRegister(&mr.buildsHelper)
+	// Metrics about runner workers health
+	registry.MustRegister(&mr.healthHelper)
+	// Metrics about configuration file accessing
+	registry.MustRegister(mr.configAccessCollector)
 	registry.MustRegister(mr)
 	// Metrics about API connections
-	registry.MustRegister(mr.networkRequestStatusesCollector)
+	registry.MustRegister(mr.apiRequestsCollector)
 	// Metrics about jobs failures
 	registry.MustRegister(mr.failuresCollector)
 	// Metrics about catched errors
@@ -289,9 +516,9 @@ func (mr *RunCommand) serveMetrics(mux *http.ServeMux) {
 	// Metrics about the program's build version.
 	registry.MustRegister(common.AppVersion.NewMetricsCollector())
 	// Go-specific metrics about the process (GC stats, goroutines, etc.).
-	registry.MustRegister(prometheus.NewGoCollector())
+	registry.MustRegister(collectors.NewGoCollector())
 	// Go-unrelated process metrics (memory usage, file descriptors, etc.).
-	registry.MustRegister(prometheus.NewProcessCollector(prometheus.ProcessCollectorOpts{}))
+	registry.MustRegister(collectors.NewProcessCollector(collectors.ProcessCollectorOpts{}))
 
 	// Register all executor provider collectors
 	for _, provider := range common.GetExecutorProviders() {
@@ -300,7 +527,18 @@ func (mr *RunCommand) serveMetrics(mux *http.ServeMux) {
 		}
 	}
 
-	mux.Handle("/metrics", promhttp.HandlerFor(registry, promhttp.HandlerOpts{}))
+	// restrictHTTPMethods should be used on all promhttp handlers
+	// In this specific instance, the handler is unintrumented, so isn't as
+	// important. But in the future, if any other promhttp handlers are added
+	// they too should be wrapped and restriced.
+	// https://gitlab.com/gitlab-org/gitlab-runner/-/issues/27194
+	mux.Handle(
+		"/metrics",
+		restrictHTTPMethods(
+			promhttp.HandlerFor(registry, promhttp.HandlerOpts{}),
+			http.MethodGet, http.MethodHead,
+		),
+	)
 }
 
 func (mr *RunCommand) serveDebugData(mux *http.ServeMux) {
@@ -315,8 +553,27 @@ func (mr *RunCommand) servePprof(mux *http.ServeMux) {
 	mux.HandleFunc("/debug/pprof/trace", pprof.Trace)
 }
 
+// restrictHTTPMethods wraps a http.Handler and returns a http.Handler that
+// restricts methods only to those provided.
+func restrictHTTPMethods(handler http.Handler, methods ...string) http.Handler {
+	supported := map[string]struct{}{}
+	for _, method := range methods {
+		supported[method] = struct{}{}
+	}
+
+	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+		if _, ok := supported[r.Method]; !ok {
+			http.Error(w, http.StatusText(http.StatusMethodNotAllowed), http.StatusMethodNotAllowed)
+			return
+		}
+
+		handler.ServeHTTP(w, r)
+	})
+}
+
 func (mr *RunCommand) setupSessionServer() {
-	if mr.config.SessionServer.ListenAddress == "" {
+	config := mr.getConfig()
+	if config.SessionServer.ListenAddress == "" {
 		mr.log().Info("[session_server].listen_address not defined, session endpoints disabled")
 		return
 	}
@@ -324,9 +581,9 @@ func (mr *RunCommand) setupSessionServer() {
 	var err error
 	mr.sessionServer, err = session.NewServer(
 		session.ServerConfig{
-			AdvertiseAddress: mr.config.SessionServer.AdvertiseAddress,
-			ListenAddress:    mr.config.SessionServer.ListenAddress,
-			ShutdownTimeout:  common.ShutdownTimeout * time.Second,
+			AdvertiseAddress: config.SessionServer.AdvertiseAddress,
+			ListenAddress:    config.SessionServer.ListenAddress,
+			ShutdownTimeout:  mr.config.GetShutdownTimeout(),
 		},
 		mr.log(),
 		certificate.X509Generator{},
@@ -344,7 +601,7 @@ func (mr *RunCommand) setupSessionServer() {
 	}()
 
 	mr.log().
-		WithField("address", mr.config.SessionServer.ListenAddress).
+		WithField("address", config.SessionServer.ListenAddress).
 		Info("Session server listening")
 }
 
@@ -357,7 +614,7 @@ func (mr *RunCommand) setupSessionServer() {
 func (mr *RunCommand) feedRunners(runners chan *common.RunnerConfig) {
 	for mr.stopSignal == nil {
 		mr.log().Debugln("Feeding runners to channel")
-		config := mr.config
+		config := mr.getConfig()
 
 		// If no runners wait full interval to test again
 		if len(config.Runners) == 0 {
@@ -380,10 +637,13 @@ func (mr *RunCommand) feedRunners(runners chan *common.RunnerConfig) {
 }
 
 func (mr *RunCommand) feedRunner(runner *common.RunnerConfig, runners chan *common.RunnerConfig) {
-	if !mr.isHealthy(runner.UniqueID()) {
+	if !mr.healthHelper.isHealthy(runner) {
+		mr.runnerWorkersFeedFailures.WithLabelValues(runner.ShortDescription(), runner.Name, runner.GetSystemID()).Inc()
 		return
 	}
 
+	mr.runnerWorkersFeeds.WithLabelValues(runner.ShortDescription(), runner.Name, runner.GetSystemID()).Inc()
+	mr.log().WithField("runner", runner.ShortDescription()).Debugln("Feeding runner to channel")
 	runners <- runner
 }
 
@@ -404,27 +664,41 @@ func (mr *RunCommand) processRunners(id int, stopWorker chan bool, runners chan
 		WithField("worker", id).
 		Debugln("Starting worker")
 
+	mr.runnerWorkerSlotOperations.WithLabelValues(workerSlotOperationStarted).Inc()
+
 	for mr.stopSignal == nil {
 		select {
 		case runner := <-runners:
 			err := mr.processRunner(id, runner, runners)
 			if err != nil {
-				mr.log().
+				logger := mr.log().
 					WithFields(logrus.Fields{
 						"runner":   runner.ShortDescription(),
 						"executor": runner.Executor,
-					}).
-					WithError(err).
-					Warn("Failed to process runner")
-			}
+					}).WithError(err)
 
-			// force GC cycle after processing build
-			runtime.GC()
+				l := logger.Warn
+				failureType := workerProcessingFailureOther
+
+				var NoFreeExecutorError *common.NoFreeExecutorError
+				if errors.As(err, &NoFreeExecutorError) {
+					l = logger.Debug
+					failureType = workerProcessingFailureNoFreeExecutor
+				}
+
+				l("Failed to process runner")
+				mr.runnerWorkerProcessingFailure.
+					WithLabelValues(failureType, runner.ShortDescription(), runner.Name, runner.GetSystemID()).
+					Inc()
+			}
 
 		case <-stopWorker:
 			mr.log().
 				WithField("worker", id).
 				Debugln("Stopping worker")
+
+			mr.runnerWorkerSlotOperations.WithLabelValues(workerSlotOperationStopped).Inc()
+
 			return
 		}
 	}
@@ -437,51 +711,68 @@ func (mr *RunCommand) processRunners(id int, stopWorker chan bool, runners chan
 // GitLab instance and finally creates and finishes the job.
 // To speed-up jobs handling before starting the job this method "requeues" the runner to another
 // worker (by feeding the channel normally handled by feedRunners).
-func (mr *RunCommand) processRunner(
-	id int,
-	runner *common.RunnerConfig,
-	runners chan *common.RunnerConfig,
-) (err error) {
+func (mr *RunCommand) processRunner(id int, runner *common.RunnerConfig, runners chan *common.RunnerConfig) error {
+	mr.log().WithField("runner", runner.ShortDescription()).Debugln("Processing runner")
+
 	provider := common.GetExecutorProvider(runner.Executor)
 	if provider == nil {
-		return
+		mr.log().
+			WithField("runner", runner.ShortDescription()).
+			Errorf("Executor %q is not known; marking Runner as unhealthy", runner.Executor)
+		mr.healthHelper.markHealth(runner, false)
+
+		return nil
 	}
 
+	mr.log().WithField("runner", runner.ShortDescription()).Debug("Acquiring executor from provider")
 	executorData, err := provider.Acquire(runner)
 	if err != nil {
 		return fmt.Errorf("failed to update executor: %w", err)
 	}
 	defer provider.Release(runner, executorData)
 
+	mr.log().WithField("runner", runner.ShortDescription()).Debug("Acquiring job slot")
 	if !mr.buildsHelper.acquireBuild(runner) {
 		logrus.WithFields(logrus.Fields{
 			"runner": runner.ShortDescription(),
 			"worker": id,
 		}).Debug("Failed to request job, runner limit met")
-		return
+
+		return nil
 	}
 	defer mr.buildsHelper.releaseBuild(runner)
 
+	return mr.processBuildOnRunner(runner, runners, provider, executorData)
+}
+
+func (mr *RunCommand) processBuildOnRunner(
+	runner *common.RunnerConfig,
+	runners chan *common.RunnerConfig,
+	provider common.ExecutorProvider,
+	executorData common.ExecutorData,
+) error {
 	buildSession, sessionInfo, err := mr.createSession(provider)
 	if err != nil {
-		return
+		return err
 	}
 
 	// Receive a new build
 	trace, jobData, err := mr.requestJob(runner, sessionInfo)
 	if err != nil || jobData == nil {
-		return
+		return err
 	}
 	defer func() { mr.traceOutcome(trace, err) }()
 
 	// Create a new build
 	build, err := common.NewBuild(*jobData, runner, mr.abortBuilds, executorData)
 	if err != nil {
-		return
+		return err
 	}
 	build.Session = buildSession
 	build.ArtifactUploader = mr.network.UploadRawArtifacts
 
+	trace.SetDebugModeEnabled(build.IsDebugModeEnabled())
+
 	// Add build to list of builds to assign numbers
 	mr.buildsHelper.addBuild(build)
 	defer mr.buildsHelper.removeBuild(build)
@@ -491,7 +782,7 @@ func (mr *RunCommand) processRunner(
 	mr.requeueRunner(runner, runners)
 
 	// Process a build
-	return build.Run(mr.config, trace)
+	return build.Run(mr.getConfig(), trace)
 }
 
 func (mr *RunCommand) traceOutcome(trace common.JobTrace, err error) {
@@ -537,6 +828,7 @@ func (mr *RunCommand) requestJob(
 	runner *common.RunnerConfig,
 	sessionInfo *common.SessionInfo,
 ) (common.JobTrace, *common.JobResponse, error) {
+	mr.log().WithField("runner", runner.ShortDescription()).Debug("Acquiring request slot")
 	if !mr.buildsHelper.acquireRequest(runner) {
 		mr.log().WithField("runner", runner.ShortDescription()).
 			Debugln("Failed to request job: runner requestConcurrency meet")
@@ -545,7 +837,7 @@ func (mr *RunCommand) requestJob(
 	defer mr.buildsHelper.releaseRequest(runner)
 
 	jobData, healthy := mr.doJobRequest(context.Background(), runner, sessionInfo)
-	mr.makeHealthy(runner.UniqueID(), healthy)
+	mr.healthHelper.markHealth(runner, healthy)
 
 	if jobData == nil {
 		return nil, nil, nil
@@ -621,7 +913,8 @@ func (mr *RunCommand) requeueRunner(runner *common.RunnerConfig, runners chan *c
 // for the fact that `concurrent` defines the upper number of jobs that can be concurrently handled
 // by GitLab Runner process.
 func (mr *RunCommand) updateWorkers(workerIndex *int, startWorker chan int, stopWorker chan bool) os.Signal {
-	concurrentLimit := mr.config.Concurrent
+	config := mr.getConfig()
+	concurrentLimit := config.Concurrent
 
 	if concurrentLimit < 1 {
 		mr.log().Fatalln("Concurrent is less than 1 - no jobs will be processed")
@@ -636,6 +929,7 @@ func (mr *RunCommand) updateWorkers(workerIndex *int, startWorker chan int, stop
 			return signaled
 		}
 		mr.currentWorkers--
+		mr.runnerWorkerSlots.Set(float64(mr.currentWorkers))
 	}
 
 	for mr.currentWorkers < concurrentLimit {
@@ -647,22 +941,32 @@ func (mr *RunCommand) updateWorkers(workerIndex *int, startWorker chan int, stop
 			return signaled
 		}
 		mr.currentWorkers++
+		mr.runnerWorkerSlots.Set(float64(mr.currentWorkers))
+
 		*workerIndex++
 	}
 
 	return nil
 }
 
+func (mr *RunCommand) stopWorkers(stopWorker chan bool) {
+	for mr.currentWorkers > 0 {
+		stopWorker <- true
+		mr.currentWorkers--
+		mr.runnerWorkerSlots.Set(float64(mr.currentWorkers))
+	}
+}
+
 func (mr *RunCommand) updateConfig() os.Signal {
 	select {
-	case <-time.After(common.ReloadConfigInterval * time.Second):
+	case <-time.After(mr.reloadConfigInterval):
 		err := mr.checkConfig()
 		if err != nil {
 			mr.log().Errorln("Failed to load config", err)
 		}
 
 	case <-mr.reloadSignal:
-		err := mr.loadConfig()
+		err := mr.reloadConfig()
 		if err != nil {
 			mr.log().Errorln("Failed to load config", err)
 		}
@@ -680,15 +984,16 @@ func (mr *RunCommand) checkConfig() (err error) {
 		return err
 	}
 
-	if !mr.config.ModTime.Before(info.ModTime()) {
+	config := mr.getConfig()
+	if !config.ModTime.Before(info.ModTime()) {
 		return nil
 	}
 
-	err = mr.loadConfig()
+	err = mr.reloadConfig()
 	if err != nil {
 		mr.log().Errorln("Failed to load config", err)
 		// don't reload the same file
-		mr.config.ModTime = info.ModTime()
+		config.ModTime = info.ModTime()
 		return
 	}
 	return nil
@@ -735,7 +1040,7 @@ func (mr *RunCommand) Stop(_ service.Service) error {
 
 	mr.log().
 		WithError(err).
-		Warning("Graceful shutdown not finished properly")
+		Warning(`Graceful shutdown not finished properly. To gracefully clean up running plugins please use SIGQUIT (ctrl-\) instead of SIGINT (ctrl-c)`)
 
 	err = mr.handleForcefulShutdown()
 	if err == nil {
@@ -795,20 +1100,21 @@ func (mr *RunCommand) handleGracefulShutdown() error {
 // It calls mr.abortAllBuilds which will broadcast abort signal which finally
 // ends with jobs termination.
 // Next it waits for one of the following events:
-// 1. Another signal was sent to process, which is handled as force exit and
-//    triggers exit of the method and finally process termination without
-//    waiting for anything else.
-// 2. ShutdownTimeout is exceeded. If waiting for shutdown will take more than
-//    defined time, the process will be forceful terminated just like in the
-//    case when second signal is sent.
-// 3. mr.runFinished was closed, which means that all termination was done
-//    properly.
+//  1. Another signal was sent to process, which is handled as force exit and
+//     triggers exit of the method and finally process termination without
+//     waiting for anything else.
+//  2. ShutdownTimeout is exceeded. If waiting for shutdown will take more than
+//     defined time, the process will be forceful terminated just like in the
+//     case when second signal is sent.
+//  3. mr.runFinished was closed, which means that all termination was done
+//     properly.
 //
 // After this method exits, Stop returns it error and finally the
 // `github.com/kardianos/service` service mechanism will finish
 // process execution.
 func (mr *RunCommand) handleForcefulShutdown() error {
 	mr.log().
+		WithField("shutdown-timeout", mr.config.GetShutdownTimeout()).
 		WithField("StopSignal", mr.stopSignal).
 		Warning("Starting forceful shutdown")
 
@@ -821,7 +1127,7 @@ func (mr *RunCommand) handleForcefulShutdown() error {
 			mr.log().WithField("stop-signal", mr.stopSignal).Warning("[handleForcefulShutdown] received stop signal")
 			return fmt.Errorf("forced exit with stop signal: %v", mr.stopSignal)
 
-		case <-time.After(common.ShutdownTimeout * time.Second):
+		case <-time.After(mr.config.GetShutdownTimeout()):
 			return errors.New("shutdown timed out")
 
 		case <-mr.runFinished:
@@ -891,11 +1197,17 @@ func (mr *RunCommand) runWait() {
 func (mr *RunCommand) Describe(ch chan<- *prometheus.Desc) {
 	ch <- concurrentDesc
 	ch <- limitDesc
+
+	mr.runnerWorkersFeeds.Describe(ch)
+	mr.runnerWorkersFeedFailures.Describe(ch)
+	mr.runnerWorkerSlots.Describe(ch)
+	mr.runnerWorkerSlotOperations.Describe(ch)
+	mr.runnerWorkerProcessingFailure.Describe(ch)
 }
 
 // Collect implements prometheus.Collector.
 func (mr *RunCommand) Collect(ch chan<- prometheus.Metric) {
-	config := mr.config
+	config := mr.getConfig()
 
 	ch <- prometheus.MustNewConstMetric(
 		concurrentDesc,
@@ -909,23 +1221,32 @@ func (mr *RunCommand) Collect(ch chan<- prometheus.Metric) {
 			prometheus.GaugeValue,
 			float64(runner.Limit),
 			runner.ShortDescription(),
+			runner.SystemIDState.GetSystemID(),
 		)
 	}
+
+	mr.runnerWorkersFeeds.Collect(ch)
+	mr.runnerWorkersFeedFailures.Collect(ch)
+	mr.runnerWorkerSlots.Collect(ch)
+	mr.runnerWorkerSlotOperations.Collect(ch)
+	mr.runnerWorkerProcessingFailure.Collect(ch)
 }
 
 func init() {
-	requestStatusesCollector := network.NewAPIRequestStatusesMap()
+	apiRequestsCollector := network.NewAPIRequestsCollector()
 
-	common.RegisterCommand2(
-		"run",
-		"run multi runner service",
-		&RunCommand{
-			ServiceName:                     defaultServiceName,
-			network:                         network.NewGitLabClientWithRequestStatusesMap(requestStatusesCollector),
-			networkRequestStatusesCollector: requestStatusesCollector,
-			prometheusLogHook:               prometheus_helper.NewLogHook(),
-			failuresCollector:               prometheus_helper.NewFailuresCollector(),
-			buildsHelper:                    newBuildsHelper(),
-		},
-	)
+	cmd := &RunCommand{
+		ServiceName:          defaultServiceName,
+		network:              network.NewGitLabClientWithAPIRequestsCollector(apiRequestsCollector),
+		apiRequestsCollector: apiRequestsCollector,
+		prometheusLogHook:    prometheus_helper.NewLogHook(),
+		failuresCollector:    prometheus_helper.NewFailuresCollector(),
+		healthHelper:         newHealthHelper(),
+		buildsHelper:         newBuildsHelper(),
+		runAt:                runAt,
+		reloadConfigInterval: common.ReloadConfigInterval,
+	}
+	cmd.configAccessCollector = newConfigAccessCollector()
+
+	common.RegisterCommand2("run", "run multi runner service", cmd)
 }
