@@ -1,15 +1,16 @@
-// +build integration
+//go:build integration
 
 package commands_test
 
 import (
 	"bufio"
 	"fmt"
-	"io/ioutil"
 	"os"
+	"regexp"
 	"runtime"
 	"strings"
 	"testing"
+	"time"
 
 	"github.com/sirupsen/logrus"
 	"github.com/sirupsen/logrus/hooks/test"
@@ -18,6 +19,8 @@ import (
 	"github.com/stretchr/testify/require"
 	"github.com/urfave/cli"
 
+	clihelpers "gitlab.com/gitlab-org/golang-cli-helpers"
+
 	"gitlab.com/gitlab-org/gitlab-runner/commands"
 	"gitlab.com/gitlab-org/gitlab-runner/common"
 	_ "gitlab.com/gitlab-org/gitlab-runner/executors/docker"
@@ -26,7 +29,6 @@ import (
 	"gitlab.com/gitlab-org/gitlab-runner/helpers"
 	"gitlab.com/gitlab-org/gitlab-runner/helpers/ssh"
 	"gitlab.com/gitlab-org/gitlab-runner/shells"
-	clihelpers "gitlab.com/gitlab-org/golang-cli-helpers"
 )
 
 const osTypeWindows = "windows"
@@ -76,7 +78,7 @@ func TestAccessLevelSetting(t *testing.T) {
 				"--access-level", string(testCase.accessLevel),
 			}
 
-			_, output, err := testRegisterCommandRun(t, network, nil, arguments...)
+			_, output, err := testRegisterCommandRun(t, network, nil, "test-runner-token", arguments...)
 
 			if testCase.failureExpected {
 				assert.EqualError(t, err, "command error: Given access-level is not valid. "+
@@ -96,9 +98,7 @@ func TestAskRunnerOverrideDefaultsForExecutors(t *testing.T) {
 	executors := []string{
 		"kubernetes",
 		"docker+machine",
-		"docker-ssh+machine",
 		"docker",
-		"docker-ssh",
 		"ssh",
 		"custom",
 		"parallels",
@@ -114,15 +114,183 @@ func TestAskRunnerOverrideDefaultsForExecutors(t *testing.T) {
 	}
 }
 
+func isValidToken(systemID string) bool {
+	ok, _ := regexp.MatchString("^[sr]_[0-9a-zA-Z]{12}$", systemID)
+	return ok
+}
+
+func TestAskRunnerUsingRunnerTokenOverrideDefaults(t *testing.T) {
+	const executor = "docker"
+
+	basicValidation := func(s *commands.RegisterCommand) {
+		assert.Equal(t, "http://gitlab.example.com/", s.URL)
+		assert.Equal(t, "glrt-testtoken", s.Token)
+		assert.Equal(t, executor, s.RunnerSettings.Executor)
+	}
+	expectedParamsFn := func(p common.RunnerCredentials) bool {
+		return p.URL == "http://gitlab.example.com/" && p.Token == "glrt-testtoken"
+	}
+
+	tests := map[string]struct {
+		answers        []string
+		arguments      []string
+		validate       func(s *commands.RegisterCommand)
+		expectedParams func(common.RunnerCredentials) bool
+	}{
+		"basic answers": {
+			answers: append([]string{
+				"http://gitlab.example.com/",
+				"glrt-testtoken",
+				"name",
+			}, executorAnswers(t, executor)...),
+			validate:       basicValidation,
+			expectedParams: expectedParamsFn,
+		},
+		"basic arguments, accepting provided": {
+			answers: make([]string, 9),
+			arguments: append(
+				executorCmdLineArgs(t, executor),
+				"--url", "http://gitlab.example.com/",
+				"-r", "glrt-testtoken",
+				"--name", "name",
+			),
+			validate:       basicValidation,
+			expectedParams: expectedParamsFn,
+		},
+		"basic arguments override": {
+			answers: append(
+				[]string{"http://gitlab.example2.com/", "glrt-testtoken2", "new-name", executor},
+				executorOverrideAnswers(t, executor)...,
+			),
+			arguments: append(
+				executorCmdLineArgs(t, executor),
+				"--url", "http://gitlab.example.com/",
+				"-r", "glrt-testtoken",
+				"--name", "name",
+			),
+			validate: func(s *commands.RegisterCommand) {
+				assert.Equal(t, "http://gitlab.example2.com/", s.URL)
+				assert.Equal(t, "glrt-testtoken2", s.Token)
+				assert.Equal(t, "new-name", s.Name)
+				assert.Equal(t, executor, s.RunnerSettings.Executor)
+				require.NotNil(t, s.RunnerSettings.Docker)
+				assert.Equal(t, "nginx:latest", s.RunnerSettings.Docker.Image)
+			},
+			expectedParams: func(p common.RunnerCredentials) bool {
+				return p.URL == "http://gitlab.example2.com/" && p.Token == "glrt-testtoken2"
+			},
+		},
+	}
+
+	for tn, tc := range tests {
+		t.Run(tn, func(t *testing.T) {
+			network := new(common.MockNetwork)
+			defer network.AssertExpectations(t)
+
+			network.On("VerifyRunner", mock.MatchedBy(tc.expectedParams), mock.MatchedBy(isValidToken)).
+				Return(&common.VerifyRunnerResponse{
+					ID:    12345,
+					Token: "glrt-testtoken",
+				}).
+				Once()
+
+			cmd := commands.NewRegisterCommandForTest(
+				bufio.NewReader(strings.NewReader(strings.Join(tc.answers, "\n")+"\n")),
+				network,
+			)
+
+			app := cli.NewApp()
+			app.Commands = []cli.Command{
+				{
+					Name:   "register",
+					Action: cmd.Execute,
+					Flags:  clihelpers.GetFlagsFromStruct(cmd),
+				},
+			}
+
+			hook := test.NewGlobal()
+			args := append(tc.arguments, "--leave-runner")
+			err := app.Run(append([]string{"runner", "register"}, args...))
+			output := commands.GetLogrusOutput(t, hook)
+
+			assert.NoError(t, err)
+			tc.validate(cmd)
+			assert.Contains(t, output, "Runner registered successfully.")
+		})
+	}
+}
+
+func TestAskRunnerUsingRunnerTokenOverrideForbiddenDefaults(t *testing.T) {
+	tests := map[string]interface{}{
+		"--access-level":     "not_protected",
+		"--run-untagged":     true,
+		"--maximum-timeout":  1,
+		"--paused":           true,
+		"--tag-list":         "tag",
+		"--maintenance-note": "note",
+	}
+
+	for tn, tc := range tests {
+		t.Run(tn, func(t *testing.T) {
+			network := new(common.MockNetwork)
+			network.On("VerifyRunner", mock.Anything, mock.MatchedBy(isValidToken)).
+				Return(&common.VerifyRunnerResponse{
+					ID:    1,
+					Token: "glrt-testtoken",
+				}).
+				Once()
+
+			answers := make([]string, 4)
+			arguments := append(
+				executorCmdLineArgs(t, "shell"),
+				"--url", "http://gitlab.example.com/",
+				"-r", "glrt-testtoken",
+				tn, fmt.Sprintf("%v", tc),
+			)
+
+			cmd := commands.NewRegisterCommandForTest(
+				bufio.NewReader(strings.NewReader(strings.Join(answers, "\n")+"\n")),
+				network,
+			)
+
+			app := cli.NewApp()
+			app.Commands = []cli.Command{
+				{
+					Name:   "register",
+					Action: cmd.Execute,
+					Flags:  clihelpers.GetFlagsFromStruct(cmd),
+				},
+			}
+
+			fatalReceived := false
+			helpers.MakeFatalToPanic()
+
+			defer func() {
+				if r := recover(); r != nil {
+					entry, _ := r.(*logrus.Entry)
+					assert.Contains(t, entry.Message, "cannot be specified when registering with a runner token.")
+					fatalReceived = true
+				}
+				assert.True(t, fatalReceived, "fatal error expected")
+			}()
+
+			_ = app.Run(append([]string{"runner", "register"}, arguments...))
+
+			t.Fail()
+		})
+	}
+}
+
 func testRegisterCommandRun(
 	t *testing.T,
 	network common.Network,
 	env []kv,
+	token string,
 	args ...string,
 ) (content, output string, err error) {
 	config := &common.RunnerConfig{
 		RunnerCredentials: common.RunnerCredentials{
-			Token: "test-runner-token",
+			Token: token,
 		},
 		RunnerSettings: common.RunnerSettings{
 			Executor: "shell",
@@ -166,7 +334,7 @@ func testRegisterCommandRun(
 		},
 	}
 
-	configFile, err := ioutil.TempFile("", "config.toml")
+	configFile, err := os.CreateTemp("", "config.toml")
 	require.NoError(t, err)
 
 	err = configFile.Close()
@@ -179,7 +347,7 @@ func testRegisterCommandRun(
 		"-n",
 		"--config", configFile.Name(),
 		"--url", "http://gitlab.example.com/",
-		"--registration-token", "test-registration-token",
+		"--registration-token", token,
 	}, args...)
 	if !contains(args, "--executor") {
 		args = append(args, "--executor", config.RunnerSettings.Executor)
@@ -187,12 +355,12 @@ func testRegisterCommandRun(
 
 	commandErr := app.Run(args)
 
-	fileContent, err := ioutil.ReadFile(configFile.Name())
+	fileContent, err := os.ReadFile(configFile.Name())
 	require.NoError(t, err)
 
 	err = commandErr
 
-	return string(fileContent), "", err
+	return string(fileContent), output, err
 }
 
 func contains(args []string, s string) bool {
@@ -221,43 +389,49 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 				"test-registration-token",
 				"name",
 				"tag,list",
+				"basic notes",
 			}, executorAnswers(t, executor)...),
 			validate: basicValidation,
 			expectedParams: func(p common.RegisterRunnerParameters) bool {
 				return p == common.RegisterRunnerParameters{
-					Description: "name",
-					Tags:        "tag,list",
-					Locked:      true,
-					Active:      true,
+					Description:     "name",
+					MaintenanceNote: "basic notes",
+					Tags:            "tag,list",
+					Locked:          true,
+					Paused:          false,
 				}
 			},
 		},
 		"basic arguments, accepting provided": {
-			answers: make([]string, 10),
+			answers: make([]string, 11),
 			arguments: append(
 				executorCmdLineArgs(t, executor),
 				"--url", "http://gitlab.example.com/",
 				"-r", "test-registration-token",
 				"--name", "name",
 				"--tag-list", "tag,list",
+				"--maintenance-note", "maintainer notes",
 				"--paused",
 				"--locked=false",
 			),
 			validate: basicValidation,
 			expectedParams: func(p common.RegisterRunnerParameters) bool {
 				return p == common.RegisterRunnerParameters{
-					Description: "name",
-					Tags:        "tag,list",
+					Description:     "name",
+					MaintenanceNote: "maintainer notes",
+					Tags:            "tag,list",
+					Paused:          true,
 				}
 			},
 		},
 		"basic arguments override": {
-			answers: append([]string{"", "", "new-name", "", ""}, executorOverrideAnswers(t, executor)...),
+			answers: append([]string{"", "", "new-name", "", "maintainer notes", ""}, executorOverrideAnswers(t, executor)...),
 			arguments: append(
 				executorCmdLineArgs(t, executor),
 				"--url", "http://gitlab.example.com/",
 				"-r", "test-registration-token",
 				"--name", "name",
+				"--maintenance-note", "notes",
 				"--tag-list", "tag,list",
 				"--paused",
 				"--locked=false",
@@ -267,8 +441,10 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 			},
 			expectedParams: func(p common.RegisterRunnerParameters) bool {
 				return p == common.RegisterRunnerParameters{
-					Description: "new-name",
-					Tags:        "tag,list",
+					Description:     "new-name",
+					MaintenanceNote: "maintainer notes",
+					Tags:            "tag,list",
+					Paused:          true,
 				}
 			},
 		},
@@ -278,6 +454,7 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 				"test-registration-token",
 				"name",
 				"",
+				"",
 			}, executorAnswers(t, executor)...),
 			validate: basicValidation,
 			expectedParams: func(p common.RegisterRunnerParameters) bool {
@@ -285,7 +462,7 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 					Description: "name",
 					RunUntagged: true,
 					Locked:      true,
-					Active:      true,
+					Paused:      false,
 				}
 			},
 		},
@@ -295,6 +472,7 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 				"test-registration-token",
 				"name",
 				"",
+				"",
 			}, executorAnswers(t, executor)...),
 			arguments: []string{"--run-untagged"},
 			validate:  basicValidation,
@@ -303,7 +481,7 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 					Description: "name",
 					RunUntagged: true,
 					Locked:      true,
-					Active:      true,
+					Paused:      false,
 				}
 			},
 		},
@@ -313,6 +491,7 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 				"test-registration-token",
 				"name",
 				"tag,list",
+				"",
 			}, executorAnswers(t, executor)...),
 			arguments: []string{"--run-untagged"},
 			validate:  basicValidation,
@@ -322,7 +501,7 @@ func testAskRunnerOverrideDefaultsForExecutor(t *testing.T, executor string) {
 					Tags:        "tag,list",
 					RunUntagged: true,
 					Locked:      true,
-					Active:      true,
+					Paused:      false,
 				}
 			},
 		},
@@ -389,15 +568,6 @@ func assertExecutorDefaultValues(t *testing.T, executor string, s *commands.Regi
 		assert.NotNil(t, s.RunnerSettings.Machine)
 		require.NotNil(t, s.RunnerSettings.Docker)
 		assert.Equal(t, "busybox:latest", s.RunnerSettings.Docker.Image)
-	case "docker-ssh":
-		assertDefaultSSHLogin(t, s.RunnerSettings.SSH)
-		require.NotNil(t, s.RunnerSettings.Docker)
-		assert.Equal(t, "busybox:latest", s.RunnerSettings.Docker.Image)
-	case "docker-ssh+machine":
-		assert.NotNil(t, s.RunnerSettings.Machine)
-		assertDefaultSSHLogin(t, s.RunnerSettings.SSH)
-		require.NotNil(t, s.RunnerSettings.Docker)
-		assert.Equal(t, "busybox:latest", s.RunnerSettings.Docker.Image)
 	case "ssh":
 		assertDefaultSSHLogin(t, s.RunnerSettings.SSH)
 		assertDefaultSSHServer(t, s.RunnerSettings.SSH)
@@ -452,15 +622,6 @@ func assertExecutorOverridenValues(t *testing.T, executor string, s *commands.Re
 		assert.NotNil(t, s.RunnerSettings.Machine)
 		require.NotNil(t, s.RunnerSettings.Docker)
 		assert.Equal(t, "nginx:latest", s.RunnerSettings.Docker.Image)
-	case "docker-ssh":
-		assertOverridenSSHLogin(t, s.RunnerSettings.SSH)
-		require.NotNil(t, s.RunnerSettings.Docker)
-		assert.Equal(t, "nginx:latest", s.RunnerSettings.Docker.Image)
-	case "docker-ssh+machine":
-		assert.NotNil(t, s.Machine)
-		assertOverridenSSHLogin(t, s.RunnerSettings.SSH)
-		require.NotNil(t, s.RunnerSettings.Docker)
-		assert.Equal(t, "nginx:latest", s.RunnerSettings.Docker.Image)
 	case "ssh":
 		assertOverridenSSHLogin(t, s.RunnerSettings.SSH)
 		assertOverridenSSHServer(t, s.RunnerSettings.SSH)
@@ -492,17 +653,15 @@ func assertOverridenSSHServer(t *testing.T, sshCfg *ssh.Config) {
 
 func executorAnswers(t *testing.T, executor string) []string {
 	values := map[string][]string{
-		"kubernetes":         {executor},
-		"custom":             {executor},
-		"shell":              {executor},
-		"docker":             {executor, "busybox:latest"},
-		"docker-windows":     {executor, "mcr.microsoft.com/windows/servercore:YYH1"},
-		"docker+machine":     {executor, "busybox:latest"},
-		"docker-ssh":         {executor, "busybox:latest", "user", "password", "/home/user/.ssh/id_rsa"},
-		"docker-ssh+machine": {executor, "busybox:latest", "user", "password", "/home/user/.ssh/id_rsa"},
-		"ssh":                {executor, "gitlab.example.com", "22", "user", "password", "/home/user/.ssh/id_rsa"},
-		"parallels":          {executor, "parallels-vm-name", "gitlab.example.com", "22"},
-		"virtualbox":         {executor, "virtualbox-vm-name", "user", "password", "/home/user/.ssh/id_rsa"},
+		"kubernetes":     {executor},
+		"custom":         {executor},
+		"shell":          {executor},
+		"docker":         {executor, "busybox:latest"},
+		"docker-windows": {executor, "mcr.microsoft.com/windows/servercore:YYH1"},
+		"docker+machine": {executor, "busybox:latest"},
+		"ssh":            {executor, "gitlab.example.com", "22", "user", "password", "/home/user/.ssh/id_rsa"},
+		"parallels":      {executor, "parallels-vm-name", "gitlab.example.com", "22"},
+		"virtualbox":     {executor, "virtualbox-vm-name", "user", "password", "/home/user/.ssh/id_rsa"},
 	}
 
 	answers, ok := values[executor]
@@ -514,17 +673,15 @@ func executorAnswers(t *testing.T, executor string) []string {
 
 func executorOverrideAnswers(t *testing.T, executor string) []string {
 	values := map[string][]string{
-		"kubernetes":         {""},
-		"custom":             {""},
-		"shell":              {""},
-		"docker":             {"nginx:latest"},
-		"docker-windows":     {"mcr.microsoft.com/windows/servercore:YYH2"},
-		"docker+machine":     {"nginx:latest"},
-		"docker-ssh":         {"nginx:latest", "root", "admin", "/root/.ssh/id_rsa"},
-		"docker-ssh+machine": {"nginx:latest", "root", "admin", "/root/.ssh/id_rsa"},
-		"ssh":                {"ssh.gitlab.example.com", "8822", "root", "admin", "/root/.ssh/id_rsa"},
-		"parallels":          {"override-parallels-vm-name", "ssh.gitlab.example.com", "8822"},
-		"virtualbox":         {"override-virtualbox-vm-name", "root", "admin", "/root/.ssh/id_rsa"},
+		"kubernetes":     {""},
+		"custom":         {""},
+		"shell":          {""},
+		"docker":         {"nginx:latest"},
+		"docker-windows": {"mcr.microsoft.com/windows/servercore:YYH2"},
+		"docker+machine": {"nginx:latest"},
+		"ssh":            {"ssh.gitlab.example.com", "8822", "root", "admin", "/root/.ssh/id_rsa"},
+		"parallels":      {"override-parallels-vm-name", "ssh.gitlab.example.com", "8822"},
+		"virtualbox":     {"override-virtualbox-vm-name", "root", "admin", "/root/.ssh/id_rsa"},
 	}
 
 	answers, ok := values[executor]
@@ -542,16 +699,6 @@ func executorCmdLineArgs(t *testing.T, executor string) []string {
 		"docker":         {"--executor", executor, "--docker-image", "busybox:latest"},
 		"docker-windows": {"--executor", executor, "--docker-image", "mcr.microsoft.com/windows/servercore:YYH1"},
 		"docker+machine": {"--executor", executor, "--docker-image", "busybox:latest"},
-		"docker-ssh": {
-			"--executor", executor, "--docker-image", "busybox:latest", "--ssh-user", "user",
-			"--ssh-password", "password",
-			"--ssh-identity-file", "/home/user/.ssh/id_rsa",
-		},
-		"docker-ssh+machine": {
-			"--executor", executor, "--docker-image", "busybox:latest", "--ssh-user", "user",
-			"--ssh-password", "password",
-			"--ssh-identity-file", "/home/user/.ssh/id_rsa",
-		},
 		"ssh": {
 			"--executor", executor, "--ssh-host", "gitlab.example.com", "--ssh-port", "22", "--ssh-user", "user",
 			"--ssh-password", "password", "--ssh-identity-file", "/home/user/.ssh/id_rsa",
@@ -589,6 +736,7 @@ func TestExecute_MergeConfigTemplate(t *testing.T) {
 
 		baseOutputConfigFmt = `concurrent = 1
 check_interval = 0
+shutdown_timeout = 0
 
 [session_server]
   session_timeout = 1800
@@ -596,14 +744,14 @@ check_interval = 0
 [[runners]]
   name = %q
   url = "http://gitlab.example.com/"
+  id = 0
   token = "test-runner-token"
+  token_obtained_at = %s
+  token_expires_at = 0001-01-01T00:00:00Z
   executor = "shell"
   shell = "pwsh"
-  [runners.custom_build_dir]
   [runners.cache]
-    [runners.cache.s3]
-    [runners.cache.gcs]
-    [runners.cache.azure]
+    MaxUploadedArchiveSize = 0
 `
 	)
 
@@ -649,6 +797,7 @@ check_interval = 0
 			errExpected: false,
 			expectedFileContentFmt: `concurrent = 1
 check_interval = 0
+shutdown_timeout = 0
 
 [session_server]
   session_timeout = 1800
@@ -656,50 +805,20 @@ check_interval = 0
 [[runners]]
   name = %q
   url = "http://gitlab.example.com/"
+  id = 0
   token = "test-runner-token"
+  token_obtained_at = %s
+  token_expires_at = 0001-01-01T00:00:00Z
   executor = "shell"
   shell = "pwsh"
-  [runners.custom_build_dir]
   [runners.cache]
-    [runners.cache.s3]
-    [runners.cache.gcs]
-    [runners.cache.azure]
-  [runners.kubernetes]
-    host = ""
-    bearer_token_overwrite_allowed = false
-    image = ""
-    namespace = ""
-    namespace_overwrite_allowed = ""
-    service_account_overwrite_allowed = ""
-    pod_annotations_overwrite_allowed = ""
-    [runners.kubernetes.affinity]
-    [runners.kubernetes.pod_security_context]
-    [runners.kubernetes.build_container_security_context]
-    [runners.kubernetes.helper_container_security_context]
-    [runners.kubernetes.service_container_security_context]
-    [runners.kubernetes.volumes]
-
-      [[runners.kubernetes.volumes.empty_dir]]
-        name = "empty_dir"
-        mount_path = "/path/to/empty_dir"
-        medium = "Memory"
-    [runners.kubernetes.dns_config]
-    [runners.kubernetes.container_lifecycle]
+    MaxUploadedArchiveSize = 0
 `,
 		},
 		"incorrect config template merge": {
-			configTemplate: configTemplateMergeInvalidConfiguration,
-			networkAssertions: func(n *common.MockNetwork) {
-				n.On("RegisterRunner", mock.Anything, mock.Anything).
-					Return(&common.RegisterRunnerResponse{
-						Token: "test-runner-token",
-					}).
-					Once()
-				n.On("UnregisterRunner", mock.Anything).
-					Return(true).
-					Once()
-			},
-			errExpected: true,
+			configTemplate:    configTemplateMergeInvalidConfiguration,
+			networkAssertions: func(n *common.MockNetwork) {},
+			errExpected:       true,
 		},
 	}
 
@@ -727,7 +846,7 @@ check_interval = 0
 
 			tt.networkAssertions(network)
 
-			fileContent, _, err := testRegisterCommandRun(t, network, nil, args...)
+			fileContent, _, err := testRegisterCommandRun(t, network, nil, "test-runner-token", args...)
 			if tt.errExpected {
 				require.Error(t, err)
 				return
@@ -736,28 +855,50 @@ check_interval = 0
 			require.NoError(t, err)
 			name, err := os.Hostname()
 			require.NoError(t, err)
-			assert.Equal(t, fmt.Sprintf(tt.expectedFileContentFmt, name), fileContent)
+			assert.Equal(t, fmt.Sprintf(tt.expectedFileContentFmt, name, commands.RegisterTimeNowDate.Format(time.RFC3339)), fileContent)
 		})
 	}
 }
 
 func TestUnregisterOnFailure(t *testing.T) {
 	tests := map[string]struct {
+		token                 string
 		leaveRunner           bool
 		registrationFails     bool
 		expectsLeftRegistered bool
 	}{
+		"ui created runner, verification succeeds, runner left registered": {
+			token:                 "glrt-test-runner-token",
+			leaveRunner:           false,
+			registrationFails:     false,
+			expectsLeftRegistered: true,
+		},
+		"ui created runner, verification fails, LeaveRunner is false, runner machine is unregistered": {
+			token:                 "glrt-test-runner-token",
+			leaveRunner:           false,
+			registrationFails:     true,
+			expectsLeftRegistered: false,
+		},
+		"ui created runner, verification fails, LeaveRunner is true, runner machine left registered": {
+			token:                 "glrt-test-runner-token",
+			leaveRunner:           true,
+			registrationFails:     true,
+			expectsLeftRegistered: true,
+		},
 		"registration succeeds, runner left registered": {
+			token:                 "test-runner-token",
 			leaveRunner:           false,
 			registrationFails:     false,
 			expectsLeftRegistered: true,
 		},
 		"registration fails, LeaveRunner is false, runner is unregistered": {
+			token:                 "test-runner-token",
 			leaveRunner:           false,
 			registrationFails:     true,
 			expectsLeftRegistered: false,
 		},
 		"registration fails, LeaveRunner is true, runner left registered": {
+			token:                 "test-runner-token",
 			leaveRunner:           true,
 			registrationFails:     true,
 			expectsLeftRegistered: true,
@@ -766,22 +907,37 @@ func TestUnregisterOnFailure(t *testing.T) {
 
 	for testName, testCase := range tests {
 		t.Run(testName, func(t *testing.T) {
+			runnerUICreated := strings.HasPrefix(testCase.token, "glrt-")
 			network := new(common.MockNetwork)
 			defer network.AssertExpectations(t)
 
-			const token = "test-runner-token"
-			network.On("RegisterRunner", mock.Anything, mock.Anything).
-				Return(&common.RegisterRunnerResponse{
-					Token: token,
-				}).
-				Once()
+			if runnerUICreated {
+				network.On("VerifyRunner", mock.Anything, mock.MatchedBy(isValidToken)).
+					Return(&common.VerifyRunnerResponse{
+						ID:    1,
+						Token: testCase.token,
+					}).
+					Once()
+			} else {
+				network.On("RegisterRunner", mock.Anything, mock.Anything).
+					Return(&common.RegisterRunnerResponse{
+						Token: testCase.token,
+					}).
+					Once()
+			}
 			if !testCase.expectsLeftRegistered {
 				credsMocker := mock.MatchedBy(func(credentials common.RunnerCredentials) bool {
-					return credentials.Token == token
+					return credentials.Token == testCase.token
 				})
-				network.On("UnregisterRunner", credsMocker).
-					Return(true).
-					Once()
+				if runnerUICreated {
+					network.On("UnregisterRunnerManager", credsMocker, mock.Anything).
+						Return(true).
+						Once()
+				} else {
+					network.On("UnregisterRunner", credsMocker).
+						Return(true).
+						Once()
+				}
 			}
 
 			var arguments []string
@@ -789,7 +945,10 @@ func TestUnregisterOnFailure(t *testing.T) {
 				arguments = append(arguments, "--leave-runner")
 			}
 
-			answers := []string{"https://gitlab.com/", token, "description", ""}
+			answers := []string{"https://gitlab.com/", testCase.token, "description"}
+			if !runnerUICreated {
+				answers = append(answers, "", "")
+			}
 			if testCase.registrationFails {
 				defer func() { _ = recover() }()
 			} else {
@@ -820,13 +979,29 @@ func TestUnregisterOnFailure(t *testing.T) {
 func TestRegisterCommand(t *testing.T) {
 	type testCase struct {
 		condition       func() bool
+		token           string
 		arguments       []string
 		environment     []kv
 		expectedConfigs []string
 	}
 
 	testCases := map[string]testCase{
+		"runner ID is included in config": {
+			token: "glrt-test-runner-token",
+			arguments: []string{
+				"--name", "test-runner",
+			},
+			expectedConfigs: []string{`id = 12345`, `token = "glrt-test-runner-token"`},
+		},
+		"registration token is accepted": {
+			token: "test-runner-token",
+			arguments: []string{
+				"--name", "test-runner",
+			},
+			expectedConfigs: []string{`id = 12345`, `token = "test-runner-token"`},
+		},
 		"feature flags are included in config": {
+			token: "glrt-test-runner-token",
 			arguments: []string{
 				"--name", "test-runner",
 				"--feature-flags", "FF_TEST_1:true",
@@ -838,6 +1013,7 @@ func TestRegisterCommand(t *testing.T) {
 		},
 		"shell defaults to pwsh on Windows with shell executor": {
 			condition: func() bool { return runtime.GOOS == osTypeWindows },
+			token:     "glrt-test-runner-token",
 			arguments: []string{
 				"--name", "test-runner",
 				"--executor", "shell",
@@ -846,6 +1022,7 @@ func TestRegisterCommand(t *testing.T) {
 		},
 		"shell defaults to pwsh on Windows with docker-windows executor": {
 			condition: func() bool { return runtime.GOOS == osTypeWindows },
+			token:     "glrt-test-runner-token",
 			arguments: []string{
 				"--name", "test-runner",
 				"--executor", "docker-windows",
@@ -855,6 +1032,7 @@ func TestRegisterCommand(t *testing.T) {
 		},
 		"shell can be overridden to powershell on Windows with shell executor": {
 			condition: func() bool { return runtime.GOOS == osTypeWindows },
+			token:     "glrt-test-runner-token",
 			arguments: []string{
 				"--name", "test-runner",
 				"--executor", "shell",
@@ -864,6 +1042,7 @@ func TestRegisterCommand(t *testing.T) {
 		},
 		"shell can be overridden to powershell on Windows with docker-windows executor": {
 			condition: func() bool { return runtime.GOOS == osTypeWindows },
+			token:     "glrt-test-runner-token",
 			arguments: []string{
 				"--name", "test-runner",
 				"--executor", "docker-windows",
@@ -873,6 +1052,7 @@ func TestRegisterCommand(t *testing.T) {
 			expectedConfigs: []string{`shell = "powershell"`},
 		},
 		"kubernetes security context namespace": {
+			token: "glrt-test-runner-token",
 			arguments: []string{
 				"--executor", "kubernetes",
 			},
@@ -905,6 +1085,29 @@ func TestRegisterCommand(t *testing.T) {
         	add = ["NET_RAW, NET_RAW1"]`,
 			},
 		},
+		"s3 cache AuthenticationType arg": {
+			token: "glrt-test-runner-token",
+			arguments: []string{
+				"--cache-s3-authentication_type=iam",
+			},
+			expectedConfigs: []string{`
+		[runners.cache.s3]
+			AuthenticationType = "iam"
+			`},
+		},
+		"s3 cache AuthenticationType env": {
+			token: "glrt-test-runner-token",
+			environment: []kv{
+				{
+					key:   "CACHE_S3_AUTHENTICATION_TYPE",
+					value: "iam",
+				},
+			},
+			expectedConfigs: []string{`
+		[runners.cache.s3]
+			AuthenticationType = "iam"
+			`},
+		},
 	}
 
 	for tn, tc := range testCases {
@@ -916,13 +1119,23 @@ func TestRegisterCommand(t *testing.T) {
 			network := new(common.MockNetwork)
 			defer network.AssertExpectations(t)
 
-			network.On("RegisterRunner", mock.Anything, mock.Anything).
-				Return(&common.RegisterRunnerResponse{
-					Token: "test-runner-token",
-				}).
-				Once()
+			if strings.HasPrefix(tc.token, "glrt-") {
+				network.On("VerifyRunner", mock.Anything, mock.MatchedBy(isValidToken)).
+					Return(&common.VerifyRunnerResponse{
+						ID:    12345,
+						Token: tc.token,
+					}).
+					Once()
+			} else {
+				network.On("RegisterRunner", mock.Anything, mock.Anything).
+					Return(&common.RegisterRunnerResponse{
+						ID:    12345,
+						Token: tc.token,
+					}).
+					Once()
+			}
 
-			gotConfig, _, err := testRegisterCommandRun(t, network, tc.environment, tc.arguments...)
+			gotConfig, _, err := testRegisterCommandRun(t, network, tc.environment, tc.token, tc.arguments...)
 			require.NoError(t, err)
 
 			for _, expectedConfig := range tc.expectedConfigs {
@@ -931,3 +1144,72 @@ func TestRegisterCommand(t *testing.T) {
 		})
 	}
 }
+
+func TestRegisterTokenExpiresAt(t *testing.T) {
+	type testCase struct {
+		token          string
+		expiration     time.Time
+		expectedConfig string
+	}
+
+	testCases := map[string]testCase{
+		"no expiration": {
+			token: "test-runner-token",
+			expectedConfig: `token = "test-runner-token"
+				token_obtained_at = %s
+				token_expires_at = 0001-01-01T00:00:00Z`,
+		},
+		"token expiration": {
+			token:      "test-runner-token",
+			expiration: time.Date(2594, 7, 21, 15, 42, 53, 0, time.UTC),
+			expectedConfig: `token = "test-runner-token"
+				token_obtained_at = %s
+				token_expires_at = 2594-07-21T15:42:53Z`,
+		},
+		"no expiration with authentication token": {
+			token: "glrt-test-runner-token",
+			expectedConfig: `token = "glrt-test-runner-token"
+				token_obtained_at = %s
+				token_expires_at = 0001-01-01T00:00:00Z`,
+		},
+		"token expiration with authentication token": {
+			token:      "glrt-test-runner-token",
+			expiration: time.Date(2594, 7, 21, 15, 42, 53, 0, time.UTC),
+			expectedConfig: `token = "glrt-test-runner-token"
+				token_obtained_at = %s
+				token_expires_at = 2594-07-21T15:42:53Z`,
+		},
+	}
+
+	for tn, tc := range testCases {
+		t.Run(tn, func(t *testing.T) {
+			network := new(common.MockNetwork)
+			defer network.AssertExpectations(t)
+
+			if strings.HasPrefix(tc.token, "glrt-") {
+				network.On("VerifyRunner", mock.Anything, mock.MatchedBy(isValidToken)).
+					Return(&common.VerifyRunnerResponse{
+						ID:             12345,
+						Token:          tc.token,
+						TokenExpiresAt: tc.expiration,
+					}).
+					Once()
+			} else {
+				network.On("RegisterRunner", mock.Anything, mock.Anything).
+					Return(&common.RegisterRunnerResponse{
+						Token:          tc.token,
+						TokenExpiresAt: tc.expiration,
+					}).
+					Once()
+			}
+
+			gotConfig, _, err := testRegisterCommandRun(t, network, []kv{}, tc.token, "--name", "test-runner")
+			require.NoError(t, err)
+
+			assert.Contains(
+				t, spaceReplacer.Replace(gotConfig),
+				spaceReplacer.Replace(fmt.Sprintf(tc.expectedConfig, commands.RegisterTimeNowDate.Format(time.RFC3339))),
+			)
+		})
+	}
+}
