@@ -1,4 +1,4 @@
-// +build !integration
+//go:build !integration
 
 package common
 
@@ -9,6 +9,7 @@ import (
 
 	"github.com/BurntSushi/toml"
 	"github.com/stretchr/testify/assert"
+	"github.com/stretchr/testify/mock"
 	"github.com/stretchr/testify/require"
 	api "k8s.io/api/core/v1"
 
@@ -122,6 +123,16 @@ func TestCacheS3Config_AuthType(t *testing.T) {
 			},
 			authType: S3AuthTypeAccessKey,
 		},
+		"Explicitly set but lowercase auth type": {
+			s3: CacheS3Config{
+				AccessKey:          "TOKEN",
+				SecretKey:          "TOKEN",
+				AuthenticationType: "access-key",
+				BucketName:         "name",
+				BucketLocation:     "us-east-1a",
+			},
+			authType: S3AuthTypeAccessKey,
+		},
 		"Explicitly set invalid auth type": {
 			s3: CacheS3Config{
 				AccessKey:          "TOKEN",
@@ -211,7 +222,7 @@ func TestConfigParse(t *testing.T) {
 				[[runners.docker.services]]
 				name = 5
 			`,
-			expectedErr: "toml: cannot load TOML value of type int64 into a Go string",
+			expectedErr: "incompatible types: TOML value has type int64; destination has type string",
 		},
 		"parse Service as table int value alias": {
 			config: `
@@ -220,7 +231,7 @@ func TestConfigParse(t *testing.T) {
 				name = "svc1"
 				alias = 5
 			`,
-			expectedErr: "toml: cannot load TOML value of type int64 into a Go string",
+			expectedErr: "incompatible types: TOML value has type int64; destination has type string",
 		},
 		"parse Service runners.docker and runners.docker.services": {
 			config: `
@@ -238,7 +249,72 @@ func TestConfigParse(t *testing.T) {
 				assert.Equal(t, "image", config.Runners[0].Docker.Image)
 			},
 		},
-		//nolint:lll
+		"parse Service runners.docker.services environment": {
+			config: `
+				[[runners]]
+				[runners.docker]
+				[[runners.docker.services]]
+				name = "svc1"
+				environment = ["ENV1=value1", "ENV2=value2"]
+			`,
+			validateConfig: func(t *testing.T, config *Config) {
+				require.Equal(t, 1, len(config.Runners))
+				require.Equal(t, 1, len(config.Runners[0].Docker.Services))
+				require.Equal(t, 2, len(config.Runners[0].Docker.Services[0].Environment))
+				assert.Equal(t, "ENV1=value1", config.Runners[0].Docker.Services[0].Environment[0])
+				assert.Equal(t, "ENV2=value2", config.Runners[0].Docker.Services[0].Environment[1])
+			},
+		},
+		"parse Docker Container Labels with string key and value": {
+			config: `
+                        [[runners]]
+                                [runners.docker]
+                                        image = "image"
+                                        [runners.docker.container_labels]
+                                                "my.docker.TestContainerlabel1" = "TestContainerlabel-1"
+                `,
+			validateConfig: func(t *testing.T, config *Config) {
+				require.Equal(t, 1, len(config.Runners))
+
+				runner := config.Runners[0]
+				require.NotNil(t, runner.RunnerSettings.Docker.ContainerLabels)
+				require.NotNil(t, runner.RunnerSettings.Docker.ContainerLabels["my.docker.TestContainerlabel1"])
+				require.Equal(
+					t,
+					"TestContainerlabel-1",
+					runner.RunnerSettings.Docker.ContainerLabels["my.docker.TestContainerlabel1"],
+				)
+			},
+		},
+		"parse Docker Container Labels with integer key and value": {
+			config: `
+                        [[runners]]
+                                [runners.docker]
+                                        image = "image"
+                                        [runners.docker.container_labels]
+                                                5 = 5
+                `,
+			expectedErr: "incompatible types: TOML value has type int64; destination has type string",
+		},
+		"parse Docker Container Labels with integer value": {
+			config: `
+                        [[runners]]
+                                [runners.docker]
+                                        image = "image"
+                                        [runners.docker.container_labels]
+                                                "my.docker.TestContainerlabel1" = 5
+                `,
+			expectedErr: "incompatible types: TOML value has type int64; destination has type string",
+		},
+		"parse Docker Container Labels with integer key": {
+			config: `
+                        [[runners]]
+                                [runners.docker]
+                                        image = "image"
+                                        [runners.docker.container_labels]
+                                                5 = "TestContainerlabel-1"
+                `,
+		},
 		"check node affinities": {
 			config: `
 				[[runners]]
@@ -333,8 +409,6 @@ func TestConfigParse(t *testing.T) {
 				assert.Equal(t, []string{"e2e-az1"}, nodeAffinity.RequiredDuringSchedulingIgnoredDuringExecution.NodeSelectorTerms[1].MatchFields[0].Values)
 			},
 		},
-
-		//nolint:lll
 		"check pod affinities": {
 			config: `
 				[[runners]]
@@ -417,7 +491,6 @@ func TestConfigParse(t *testing.T) {
 				assert.Equal(t, []string{"S2"}, preferredMatchExp.Values)
 			},
 		},
-		//nolint:lll
 		"check pod anti affinities": {
 			config: `
 				[[runners]]
@@ -741,6 +814,32 @@ func TestConfigParse(t *testing.T) {
 				assert.Equal(t, "localhost", lifecycleCfg.PreStop.TCPSocket.Host)
 			},
 		},
+		"setting Priority Class to priority-1": {
+			config: `
+				[[runners]]
+					[runners.kubernetes]
+						priority_class_name = 'priority-1'
+			`,
+			validateConfig: func(t *testing.T, config *Config) {
+				require.Len(t, config.Runners, 1)
+
+				priorityClassName := config.Runners[0].Kubernetes.PriorityClassName
+				assert.Equal(t, "priority-1", priorityClassName)
+			},
+		},
+		"setting scheduler_name to foobar": {
+			config: `
+				[[runners]]
+					[runners.kubernetes]
+						scheduler_name = 'foobar'
+			`,
+			validateConfig: func(t *testing.T, config *Config) {
+				require.Len(t, config.Runners, 1)
+
+				schedulerName := config.Runners[0].Kubernetes.SchedulerName
+				assert.Equal(t, "foobar", schedulerName)
+			},
+		},
 	}
 
 	for tn, tt := range tests {
@@ -748,7 +847,7 @@ func TestConfigParse(t *testing.T) {
 			cfg := NewConfig()
 			_, err := toml.Decode(tt.config, cfg)
 			if tt.expectedErr != "" {
-				assert.EqualError(t, err, tt.expectedErr)
+				assert.ErrorContains(t, err, tt.expectedErr)
 				return
 			}
 
@@ -871,6 +970,14 @@ func TestService_ToImageDefinition(t *testing.T) {
 			service:       Service{Name: "name", Alias: "alias"},
 			expectedImage: Image{Name: "name", Alias: "alias"},
 		},
+		"only aliases": {
+			service:       Service{Alias: "alias-1 alias-2"},
+			expectedImage: Image{Alias: "alias-1 alias-2"},
+		},
+		"name and aliases": {
+			service:       Service{Name: "name", Alias: "alias-1 alias-2"},
+			expectedImage: Image{Name: "name", Alias: "alias-1 alias-2"},
+		},
 		"command specified": {
 			service:       Service{Name: "name", Command: []string{"executable", "param1", "param2"}},
 			expectedImage: Image{Name: "name", Command: []string{"executable", "param1", "param2"}},
@@ -891,6 +998,13 @@ func TestService_ToImageDefinition(t *testing.T) {
 				Entrypoint: []string{"executable", "param3", "param4"},
 			},
 		},
+		"environment specified": {
+			service: Service{Name: "name", Environment: []string{"ENV1=value1", "ENV2=value2"}},
+			expectedImage: Image{Name: "name", Variables: JobVariables{
+				{Key: "ENV1", Value: "value1", Internal: true},
+				{Key: "ENV2", Value: "value2", Internal: true},
+			}},
+		},
 	}
 
 	for tn, tt := range tests {
@@ -1151,6 +1265,109 @@ func TestDockerConfig_GetPullPolicies(t *testing.T) {
 	}
 }
 
+func TestDockerConfig_GetAllowedPullPolicies(t *testing.T) {
+	tests := map[string]struct {
+		config               DockerConfig
+		expectedPullPolicies []DockerPullPolicy
+		expectedErr          bool
+	}{
+		"nil allowed_pull_policies": {
+			config:               DockerConfig{},
+			expectedPullPolicies: []DockerPullPolicy{PullPolicyAlways},
+			expectedErr:          false,
+		},
+		"empty allowed_pull_policies": {
+			config:               DockerConfig{AllowedPullPolicies: []DockerPullPolicy{}},
+			expectedPullPolicies: []DockerPullPolicy{PullPolicyAlways},
+			expectedErr:          false,
+		},
+		"empty string allowed_pull_policies": {
+			config:      DockerConfig{AllowedPullPolicies: []DockerPullPolicy{""}},
+			expectedErr: true,
+		},
+		"known elements in allowed_pull_policies": {
+			config: DockerConfig{
+				AllowedPullPolicies: []DockerPullPolicy{PullPolicyAlways, PullPolicyNever},
+			},
+			expectedPullPolicies: []DockerPullPolicy{PullPolicyAlways, PullPolicyNever},
+			expectedErr:          false,
+		},
+		"invalid allowed_pull_policies": {
+			config:      DockerConfig{AllowedPullPolicies: []DockerPullPolicy{"invalid"}},
+			expectedErr: true,
+		},
+	}
+
+	for tn, tt := range tests {
+		t.Run(tn, func(t *testing.T) {
+			policies, err := tt.config.GetAllowedPullPolicies()
+
+			if tt.expectedErr {
+				assert.Error(t, err)
+				return
+			}
+
+			assert.NoError(t, err)
+			assert.Equal(t, tt.expectedPullPolicies, policies)
+		})
+	}
+}
+
+func TestKubernetesConfig_GetAllowedPullPolicies(t *testing.T) {
+	tests := map[string]struct {
+		config               KubernetesConfig
+		expectedPullPolicies []api.PullPolicy
+		expectedErr          bool
+	}{
+		"nil allowed_pull_policies": {
+			config:               KubernetesConfig{},
+			expectedPullPolicies: []api.PullPolicy{""},
+			expectedErr:          false,
+		},
+		"empty allowed_pull_policies": {
+			config: KubernetesConfig{
+				AllowedPullPolicies: []DockerPullPolicy{},
+			},
+			expectedPullPolicies: []api.PullPolicy{""},
+			expectedErr:          false,
+		},
+		"empty string allowed_pull_policies": {
+			config: KubernetesConfig{
+				AllowedPullPolicies: []DockerPullPolicy{""},
+			},
+			expectedPullPolicies: []api.PullPolicy{""},
+			expectedErr:          false,
+		},
+		"known elements in allowed_pull_policies": {
+			config: KubernetesConfig{
+				AllowedPullPolicies: []DockerPullPolicy{PullPolicyAlways, PullPolicyNever},
+			},
+			expectedPullPolicies: []api.PullPolicy{api.PullAlways, api.PullNever},
+			expectedErr:          false,
+		},
+		"invalid allowed_pull_policies": {
+			config: KubernetesConfig{
+				AllowedPullPolicies: []DockerPullPolicy{"invalid"},
+			},
+			expectedErr: true,
+		},
+	}
+
+	for tn, tt := range tests {
+		t.Run(tn, func(t *testing.T) {
+			policies, err := tt.config.GetAllowedPullPolicies()
+
+			if tt.expectedErr {
+				assert.Error(t, err)
+				return
+			}
+
+			assert.NoError(t, err)
+			assert.Equal(t, tt.expectedPullPolicies, policies)
+		})
+	}
+}
+
 func TestKubernetesConfig_GetPullPolicies(t *testing.T) {
 	tests := map[string]struct {
 		config               KubernetesConfig
@@ -1200,6 +1417,42 @@ func TestKubernetesConfig_GetPullPolicies(t *testing.T) {
 	}
 }
 
+func TestKubernetesConfig_ConvertFromDockerPullPolicy(t *testing.T) {
+	tests := map[string]struct {
+		config               KubernetesConfig
+		dockerPullPolicies   []DockerPullPolicy
+		expectedPullPolicies []api.PullPolicy
+		expectedErr          bool
+	}{
+		"valid list": {
+			config:               KubernetesConfig{},
+			dockerPullPolicies:   []DockerPullPolicy{PullPolicyAlways, PullPolicyIfNotPresent, PullPolicyNever},
+			expectedPullPolicies: []api.PullPolicy{api.PullAlways, api.PullIfNotPresent, api.PullNever},
+			expectedErr:          false,
+		},
+		"has an invalid pull policy": {
+			config:               KubernetesConfig{},
+			dockerPullPolicies:   []DockerPullPolicy{PullPolicyAlways, "invalid"},
+			expectedPullPolicies: []api.PullPolicy{""},
+			expectedErr:          true,
+		},
+	}
+
+	for tn, tt := range tests {
+		t.Run(tn, func(t *testing.T) {
+			policies, err := tt.config.ConvertFromDockerPullPolicy(tt.dockerPullPolicies)
+
+			if tt.expectedErr {
+				assert.Error(t, err)
+				return
+			}
+
+			assert.NoError(t, err)
+			assert.Equal(t, tt.expectedPullPolicies, policies)
+		})
+	}
+}
+
 func TestStringOrArray_UnmarshalTOML(t *testing.T) {
 	tests := map[string]struct {
 		toml           string
@@ -1430,6 +1683,70 @@ func TestContainerSecurityContext(t *testing.T) {
 				}
 			},
 		},
+		"SELinux type label - container security context": {
+			getSecurityContext: func(c *KubernetesConfig) *api.SecurityContext {
+				return c.GetContainerSecurityContext(KubernetesContainerSecurityContext{
+					SELinuxType: "spc_t",
+				})
+			},
+			getExpectedContainerSecurityContext: func() *api.SecurityContext {
+				return &api.SecurityContext{
+					SELinuxOptions: &api.SELinuxOptions{Type: "spc_t"},
+				}
+			},
+		},
+		"proc mount - blank": {
+			getSecurityContext: func(c *KubernetesConfig) *api.SecurityContext {
+				return c.GetContainerSecurityContext(KubernetesContainerSecurityContext{
+					ProcMount: "",
+				})
+			},
+			getExpectedContainerSecurityContext: func() *api.SecurityContext {
+				return &api.SecurityContext{
+					ProcMount: nil,
+				}
+			},
+		},
+		"proc mount - invalid": {
+			getSecurityContext: func(c *KubernetesConfig) *api.SecurityContext {
+				return c.GetContainerSecurityContext(KubernetesContainerSecurityContext{
+					ProcMount: "invalid",
+				})
+			},
+			getExpectedContainerSecurityContext: func() *api.SecurityContext {
+				return &api.SecurityContext{
+					ProcMount: nil,
+				}
+			},
+		},
+		"proc mount - default": {
+			getSecurityContext: func(c *KubernetesConfig) *api.SecurityContext {
+				return c.GetContainerSecurityContext(KubernetesContainerSecurityContext{
+					ProcMount: "default",
+				})
+			},
+			getExpectedContainerSecurityContext: func() *api.SecurityContext {
+				pm := api.DefaultProcMount
+
+				return &api.SecurityContext{
+					ProcMount: &pm,
+				}
+			},
+		},
+		"proc mount - unmasked": {
+			getSecurityContext: func(c *KubernetesConfig) *api.SecurityContext {
+				return c.GetContainerSecurityContext(KubernetesContainerSecurityContext{
+					ProcMount: "unmasked",
+				})
+			},
+			getExpectedContainerSecurityContext: func() *api.SecurityContext {
+				pm := api.UnmaskedProcMount
+
+				return &api.SecurityContext{
+					ProcMount: &pm,
+				}
+			},
+		},
 	}
 
 	for tn, tt := range tests {
@@ -1442,6 +1759,76 @@ func TestContainerSecurityContext(t *testing.T) {
 	}
 }
 
+func TestKubernetesPodSpecContents(t *testing.T) {
+	tests := map[string]struct {
+		patchPath     string
+		patchContents string
+		patchType     KubernetesPodSpecPatchType
+
+		expectedContents string
+		expectedType     KubernetesPodSpecPatchType
+		expectedErr      error
+	}{
+		"yaml to json": {
+			patchContents:    `hostname: "test"`,
+			expectedContents: `{"hostname":"test"}`,
+			expectedType:     PatchTypeStrategicMergePatchType,
+		},
+		"json without format to json": {
+			patchContents:    `{"hostname":"test"}`,
+			expectedContents: `{"hostname":"test"}`,
+			expectedType:     PatchTypeStrategicMergePatchType,
+		},
+		"json to json": {
+			patchContents:    `{"hostname": {"test": "value"}}`,
+			expectedContents: `{"hostname":{"test":"value"}}`,
+			expectedType:     PatchTypeStrategicMergePatchType,
+		},
+		"invalid json": {
+			patchContents: `{"hostname": {{}"test": "value"}}`,
+			expectedType:  PatchTypeStrategicMergePatchType,
+			expectedErr:   errPatchConversion,
+		},
+		"invalid yaml": {
+			patchContents: `[invalid yaml`,
+			expectedErr:   errPatchConversion,
+		},
+		"missing file": {
+			patchPath:   "missing/file",
+			expectedErr: errPatchFileFail,
+		},
+		"patch_path and patch ambiguous": {
+			patchPath:     "missing/file",
+			patchContents: `{"hostname": {"test": "value"}}`,
+			expectedErr:   errPatchAmbiguous,
+		},
+		"explicit patch type": {
+			patchContents:    `hostname: "test"`,
+			patchType:        PatchTypeMergePatchType,
+			expectedContents: `{"hostname":"test"}`,
+			expectedType:     PatchTypeMergePatchType,
+		},
+	}
+
+	for tn, tc := range tests {
+		t.Run(tn, func(t *testing.T) {
+			s := KubernetesPodSpec{
+				PatchPath: tc.patchPath,
+				Patch:     tc.patchContents,
+				PatchType: tc.patchType,
+			}
+			patchBytes, patchType, err := s.PodSpecPatch()
+			if tc.expectedErr != nil {
+				require.ErrorIs(t, err, tc.expectedErr)
+			} else {
+				require.NoError(t, err)
+				assert.Equal(t, tc.expectedContents, string(patchBytes))
+				assert.Equal(t, tc.expectedType, patchType)
+			}
+		})
+	}
+}
+
 func TestContainerSecurityCapabilities(t *testing.T) {
 	tests := map[string]struct {
 		getCapabilitiesFn    func(c *KubernetesConfig) *api.Capabilities
@@ -1703,3 +2090,60 @@ func TestKubernetesTerminationPeriod(t *testing.T) {
 		})
 	}
 }
+
+func Test_GetPullPolicySource(t *testing.T) {
+	tests := map[string]struct {
+		want              PullPolicySource
+		imagePullPolicies []DockerPullPolicy
+		pullPolicy        StringOrArray
+	}{
+		"gitlab-ci.yaml only": {
+			imagePullPolicies: []DockerPullPolicy{PullPolicyIfNotPresent},
+			pullPolicy:        StringOrArray{},
+			want:              PullPolicySourceGitLabCI,
+		},
+		"config.toml only": {
+			imagePullPolicies: []DockerPullPolicy{},
+			pullPolicy:        StringOrArray{PullPolicyIfNotPresent},
+			want:              PullPolicySourceRunner,
+		},
+		"both": {
+			imagePullPolicies: []DockerPullPolicy{PullPolicyIfNotPresent},
+			pullPolicy:        StringOrArray{PullPolicyNever},
+			want:              PullPolicySourceGitLabCI,
+		},
+		"not configured": {
+			imagePullPolicies: []DockerPullPolicy{},
+			pullPolicy:        StringOrArray{},
+			want:              PullPolicySourceDefault,
+		},
+	}
+
+	for name, tt := range tests {
+		t.Run(name, func(t *testing.T) {
+			assert.Equal(t, tt.want, GetPullPolicySource(tt.imagePullPolicies, tt.pullPolicy))
+		})
+	}
+}
+
+func TestConfig_SaveConfig(t *testing.T) {
+	const (
+		configFileName = "config-file"
+	)
+
+	oldTime := time.Now().Add(-1 * time.Hour)
+
+	cs := new(MockConfigSaver)
+	defer cs.AssertExpectations(t)
+
+	cs.On("Save", configFileName, mock.Anything).Return(nil).Once()
+
+	c := new(Config)
+	c.ModTime = oldTime
+	c.configSaver = cs
+
+	err := c.SaveConfig(configFileName)
+	require.NoError(t, err)
+
+	assert.NotEqual(t, oldTime, c.ModTime, "Expected ModTime field of Config struct to be updated")
+}
