@@ -57,11 +57,13 @@ const (
 	UploadForbidden
 	UploadFailed
 	UploadServiceUnavailable
+	UploadRedirected
 )
 
 const (
 	DownloadSucceeded DownloadState = iota
 	DownloadForbidden
+	DownloadUnauthorized
 	DownloadFailed
 	DownloadNotFound
 )
@@ -72,6 +74,7 @@ type FeaturesInfo struct {
 	Services                bool `json:"services"`
 	Artifacts               bool `json:"artifacts"`
 	Cache                   bool `json:"cache"`
+	FallbackCacheKeys       bool `json:"fallback_cache_keys"`
 	Shared                  bool `json:"shared"`
 	UploadMultipleArtifacts bool `json:"upload_multiple_artifacts"`
 	UploadRawArtifacts      bool `json:"upload_raw_artifacts"`
@@ -89,6 +92,8 @@ type FeaturesInfo struct {
 	VaultSecrets            bool `json:"vault_secrets"`
 	Cancelable              bool `json:"cancelable"`
 	ReturnExitCode          bool `json:"return_exit_code"`
+	ServiceVariables        bool `json:"service_variables"`
+	ServiceMultipleAliases  bool `json:"service_multiple_aliases"`
 }
 
 type ConfigInfo struct {
@@ -96,13 +101,14 @@ type ConfigInfo struct {
 }
 
 type RegisterRunnerParameters struct {
-	Description    string `json:"description,omitempty"`
-	Tags           string `json:"tag_list,omitempty"`
-	RunUntagged    bool   `json:"run_untagged"`
-	Locked         bool   `json:"locked"`
-	AccessLevel    string `json:"access_level,omitempty"`
-	MaximumTimeout int    `json:"maximum_timeout,omitempty"`
-	Active         bool   `json:"active"`
+	Description     string `json:"description,omitempty"`
+	MaintenanceNote string `json:"maintenance_note,omitempty"`
+	Tags            string `json:"tag_list,omitempty"`
+	RunUntagged     bool   `json:"run_untagged"`
+	Locked          bool   `json:"locked"`
+	AccessLevel     string `json:"access_level,omitempty"`
+	MaximumTimeout  int    `json:"maximum_timeout,omitempty"`
+	Paused          bool   `json:"paused"`
 }
 
 type RegisterRunnerRequest struct {
@@ -112,17 +118,41 @@ type RegisterRunnerRequest struct {
 }
 
 type RegisterRunnerResponse struct {
-	Token string `json:"token,omitempty"`
+	ID             int64     `json:"id,omitempty"`
+	Token          string    `json:"token,omitempty"`
+	TokenExpiresAt time.Time `json:"token_expires_at,omitempty"`
 }
 
 type VerifyRunnerRequest struct {
-	Token string `json:"token,omitempty"`
+	Token    string `json:"token,omitempty"`
+	SystemID string `json:"system_id,omitempty"`
+}
+
+type VerifyRunnerResponse struct {
+	ID             int64     `json:"id,omitempty"`
+	Token          string    `json:"token,omitempty"`
+	TokenExpiresAt time.Time `json:"token_expires_at,omitempty"`
 }
 
 type UnregisterRunnerRequest struct {
 	Token string `json:"token,omitempty"`
 }
 
+type UnregisterRunnerManagerRequest struct {
+	Token    string `json:"token,omitempty"`
+	SystemID string `json:"system_id"`
+}
+
+type ResetTokenRequest struct {
+	Token string `json:"token,omitempty"`
+}
+
+type ResetTokenResponse struct {
+	Token           string `json:"token,omitempty"`
+	TokenObtainedAt time.Time
+	TokenExpiresAt  time.Time `json:"token_expires_at,omitempty"`
+}
+
 type VersionInfo struct {
 	Name         string       `json:"name,omitempty"`
 	Version      string       `json:"version,omitempty"`
@@ -138,6 +168,7 @@ type VersionInfo struct {
 type JobRequest struct {
 	Info       VersionInfo  `json:"info,omitempty"`
 	Token      string       `json:"token,omitempty"`
+	SystemID   string       `json:"system_id,omitempty"`
 	LastUpdate string       `json:"last_update,omitempty"`
 	Session    *SessionInfo `json:"session,omitempty"`
 }
@@ -151,7 +182,7 @@ type SessionInfo struct {
 type JobInfo struct {
 	Name        string `json:"name"`
 	Stage       string `json:"stage"`
-	ProjectID   int    `json:"project_id"`
+	ProjectID   int64  `json:"project_id"`
 	ProjectName string `json:"project_name"`
 }
 
@@ -213,13 +244,17 @@ type Step struct {
 type Steps []Step
 
 type Image struct {
-	Name       string   `json:"name"`
-	Alias      string   `json:"alias,omitempty"`
-	Command    []string `json:"command,omitempty"`
-	Entrypoint []string `json:"entrypoint,omitempty"`
-	Ports      []Port   `json:"ports,omitempty"`
+	Name         string             `json:"name"`
+	Alias        string             `json:"alias,omitempty"`
+	Command      []string           `json:"command,omitempty"`
+	Entrypoint   []string           `json:"entrypoint,omitempty"`
+	Ports        []Port             `json:"ports,omitempty"`
+	Variables    JobVariables       `json:"variables,omitempty"`
+	PullPolicies []DockerPullPolicy `json:"pull_policy,omitempty"`
 }
 
+func (i *Image) Aliases() []string { return strings.Fields(strings.ReplaceAll(i.Alias, ",", " ")) }
+
 type Port struct {
 	Number   int    `json:"number,omitempty"`
 	Protocol string `json:"protocol,omitempty"`
@@ -271,14 +306,16 @@ type Artifact struct {
 type Artifacts []Artifact
 
 type Cache struct {
-	Key       string        `json:"key"`
-	Untracked bool          `json:"untracked"`
-	Policy    CachePolicy   `json:"policy"`
-	Paths     ArtifactPaths `json:"paths"`
-	When      CacheWhen     `json:"when"`
+	Key          string            `json:"key"`
+	Untracked    bool              `json:"untracked"`
+	Policy       CachePolicy       `json:"policy"`
+	Paths        ArtifactPaths     `json:"paths"`
+	When         CacheWhen         `json:"when"`
+	FallbackKeys CacheFallbackKeys `json:"fallback_keys"`
 }
 
 type CacheWhen string
+type CacheFallbackKeys []string
 
 const (
 	CacheWhenOnFailure CacheWhen = "on_failure"
@@ -328,7 +365,7 @@ type DependencyArtifactsFile struct {
 }
 
 type Dependency struct {
-	ID            int                     `json:"id"`
+	ID            int64                   `json:"id"`
 	Token         string                  `json:"token"`
 	Name          string                  `json:"name"`
 	ArtifactsFile DependencyArtifactsFile `json:"artifacts_file"`
@@ -337,12 +374,37 @@ type Dependency struct {
 type Dependencies []Dependency
 
 type GitlabFeatures struct {
-	TraceSections  bool               `json:"trace_sections"`
-	FailureReasons []JobFailureReason `json:"failure_reasons"`
+	TraceSections     bool               `json:"trace_sections"`
+	TokenMaskPrefixes []string           `json:"token_mask_prefixes"`
+	FailureReasons    []JobFailureReason `json:"failure_reasons"`
+}
+
+type Hooks []Hook
+
+type Hook struct {
+	Name   HookName   `json:"name"`
+	Script StepScript `json:"script"`
+}
+
+type HookName string
+
+const (
+	HookPreGetSourcesScript  HookName = "pre_get_sources_script"
+	HookPostGetSourcesScript HookName = "post_get_sources_script"
+)
+
+func (hooks Hooks) Get(name HookName) Hook {
+	for _, hook := range hooks {
+		if hook.Name == name {
+			return hook
+		}
+	}
+
+	return Hook{}
 }
 
 type JobResponse struct {
-	ID            int            `json:"id"`
+	ID            int64          `json:"id"`
 	Token         string         `json:"token"`
 	AllowGitFetch bool           `json:"allow_git_fetch"`
 	JobInfo       JobInfo        `json:"job_info"`
@@ -358,6 +420,7 @@ type JobResponse struct {
 	Dependencies  Dependencies   `json:"dependencies"`
 	Features      GitlabFeatures `json:"features"`
 	Secrets       Secrets        `json:"secrets,omitempty"`
+	Hooks         Hooks          `json:"hooks,omitempty"`
 
 	TLSCAChain  string `json:"-"`
 	TLSAuthCert string `json:"-"`
@@ -380,8 +443,9 @@ type VaultSecret struct {
 }
 
 type VaultServer struct {
-	URL  string    `json:"url"`
-	Auth VaultAuth `json:"auth"`
+	URL       string    `json:"url"`
+	Auth      VaultAuth `json:"auth"`
+	Namespace string    `json:"namespace"`
 }
 
 type VaultAuth struct {
@@ -459,6 +523,7 @@ func (s *VaultSecret) SecretField() string {
 
 func (s *VaultServer) expandVariables(vars JobVariables) {
 	s.URL = vars.ExpandValue(s.URL)
+	s.Namespace = vars.ExpandValue(s.Namespace)
 
 	s.Auth.expandVariables(vars)
 }
@@ -502,9 +567,8 @@ type JobTraceOutput struct {
 	Bytesize int    `json:"bytesize,omitempty"`
 }
 
-//nolint:lll
 type JobCredentials struct {
-	ID          int    `long:"id" env:"CI_JOB_ID" description:"The build ID to download and upload artifacts for"`
+	ID          int64  `long:"id" env:"CI_JOB_ID" description:"The build ID to download and upload artifacts for"`
 	Token       string `long:"token" env:"CI_JOB_TOKEN" required:"true" description:"Build token"`
 	URL         string `long:"url" env:"CI_SERVER_URL" required:"true" description:"GitLab CI URL"`
 	TLSCAFile   string `long:"tls-ca-file" env:"CI_SERVER_TLS_CA_FILE" description:"File containing the certificates to verify the peer when using HTTPS"`
@@ -533,7 +597,7 @@ func (j *JobCredentials) GetToken() string {
 }
 
 type UpdateJobInfo struct {
-	ID            int
+	ID            int64
 	State         JobState
 	FailureReason JobFailureReason
 	Output        JobTraceOutput
@@ -551,6 +615,7 @@ type FailuresCollector interface {
 	RecordFailure(reason JobFailureReason, runnerDescription string)
 }
 
+//go:generate mockery --name=JobTrace --inpackage
 type JobTrace interface {
 	io.Writer
 	Success()
@@ -560,7 +625,8 @@ type JobTrace interface {
 	SetAbortFunc(abortFunc context.CancelFunc)
 	Abort() bool
 	SetFailuresCollector(fc FailuresCollector)
-	SetMasked(values []string)
+	SetMasked(maskOptions MaskOptions)
+	SetDebugModeEnabled(isEnabled bool)
 	IsStdout() bool
 }
 
@@ -585,14 +651,19 @@ func NewPatchTraceResult(sentOffset int, state PatchState, newUpdateInterval int
 	}
 }
 
+//go:generate mockery --name=Network --inpackage
 type Network interface {
 	RegisterRunner(config RunnerCredentials, parameters RegisterRunnerParameters) *RegisterRunnerResponse
-	VerifyRunner(config RunnerCredentials) bool
+	VerifyRunner(config RunnerCredentials, systemID string) *VerifyRunnerResponse
 	UnregisterRunner(config RunnerCredentials) bool
+	UnregisterRunnerManager(config RunnerCredentials, systemID string) bool
+	ResetToken(runner RunnerCredentials, systemID string) *ResetTokenResponse
+	ResetTokenWithPAT(runner RunnerCredentials, systemID string, pat string) *ResetTokenResponse
 	RequestJob(ctx context.Context, config RunnerConfig, sessionInfo *SessionInfo) (*JobResponse, bool)
 	UpdateJob(config RunnerConfig, jobCredentials *JobCredentials, jobInfo UpdateJobInfo) UpdateJobResult
-	PatchTrace(config RunnerConfig, jobCredentials *JobCredentials, content []byte, startOffset int) PatchTraceResult
+	PatchTrace(config RunnerConfig, jobCredentials *JobCredentials, content []byte,
+		startOffset int, debugModeEnabled bool) PatchTraceResult
 	DownloadArtifacts(config JobCredentials, artifactsFile io.WriteCloser, directDownload *bool) DownloadState
-	UploadRawArtifacts(config JobCredentials, reader io.ReadCloser, options ArtifactsOptions) UploadState
+	UploadRawArtifacts(config JobCredentials, reader io.ReadCloser, options ArtifactsOptions) (UploadState, string)
 	ProcessJob(config RunnerConfig, buildCredentials *JobCredentials) (JobTrace, error)
 }
