@@ -6,7 +6,6 @@ import (
 	"encoding/json"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 
@@ -64,13 +63,18 @@ func (c *ConfigExecOutput) InjectInto(executor *executor) {
 	if c.JobEnv != nil {
 		executor.jobEnv = *c.JobEnv
 	}
+
+	if c.Shell != nil {
+		executor.Config.Shell = *c.Shell
+	}
 }
 
 type executor struct {
 	executors.AbstractExecutor
 
-	config  *config
-	tempDir string
+	config          *config
+	tempDir         string
+	jobResponseFile string
 
 	driverInfo *api.DriverInfo
 
@@ -85,7 +89,12 @@ func (e *executor) Prepare(options common.ExecutorPrepareOptions) error {
 		return err
 	}
 
-	e.tempDir, err = ioutil.TempDir("", "custom-executor")
+	e.tempDir, err = os.MkdirTemp("", "custom-executor")
+	if err != nil {
+		return err
+	}
+
+	e.jobResponseFile, err = e.createJobResponseFile()
 	if err != nil {
 		return err
 	}
@@ -135,6 +144,23 @@ func (e *executor) prepareConfig() error {
 	return nil
 }
 
+func (e *executor) createJobResponseFile() (string, error) {
+	responseFile := filepath.Join(e.tempDir, "response.json")
+	file, err := os.OpenFile(responseFile, os.O_RDWR|os.O_CREATE|os.O_TRUNC, 0o600)
+	if err != nil {
+		return "", fmt.Errorf("creating job response file %q: %w", responseFile, err)
+	}
+	defer func() { _ = file.Close() }()
+
+	encoder := json.NewEncoder(file)
+	err = encoder.Encode(e.Build.JobResponse)
+	if err != nil {
+		return "", fmt.Errorf("encoding job response file: %w", err)
+	}
+
+	return responseFile, nil
+}
+
 func (e *executor) dynamicConfig() error {
 	if e.config.ConfigExec == "" {
 		return nil
@@ -227,7 +253,11 @@ func (e *executor) prepareCommand(ctx context.Context, opts prepareCommandOpts)
 		cmdOpts.Env = append(cmdOpts.Env, fmt.Sprintf("CUSTOM_ENV_%s=%s", variable.Key, variable.Value))
 	}
 
-	return commandFactory(ctx, opts.executable, opts.args, cmdOpts)
+	options := command.Options{
+		JobResponseFile: e.jobResponseFile,
+	}
+
+	return commandFactory(ctx, opts.executable, opts.args, cmdOpts, options)
 }
 
 func (e *executor) getCIJobServicesEnv() common.JobVariable {
@@ -239,7 +269,7 @@ func (e *executor) getCIJobServicesEnv() common.JobVariable {
 	for _, service := range e.Build.Services {
 		services = append(services, jsonService{
 			Name:       service.Name,
-			Alias:      service.Alias,
+			Alias:      append(service.Aliases(), "")[0],
 			Entrypoint: service.Entrypoint,
 			Command:    service.Command,
 		})
@@ -257,22 +287,22 @@ func (e *executor) getCIJobServicesEnv() common.JobVariable {
 }
 
 func (e *executor) Run(cmd common.ExecutorCommand) error {
-	scriptDir, err := ioutil.TempDir(e.tempDir, "script")
+	scriptDir, err := os.MkdirTemp(e.tempDir, "script")
 	if err != nil {
 		return err
 	}
 
 	scriptFile := filepath.Join(scriptDir, "script."+e.BuildShell.Extension)
-	err = ioutil.WriteFile(scriptFile, []byte(cmd.Script), 0700)
+	err = os.WriteFile(scriptFile, []byte(cmd.Script), 0o700)
 	if err != nil {
 		return err
 	}
 
-	// TODO: Remove this translation in 14.0 - https://gitlab.com/gitlab-org/gitlab-runner/-/issues/26426
+	// TODO: Remove this translation - https://gitlab.com/groups/gitlab-org/-/epics/6112
 	stage := cmd.Stage
 	if stage == "step_script" {
-		e.BuildLogger.Warningln("Starting with version 14.0 the 'build_script' stage " +
-			"will be replaced with 'step_script': https://gitlab.com/gitlab-org/gitlab-runner/-/issues/26426")
+		e.BuildLogger.Warningln("Starting with version 17.0 the 'build_script' stage " +
+			"will be replaced with 'step_script': https://gitlab.com/groups/gitlab-org/-/epics/6112")
 		stage = "build_script"
 	}
 
