@@ -6,6 +6,7 @@ import (
 	"errors"
 	"fmt"
 	"io"
+	"net"
 	"os"
 	"path/filepath"
 	"runtime"
@@ -13,12 +14,12 @@ import (
 	"strconv"
 	"strings"
 	"sync"
-	"time"
 
 	"github.com/docker/cli/opts"
 	"github.com/docker/docker/api/types"
 	"github.com/docker/docker/api/types/container"
 	"github.com/docker/docker/api/types/network"
+	"github.com/docker/docker/client"
 	"github.com/docker/docker/pkg/stdcopy"
 	"github.com/sirupsen/logrus"
 
@@ -34,9 +35,7 @@ import (
 	"gitlab.com/gitlab-org/gitlab-runner/executors/docker/internal/wait"
 	"gitlab.com/gitlab-org/gitlab-runner/helpers"
 	"gitlab.com/gitlab-org/gitlab-runner/helpers/container/helperimage"
-	"gitlab.com/gitlab-org/gitlab-runner/helpers/container/services"
 	"gitlab.com/gitlab-org/gitlab-runner/helpers/docker"
-	"gitlab.com/gitlab-org/gitlab-runner/helpers/featureflags"
 	"gitlab.com/gitlab-org/gitlab-runner/helpers/limitwriter"
 	"gitlab.com/gitlab-org/gitlab-runner/shells"
 )
@@ -61,6 +60,12 @@ const (
 	labelWaitType    = "wait"
 )
 
+// internalFakeTunnelHostname is an internal hostname we provide the Docker client
+// when we provide a tunnelled dialer implementation. Because we're overriding
+// the dialer, this domain should never be used by the client, but we use the
+// reserved TLD ".invalid" for safety.
+const internalFakeTunnelHostname = "http://internel.tunnel.invalid"
+
 var neverRestartPolicy = container.RestartPolicy{Name: "no"}
 
 var (
@@ -78,7 +83,6 @@ type executor struct {
 
 	temporary []string // IDs of containers that should be removed
 
-	builds   []string // IDs of successfully created build containers
 	services []*types.Container
 
 	links []string
@@ -129,17 +133,24 @@ func init() {
 	}
 }
 
-func (e *executor) getServiceVariables() []string {
-	return e.Build.GetAllVariables().PublicOrInternal().StringList()
+func (e *executor) getServiceVariables(serviceDefinition common.Image) []string {
+	variables := e.Build.GetAllVariables().PublicOrInternal()
+	variables = append(variables, serviceDefinition.Variables...)
+
+	return variables.Expand().StringList()
 }
 
-func (e *executor) expandAndGetDockerImage(imageName string, allowedImages []string) (*types.ImageInspect, error) {
+func (e *executor) expandAndGetDockerImage(
+	imageName string,
+	allowedImages []string,
+	imagePullPolicies []common.DockerPullPolicy,
+) (*types.ImageInspect, error) {
 	imageName, err := e.expandImageName(imageName, allowedImages)
 	if err != nil {
 		return nil, err
 	}
 
-	image, err := e.pullManager.GetDockerImage(imageName)
+	image, err := e.pullManager.GetDockerImage(imageName, imagePullPolicies)
 	if err != nil {
 		return nil, err
 	}
@@ -185,16 +196,16 @@ func (e *executor) loadPrebuiltImage(path, ref, tag string) (*types.ImageInspect
 }
 
 func (e *executor) getPrebuiltImage() (*types.ImageInspect, error) {
-	if imageNameFromConfig := e.Config.Docker.HelperImage; imageNameFromConfig != "" {
-		imageNameFromConfig = common.AppVersion.Variables().ExpandValue(imageNameFromConfig)
-
+	if imageNameFromConfig := e.ExpandValue(e.Config.Docker.HelperImage); imageNameFromConfig != "" {
 		e.Debugln(
 			"Pull configured helper_image for predefined container instead of import bundled image",
 			imageNameFromConfig,
 			"...",
 		)
 
-		return e.pullManager.GetDockerImage(imageNameFromConfig)
+		e.Println("Using helper image: ", imageNameFromConfig, " (overridden, default would be ", e.helperImageInfo, ")")
+
+		return e.pullManager.GetDockerImage(imageNameFromConfig, nil)
 	}
 
 	e.Debugln(fmt.Sprintf("Looking for prebuilt image %s...", e.helperImageInfo))
@@ -209,13 +220,9 @@ func (e *executor) getPrebuiltImage() (*types.ImageInspect, error) {
 		return loadedImage, nil
 	}
 
-	if !e.Build.IsFeatureFlagOn(featureflags.GitLabRegistryHelperImage) {
-		e.Warningln(helperimage.DockerHubWarningMessage)
-	}
+	e.Println("Using helper image: ", e.helperImageInfo.String())
 
-	// Fall back to getting image from registry
-	e.Debugln(fmt.Sprintf("Loading image form registry: %s", e.helperImageInfo))
-	return e.pullManager.GetDockerImage(e.helperImageInfo.String())
+	return e.pullManager.GetDockerImage(e.helperImageInfo.String(), nil)
 }
 
 func (e *executor) getLocalHelperImage() *types.ImageInspect {
@@ -225,7 +232,7 @@ func (e *executor) getLocalHelperImage() *types.ImageInspect {
 
 	var flavor string
 	if e.Config.Docker != nil {
-		flavor = e.Config.Docker.HelperImageFlavor
+		flavor = e.ExpandValue(e.Config.Docker.HelperImageFlavor)
 	}
 
 	architecture := e.helperImageInfo.Architecture
@@ -266,8 +273,10 @@ func (e *executor) getBuildImage() (*types.ImageInspect, error) {
 		return nil, err
 	}
 
+	imagePullPolicies := e.Build.Image.PullPolicies
+
 	// Fetch image
-	image, err := e.pullManager.GetDockerImage(imageName)
+	image, err := e.pullManager.GetDockerImage(imageName, imagePullPolicies)
 	if err != nil {
 		return nil, err
 	}
@@ -354,7 +363,7 @@ func (e *executor) createService(
 	}
 
 	e.Println("Starting service", service+":"+version, "...")
-	serviceImage, err := e.pullManager.GetDockerImage(image)
+	serviceImage, err := e.pullManager.GetDockerImage(image, serviceDefinition.PullPolicies)
 	if err != nil {
 		return nil, err
 	}
@@ -374,7 +383,7 @@ func (e *executor) createService(
 	config := &container.Config{
 		Image:  serviceImage.ID,
 		Labels: e.labeler.Labels(labels),
-		Env:    append(e.getServiceVariables(), e.BuildShell.Environment...),
+		Env:    e.getServiceVariables(serviceDefinition),
 	}
 
 	if len(serviceDefinition.Command) > 0 {
@@ -402,12 +411,18 @@ func (e *executor) createService(
 }
 
 func (e *executor) createHostConfigForService() *container.HostConfig {
+	privileged := e.Config.Docker.Privileged
+	if e.Config.Docker.ServicesPrivileged != nil {
+		privileged = *e.Config.Docker.ServicesPrivileged
+	}
+
 	return &container.HostConfig{
 		DNS:           e.Config.Docker.DNS,
 		DNSSearch:     e.Config.Docker.DNSSearch,
 		RestartPolicy: neverRestartPolicy,
 		ExtraHosts:    e.Config.Docker.ExtraHosts,
-		Privileged:    e.Config.Docker.Privileged,
+		Privileged:    privileged,
+		SecurityOpt:   e.Config.Docker.ServicesSecurityOpt,
 		Runtime:       e.Config.Docker.Runtime,
 		UsernsMode:    container.UsernsMode(e.Config.Docker.UsernsMode),
 		NetworkMode:   e.networkMode,
@@ -441,112 +456,12 @@ func (e *executor) getProjectUniqRandomizedName() string {
 	return e.projectUniqRandomizedName
 }
 
-func (e *executor) getServicesDefinitions() (common.Services, error) {
-	var internalServiceImages []string
-	serviceDefinitions := common.Services{}
-
-	for _, service := range e.Config.Docker.Services {
-		internalServiceImages = append(internalServiceImages, service.Name)
-		serviceDefinitions = append(serviceDefinitions, service.ToImageDefinition())
-	}
-
-	for _, service := range e.Build.Services {
-		serviceName := e.Build.GetAllVariables().ExpandValue(service.Name)
-		err := e.verifyAllowedImage(serviceName, "services", e.Config.Docker.AllowedServices, internalServiceImages)
-		if err != nil {
-			return nil, err
-		}
-
-		service.Name = serviceName
-		serviceDefinitions = append(serviceDefinitions, service)
-	}
-
-	return serviceDefinitions, nil
-}
-
-func (e *executor) waitForServices() {
-	waitForServicesTimeout := e.Config.Docker.WaitForServicesTimeout
-	if waitForServicesTimeout == 0 {
-		waitForServicesTimeout = common.DefaultWaitForServicesTimeout
-	}
-
-	// wait for all services to came up
-	if waitForServicesTimeout > 0 && len(e.services) > 0 {
-		e.Println("Waiting for services to be up and running...")
-		wg := sync.WaitGroup{}
-		for _, service := range e.services {
-			wg.Add(1)
-			go func(service *types.Container) {
-				_ = e.waitForServiceContainer(service, time.Duration(waitForServicesTimeout)*time.Second)
-				wg.Done()
-			}(service)
-		}
-		wg.Wait()
-	}
-}
-
-func (e *executor) buildServiceLinks(linksMap map[string]*types.Container) (links []string) {
-	for linkName, linkee := range linksMap {
-		newContainer, err := e.client.ContainerInspect(e.Context, linkee.ID)
-		if err != nil {
-			continue
-		}
-		if newContainer.State.Running {
-			links = append(links, linkee.ID+":"+linkName)
-		}
-	}
-	return
-}
-
-func (e *executor) createFromServiceDefinition(
-	serviceIndex int,
-	serviceDefinition common.Image,
-	linksMap map[string]*types.Container,
-) error {
-	var container *types.Container
-
-	serviceMeta := services.SplitNameAndVersion(serviceDefinition.Name)
-
-	if serviceDefinition.Alias != "" {
-		serviceMeta.Aliases = append(serviceMeta.Aliases, serviceDefinition.Alias)
-	}
-
-	for _, linkName := range serviceMeta.Aliases {
-		if linksMap[linkName] != nil {
-			e.Warningln("Service", serviceDefinition.Name, "is already created. Ignoring.")
-			continue
-		}
-
-		// Create service if not yet created
-		if container == nil {
-			var err error
-			container, err = e.createService(
-				serviceIndex,
-				serviceMeta.Service,
-				serviceMeta.Version,
-				serviceMeta.ImageName,
-				serviceDefinition,
-				serviceMeta.Aliases,
-			)
-			if err != nil {
-				return err
-			}
-
-			e.Debugln("Created service", serviceDefinition.Name, "as", container.ID)
-			e.services = append(e.services, container)
-			e.temporary = append(e.temporary, container.ID)
-		}
-		linksMap[linkName] = container
-	}
-	return nil
-}
-
 func (e *executor) createBuildNetwork() error {
 	if e.networksManager == nil {
 		return errNetworksManagerUndefined
 	}
 
-	networkMode, err := e.networksManager.Create(e.Context, e.Config.Docker.NetworkMode)
+	networkMode, err := e.networksManager.Create(e.Context, e.Config.Docker.NetworkMode, e.Config.Docker.EnableIPv6)
 	if err != nil {
 		return err
 	}
@@ -582,34 +497,6 @@ func (e *executor) cleanupNetwork(ctx context.Context) error {
 	return e.networksManager.Cleanup(ctx)
 }
 
-func (e *executor) createServices() (err error) {
-	e.SetCurrentStage(ExecutorStageCreatingServices)
-	e.Debugln("Creating services...")
-
-	servicesDefinitions, err := e.getServicesDefinitions()
-	if err != nil {
-		return
-	}
-
-	linksMap := make(map[string]*types.Container)
-
-	for index, serviceDefinition := range servicesDefinitions {
-		err = e.createFromServiceDefinition(index, serviceDefinition, linksMap)
-		if err != nil {
-			return
-		}
-	}
-
-	e.waitForServices()
-
-	if e.networkMode.IsBridge() || e.networkMode.NetworkName() == "" {
-		e.Debugln("Building service links...")
-		e.links = e.buildServiceLinks(linksMap)
-	}
-
-	return
-}
-
 func (e *executor) createContainer(
 	containerType string,
 	imageDefinition common.Image,
@@ -620,7 +507,7 @@ func (e *executor) createContainer(
 		return nil, errVolumesManagerUndefined
 	}
 
-	image, err := e.expandAndGetDockerImage(imageDefinition.Name, allowedInternalImages)
+	image, err := e.expandAndGetDockerImage(imageDefinition.Name, allowedInternalImages, imageDefinition.PullPolicies)
 	if err != nil {
 		return nil, err
 	}
@@ -630,9 +517,15 @@ func (e *executor) createContainer(
 		hostname = e.Build.ProjectUniqueName()
 	}
 
-	// Always create unique, but sequential name
-	containerIndex := len(e.builds)
-	containerName := e.getProjectUniqRandomizedName() + "-" + containerType + "-" + strconv.Itoa(containerIndex)
+	// Build and predefined container names are comprised of:
+	// - A runner project scoped ID (runner-<description>-project-<project_id>-concurrent-<concurrent>)
+	// - A unique randomized ID for each execution
+	// - The container's type (build, predefined)
+	//
+	// For example: runner-linux-project-123-concurrent-2-0a1b2c3d-predefined
+	//
+	// A container of the same type is created _once_ per execution and re-used.
+	containerName := e.getProjectUniqRandomizedName() + "-" + containerType
 
 	config := e.createContainerConfig(containerType, imageDefinition, image.ID, hostname, cmd)
 
@@ -649,22 +542,15 @@ func (e *executor) createContainer(
 
 	e.Debugln("Creating container", containerName, "...")
 	resp, err := e.client.ContainerCreate(e.Context, config, hostConfig, networkConfig, containerName)
+	if resp.ID != "" {
+		e.temporary = append(e.temporary, resp.ID)
+	}
 	if err != nil {
-		if resp.ID != "" {
-			e.temporary = append(e.temporary, resp.ID)
-		}
 		return nil, err
 	}
 
 	inspect, err := e.client.ContainerInspect(e.Context, resp.ID)
-	if err != nil {
-		e.temporary = append(e.temporary, resp.ID)
-		return nil, err
-	}
-
-	e.builds = append(e.builds, resp.ID)
-	e.temporary = append(e.temporary, resp.ID)
-	return &inspect, nil
+	return &inspect, err
 }
 
 func (e *executor) createContainerConfig(
@@ -674,20 +560,28 @@ func (e *executor) createContainerConfig(
 	hostname string,
 	cmd []string,
 ) *container.Config {
+	labels := e.prepareContainerLabels(map[string]string{"type": containerType})
 	config := &container.Config{
 		Image:        imageID,
 		Hostname:     hostname,
 		Cmd:          cmd,
-		Labels:       e.labeler.Labels(map[string]string{"type": containerType}),
+		Labels:       labels,
 		Tty:          false,
 		AttachStdin:  true,
 		AttachStdout: true,
 		AttachStderr: true,
 		OpenStdin:    true,
 		StdinOnce:    true,
-		Env:          append(e.Build.GetAllVariables().StringList(), e.BuildShell.Environment...),
+		Env:          e.Build.GetAllVariables().StringList(),
 	}
+
+	// user config should only be set in build containers
+	if containerType == buildContainerType {
+		config.User = e.Config.Docker.User
+	}
+
 	config.Entrypoint = e.overwriteEntrypoint(&imageDefinition)
+	config.MacAddress = e.Config.Docker.MacAddress
 
 	return config
 }
@@ -698,6 +592,12 @@ func (e *executor) createHostConfig() (*container.HostConfig, error) {
 		return nil, err
 	}
 
+	isolation := container.Isolation(e.Config.Docker.Isolation)
+	if !isolation.IsValid() {
+		return nil, fmt.Errorf("the isolation value %q is not valid. "+
+			"the valid values are: 'process', 'hyperv', 'default' and an empty string", isolation)
+	}
+
 	return &container.HostConfig{
 		Resources: container.Resources{
 			Memory:            e.Config.Docker.GetMemory(),
@@ -709,6 +609,7 @@ func (e *executor) createHostConfig() (*container.HostConfig, error) {
 			Devices:           e.devices,
 			DeviceRequests:    e.deviceRequests,
 			OomKillDisable:    e.Config.Docker.GetOomKillDisable(),
+			DeviceCgroupRules: e.Config.Docker.DeviceCgroupRules,
 		},
 		DNS:           e.Config.Docker.DNS,
 		DNSSearch:     e.Config.Docker.DNSSearch,
@@ -721,10 +622,12 @@ func (e *executor) createHostConfig() (*container.HostConfig, error) {
 		RestartPolicy: neverRestartPolicy,
 		ExtraHosts:    e.Config.Docker.ExtraHosts,
 		NetworkMode:   e.networkMode,
+		IpcMode:       container.IpcMode(e.Config.Docker.IpcMode),
 		Links:         append(e.Config.Docker.Links, e.links...),
 		Binds:         e.volumesManager.Binds(),
 		OomScoreAdj:   e.Config.Docker.OomScoreAdjust,
 		ShmSize:       e.Config.Docker.ShmSize,
+		Isolation:     isolation,
 		VolumeDriver:  e.Config.Docker.VolumeDriver,
 		VolumesFrom:   e.Config.Docker.VolumesFrom,
 		LogConfig: container.LogConfig{
@@ -738,7 +641,13 @@ func (e *executor) createHostConfig() (*container.HostConfig, error) {
 func (e *executor) startAndWatchContainer(ctx context.Context, id string, input io.Reader) error {
 	dockerExec := exec.NewDocker(e.Context, e.client, e.waiter, e.Build.Log())
 
-	return dockerExec.Exec(ctx, id, input, e.Trace)
+	streams := exec.IOStreams{
+		Stdin:  input,
+		Stderr: e.Trace,
+		Stdout: e.Trace,
+	}
+
+	return dockerExec.Exec(ctx, id, streams)
 }
 
 func (e *executor) removeContainer(ctx context.Context, id string) error {
@@ -808,9 +717,10 @@ func (e *executor) verifyAllowedImage(image, optionName string, allowedImages, i
 }
 
 func (e *executor) expandImageName(imageName string, allowedInternalImages []string) (string, error) {
+	defaultDockerImage := e.ExpandValue(e.Config.Docker.Image)
 	if imageName != "" {
-		image := e.Build.GetAllVariables().ExpandValue(imageName)
-		allowedInternalImages = append(allowedInternalImages, e.Config.Docker.Image)
+		image := e.ExpandValue(imageName)
+		allowedInternalImages = append(allowedInternalImages, defaultDockerImage)
 		err := e.verifyAllowedImage(image, "images", e.Config.Docker.AllowedImages, allowedInternalImages)
 		if err != nil {
 			return "", err
@@ -818,11 +728,11 @@ func (e *executor) expandImageName(imageName string, allowedInternalImages []str
 		return image, nil
 	}
 
-	if e.Config.Docker.Image == "" {
+	if defaultDockerImage == "" {
 		return "", errors.New("no Docker image specified to run the build in")
 	}
 
-	return e.Config.Docker.Image, nil
+	return defaultDockerImage, nil
 }
 
 func (e *executor) overwriteEntrypoint(image *common.Image) []string {
@@ -837,17 +747,48 @@ func (e *executor) overwriteEntrypoint(image *common.Image) []string {
 	return nil
 }
 
-func (e *executor) connectDocker() error {
-	client, err := docker.New(e.Config.Docker.Credentials, "")
-	if err != nil {
-		return err
-	}
-	e.client = client
+func (e *executor) connectDocker(options common.ExecutorPrepareOptions) error {
+	var opts []client.Opt
 
-	e.info, err = client.Info(e.Context)
+	environment, ok := e.Build.ExecutorData.(executors.Environment)
+	if ok {
+		// We tunnel the docker connection for remote environments.
+		//
+		// To do this, we create a new dial context for Docker's client, whilst
+		// also overridding the daemon hostname it would typically use (if it were to use
+		// its own dialer).
+		//
+		// We do this because tunneling "npipe" and "unix" connections is possible, but if
+		// we don't give docker a "fake" host, it'll complain when we tunnel to an npipe
+		// connection from Linux, or a unix connection from Windows.
+		originalHost := e.Config.Docker.Credentials.Host
+		e.Config.Docker.Credentials.Host = internalFakeTunnelHostname
+
+		opts = append(
+			opts,
+			client.WithDialContext(e.environmentDialContext(options, environment, originalHost)),
+		)
+	}
+
+	dockerClient, err := docker.New(e.Config.Docker.Credentials, opts...)
 	if err != nil {
 		return err
 	}
+	e.client = dockerClient
+
+	e.info, err = e.client.Info(e.Context)
+	if err != nil {
+		return err
+	}
+
+	e.Debugln(fmt.Sprintf(
+		"Connected to docker daemon (api version: %s, server version: %s, kernel: %s, os: %s/%s)",
+		e.client.ClientVersion(),
+		e.info.ServerVersion,
+		e.info.KernelVersion,
+		e.info.OSType,
+		e.info.Architecture,
+	))
 
 	err = e.validateOSType()
 	if err != nil {
@@ -859,22 +800,46 @@ func (e *executor) connectDocker() error {
 	return err
 }
 
+func (e *executor) environmentDialContext(
+	options common.ExecutorPrepareOptions,
+	environment executors.Environment,
+	host string,
+) func(ctx context.Context, network, addr string) (net.Conn, error) {
+	return func(ctx context.Context, network, addr string) (net.Conn, error) {
+		if host == "" {
+			host = os.Getenv("DOCKER_HOST")
+		}
+		if host == "" {
+			host = client.DefaultDockerHost
+		}
+		u, err := client.ParseHostURL(host)
+		if err != nil {
+			return nil, fmt.Errorf("parsing docker host: %w", err)
+		}
+
+		c, err := environment.Prepare(e.Context, e.BuildLogger, options)
+		if err != nil {
+			return nil, fmt.Errorf("preparing environment: %w", err)
+		}
+
+		conn, err := c.Dial(u.Scheme, u.Host)
+		if err != nil {
+			return nil, fmt.Errorf("dialing environment connection: %w", err)
+		}
+
+		return docker.NewWrappedConnCloser(conn, c), nil
+	}
+}
+
 // validateOSType checks if the ExecutorOptions metadata matches with the docker
 // info response.
 func (e *executor) validateOSType() error {
-	executorOSType := e.ExecutorOptions.Metadata[metadataOSType]
-	if executorOSType == "" {
-		return common.MakeBuildError("%s does not have any OSType specified", e.Config.Executor)
+	switch e.info.OSType {
+	case osTypeLinux, osTypeWindows:
+		return nil
 	}
 
-	if executorOSType != e.info.OSType {
-		return common.MakeBuildError(
-			"executor requires OSType=%s, but Docker Engine supports only OSType=%s",
-			executorOSType, e.info.OSType,
-		)
-	}
-
-	return nil
+	return fmt.Errorf("unsupported os type: %s", e.info.OSType)
 }
 
 func (e *executor) createDependencies() error {
@@ -971,7 +936,7 @@ func (e *executor) Prepare(options common.ExecutorPrepareOptions) error {
 
 	e.AbstractExecutor.PrepareConfiguration(options)
 
-	err := e.connectDocker()
+	err := e.connectDocker(options)
 	if err != nil {
 		return err
 	}
@@ -981,6 +946,9 @@ func (e *executor) Prepare(options common.ExecutorPrepareOptions) error {
 		return err
 	}
 
+	// setup default executor options based on OS type
+	e.setupDefaultExecutorOptions(e.helperImageInfo.OSType)
+
 	err = e.prepareBuildsDir(options)
 	if err != nil {
 		return err
@@ -1009,14 +977,56 @@ func (e *executor) Prepare(options common.ExecutorPrepareOptions) error {
 	return nil
 }
 
+func (e *executor) setupDefaultExecutorOptions(os string) {
+	switch os {
+	case helperimage.OSTypeWindows:
+		e.DefaultBuildsDir = `C:\builds`
+		e.DefaultCacheDir = `C:\cache`
+
+		e.ExecutorOptions.Shell.Shell = shells.SNPowershell
+		e.ExecutorOptions.Shell.RunnerCommand = "gitlab-runner-helper"
+
+		if e.volumeParser == nil {
+			e.volumeParser = parser.NewWindowsParser()
+		}
+
+		if e.newVolumePermissionSetter == nil {
+			e.newVolumePermissionSetter = func() (permission.Setter, error) {
+				return permission.NewDockerWindowsSetter(), nil
+			}
+		}
+
+	default:
+		e.DefaultBuildsDir = `/builds`
+		e.DefaultCacheDir = `/cache`
+
+		e.ExecutorOptions.Shell.Shell = "bash"
+		e.ExecutorOptions.Shell.RunnerCommand = "/usr/bin/gitlab-runner-helper"
+
+		if e.volumeParser == nil {
+			e.volumeParser = parser.NewLinuxParser()
+		}
+
+		if e.newVolumePermissionSetter == nil {
+			e.newVolumePermissionSetter = func() (permission.Setter, error) {
+				helperImage, err := e.getPrebuiltImage()
+				if err != nil {
+					return nil, err
+				}
+
+				return permission.NewDockerLinuxSetter(e.client, e.Build.Log(), helperImage), nil
+			}
+		}
+	}
+}
+
 func (e *executor) prepareHelperImage() (helperimage.Info, error) {
 	return helperimage.Get(common.REVISION, helperimage.Config{
-		OSType:          e.info.OSType,
-		Architecture:    e.info.Architecture,
-		OperatingSystem: e.info.OperatingSystem,
-		Shell:           e.Config.Shell,
-		GitLabRegistry:  e.Build.IsFeatureFlagOn(featureflags.GitLabRegistryHelperImage),
-		Flavor:          e.Config.Docker.HelperImageFlavor,
+		OSType:        e.info.OSType,
+		Architecture:  e.info.Architecture,
+		KernelVersion: e.info.KernelVersion,
+		Shell:         e.Config.Shell,
+		Flavor:        e.ExpandValue(e.Config.Docker.HelperImageFlavor),
 	})
 }
 
@@ -1111,84 +1121,6 @@ func (e *executor) cleanupVolume(ctx context.Context) error {
 	return nil
 }
 
-type serviceHealthCheckError struct {
-	Inner error
-	Logs  string
-}
-
-func (e *serviceHealthCheckError) Error() string {
-	if e.Inner == nil {
-		return "serviceHealthCheckError"
-	}
-
-	return e.Inner.Error()
-}
-
-func (e *executor) runServiceHealthCheckContainer(service *types.Container, timeout time.Duration) error {
-	waitImage, err := e.getPrebuiltImage()
-	if err != nil {
-		return fmt.Errorf("getPrebuiltImage: %w", err)
-	}
-
-	containerName := service.Names[0] + "-wait-for-service"
-
-	environment, err := e.addServiceHealthCheckEnvironment(service)
-	if err != nil {
-		return err
-	}
-
-	cmd := []string{"gitlab-runner-helper", "health-check"}
-
-	config := e.createConfigForServiceHealthCheckContainer(service, cmd, waitImage, environment)
-	hostConfig := e.createHostConfigForServiceHealthCheck(service)
-
-	e.Debugln(fmt.Sprintf("Creating service healthcheck container %s...", containerName))
-	resp, err := e.client.ContainerCreate(e.Context, config, hostConfig, nil, containerName)
-	if err != nil {
-		return fmt.Errorf("create service container: %w", err)
-	}
-	defer func() { _ = e.removeContainer(e.Context, resp.ID) }()
-
-	e.Debugln(fmt.Sprintf("Starting service healthcheck container %s (%s)...", containerName, resp.ID))
-	err = e.client.ContainerStart(e.Context, resp.ID, types.ContainerStartOptions{})
-	if err != nil {
-		return fmt.Errorf("start service container: %w", err)
-	}
-
-	ctx, cancel := context.WithTimeout(e.Context, timeout)
-	defer cancel()
-
-	err = e.waiter.Wait(ctx, resp.ID)
-	if err == nil {
-		return nil
-	}
-
-	if errors.Is(err, context.DeadlineExceeded) {
-		err = fmt.Errorf("service %q timeout", containerName)
-	} else {
-		err = fmt.Errorf("service %q health check: %w", containerName, err)
-	}
-
-	return &serviceHealthCheckError{
-		Inner: err,
-		Logs:  e.readContainerLogs(resp.ID),
-	}
-}
-
-func (e *executor) createConfigForServiceHealthCheckContainer(
-	service *types.Container,
-	cmd []string,
-	waitImage *types.ImageInspect,
-	environment []string,
-) *container.Config {
-	return &container.Config{
-		Cmd:    cmd,
-		Image:  waitImage.ID,
-		Labels: e.labeler.Labels(map[string]string{"type": labelWaitType, "wait": service.ID}),
-		Env:    environment,
-	}
-}
-
 func (e *executor) createHostConfigForServiceHealthCheck(service *types.Container) *container.HostConfig {
 	return &container.HostConfig{
 		RestartPolicy: neverRestartPolicy,
@@ -1226,66 +1158,56 @@ func (e *executor) addServiceHealthCheckEnvironment(service *types.Container) ([
 			return nil, fmt.Errorf("service %q has no exposed ports", service.Names[0])
 		}
 
-		environment = append(environment, fmt.Sprintf("WAIT_FOR_SERVICE_TCP_PORT=%d", ports[0]))
+		for _, port := range ports {
+			environment = append(environment, fmt.Sprintf("WAIT_FOR_SERVICE_%d_TCP_PORT=%d", port, port))
+		}
 	}
 
 	return environment, nil
 }
 
+//nolint:gocognit
 func (e *executor) getContainerExposedPorts(container *types.Container) ([]int, error) {
-	var ports []int
-
 	inspect, err := e.client.ContainerInspect(e.Context, container.ID)
 	if err != nil {
 		return nil, err
 	}
 
+	for _, env := range inspect.Config.Env {
+		key, val, ok := strings.Cut(env, "=")
+		if !ok {
+			continue
+		}
+
+		if strings.EqualFold(key, "HEALTHCHECK_TCP_PORT") {
+			port, err := strconv.ParseInt(val, 10, 32)
+			if err != nil {
+				return nil, fmt.Errorf("invalid health check tcp port: %v", val)
+			}
+
+			return []int{int(port)}, nil
+		}
+	}
+
+	// maxPortsCheck is the maximum number of ports that we'll check to see
+	// if a service is running
+	const maxPortsCheck = 20
+
+	var ports []int
 	for port := range inspect.Config.ExposedPorts {
-		start, _, err := port.Range()
+		start, end, err := port.Range()
 		if err == nil && port.Proto() == "tcp" {
-			ports = append(ports, start)
+			for i := start; i <= end && len(ports) < maxPortsCheck; i++ {
+				ports = append(ports, i)
+			}
 		}
 	}
 
 	sort.Ints(ports)
+
 	return ports, nil
 }
 
-func (e *executor) waitForServiceContainer(service *types.Container, timeout time.Duration) error {
-	err := e.runServiceHealthCheckContainer(service, timeout)
-	if err == nil {
-		return nil
-	}
-
-	var buffer bytes.Buffer
-	buffer.WriteString("\n")
-	buffer.WriteString(
-		helpers.ANSI_YELLOW + "*** WARNING:" + helpers.ANSI_RESET + " Service " + service.Names[0] +
-			" probably didn't start properly.\n")
-	buffer.WriteString("\n")
-	buffer.WriteString("Health check error:\n")
-	buffer.WriteString(strings.TrimSpace(err.Error()))
-	buffer.WriteString("\n")
-
-	if healtCheckErr, ok := err.(*serviceHealthCheckError); ok {
-		buffer.WriteString("\n")
-		buffer.WriteString("Health check container logs:\n")
-		buffer.WriteString(healtCheckErr.Logs)
-		buffer.WriteString("\n")
-	}
-
-	buffer.WriteString("\n")
-	buffer.WriteString("Service container logs:\n")
-	buffer.WriteString(e.readContainerLogs(service.ID))
-	buffer.WriteString("\n")
-
-	buffer.WriteString("\n")
-	buffer.WriteString(helpers.ANSI_YELLOW + "*********" + helpers.ANSI_RESET + "\n")
-	buffer.WriteString("\n")
-	_, _ = io.Copy(e.Trace, &buffer)
-	return err
-}
-
 func (e *executor) readContainerLogs(containerID string) string {
 	var buf bytes.Buffer
 
@@ -1308,3 +1230,13 @@ func (e *executor) readContainerLogs(containerID string) string {
 	_, _ = stdcopy.StdCopy(w, w, hijacked)
 	return strings.TrimSpace(buf.String())
 }
+
+func (e *executor) prepareContainerLabels(otherLabels map[string]string) map[string]string {
+	l := e.labeler.Labels(otherLabels)
+
+	for k, v := range e.Config.Docker.ContainerLabels {
+		l[k] = e.Build.Variables.ExpandValue(v)
+	}
+
+	return l
+}
