@@ -1,4 +1,4 @@
-// +build integration
+//go:build integration
 
 package docker_test
 
@@ -24,7 +24,6 @@ import (
 
 	"github.com/docker/docker/api/types"
 	"github.com/docker/docker/api/types/filters"
-	"github.com/hashicorp/go-version"
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/require"
 
@@ -41,13 +40,16 @@ import (
 
 var getWindowsImageOnce sync.Once
 var windowsImage string
+var systemIDState = common.NewSystemIDState()
 
 var windowsDockerImageTagMappings = map[string]string{
-	windows.V20H2: "20H2",
+	windows.V1809: "ltsc2019",
+	windows.V21H2: "ltsc2022",
 }
 
 func TestMain(m *testing.M) {
 	execDocker.PrebuiltImagesPaths = []string{"../../out/helper-images/"}
+	_ = systemIDState.EnsureSystemID()
 
 	os.Exit(m.Run())
 }
@@ -159,10 +161,23 @@ func getBuildForOS(t *testing.T, getJobResp func() (common.JobResponse, error))
 	jobResp, err := getJobResp()
 	require.NoError(t, err)
 
-	return common.Build{
+	build := common.Build{
 		JobResponse: jobResp,
 		Runner:      getRunnerConfigForOS(t),
 	}
+
+	// There's a problem with some of our Windows docker tests and git-lfs that makes them flaky.
+	//
+	// To avoid this, for now we're skipping git-lfs calls where possible. This can eventually be removed.
+	// see: https://gitlab.com/gitlab-org/gitlab-runner/-/issues/30865
+	if runtime.GOOS == "windows" {
+		build.Variables = append(build.Variables, common.JobVariable{
+			Key:   "GIT_LFS_SKIP_SMUDGE",
+			Value: "1",
+		})
+	}
+
+	return build
 }
 
 func getRunnerConfigForOS(t *testing.T) *common.RunnerConfig {
@@ -171,11 +186,12 @@ func getRunnerConfigForOS(t *testing.T) *common.RunnerConfig {
 	shell := "bash"
 
 	if runtime.GOOS == "windows" {
-		executor = "docker-windows"
 		shell = shells.SNPowershell
 		image = getWindowsImage(t)
 	}
 
+	require.NoError(t, systemIDState.EnsureSystemID())
+
 	return &common.RunnerConfig{
 		RunnerSettings: common.RunnerSettings{
 			Executor: executor,
@@ -184,19 +200,21 @@ func getRunnerConfigForOS(t *testing.T) *common.RunnerConfig {
 				Image:      image,
 				PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 			},
+			Cache: &common.CacheConfig{},
 		},
 		RunnerCredentials: common.RunnerCredentials{
 			Token: fmt.Sprintf("%x", md5.Sum([]byte(t.Name()))),
 		},
+		SystemIDState: systemIDState,
 	}
 }
 
-// windowsDockerImageTag checks the specified operatingSystem to see if it's one of the
-// supported Windows version. If true, it maps the os version to the corresponding mcr.microsoft.com Docker image tag.
+// windowsDockerImageTag checks the specified kernel version to see if it's one of the
+// supported Windows version. If true, it maps a compatible mcr.microsoft.com Docker image tag.
 // UnsupportedWindowsVersionError is returned when no supported Windows version
 // is found in the string.
-func windowsDockerImageTag(operatingSystem string) (string, error) {
-	version, err := windows.Version(operatingSystem)
+func windowsDockerImageTag(version string) (string, error) {
+	version, err := windows.Version(version)
 	if err != nil {
 		return "", err
 	}
@@ -211,15 +229,16 @@ func windowsDockerImageTag(operatingSystem string) (string, error) {
 
 func getWindowsImage(t *testing.T) string {
 	getWindowsImageOnce.Do(func() {
-		client, err := docker.New(docker.Credentials{}, "")
+		client, err := docker.New(docker.Credentials{})
 		require.NoError(t, err, "creating docker client")
 		defer client.Close()
 
 		info, err := client.Info(context.Background())
 		require.NoError(t, err, "docker info")
 
-		dockerImageTag, err := windowsDockerImageTag(info.OperatingSystem)
+		dockerImageTag, err := windowsDockerImageTag(info.KernelVersion)
 		require.NoError(t, err)
+
 		windowsImage = fmt.Sprintf(common.TestWindowsImage, dockerImageTag)
 	})
 
@@ -347,6 +366,35 @@ func TestDockerCommandNoRootImage(t *testing.T) {
 					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 				},
 			},
+			SystemIDState: systemIDState,
+		},
+	}
+
+	err = build.Run(&common.Config{}, &common.Trace{Writer: os.Stdout})
+	assert.NoError(t, err)
+}
+
+func TestDockerCommandEntrypointWithStderrOutput(t *testing.T) {
+	test.SkipIfGitLabCIOn(t, test.OSWindows)
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	resp, err := common.GetRemoteSuccessfulBuild()
+	assert.NoError(t, err)
+
+	resp.Image.Name = common.TestAlpineEntrypointStderrImage
+	build := &common.Build{
+		JobResponse: resp,
+		Runner: &common.RunnerConfig{
+			RunnerSettings: common.RunnerSettings{
+				Executor: "docker",
+				Docker: &common.DockerConfig{
+					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
+				},
+				FeatureFlags: map[string]bool{
+					featureflags.DisableUmaskForDockerExecutor: true,
+				},
+			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -374,6 +422,7 @@ func TestDockerCommandOwnershipOverflow(t *testing.T) {
 					featureflags.DisableUmaskForDockerExecutor: true,
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -418,6 +467,7 @@ func TestDockerCommandWithAllowedImagesRun(t *testing.T) {
 					PullPolicy:      common.StringOrArray{common.PullPolicyIfNotPresent},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -480,6 +530,7 @@ func TestDockerCommandDisableEntrypointOverwrite(t *testing.T) {
 							DisableEntrypointOverwrite: test.disabled,
 						},
 					},
+					SystemIDState: systemIDState,
 				},
 			}
 
@@ -501,22 +552,6 @@ func TestDockerCommandDisableEntrypointOverwrite(t *testing.T) {
 	}
 }
 
-func isDockerOlderThan17_07(t *testing.T) bool {
-	client, err := docker.New(docker.Credentials{}, "")
-	require.NoError(t, err, "should be able to connect to docker")
-
-	types, err := client.Info(context.Background())
-	require.NoError(t, err, "should be able to get docker info")
-
-	localVersion, err := version.NewVersion(types.ServerVersion)
-	require.NoError(t, err)
-
-	checkedVersion, err := version.NewVersion("17.07.0-ce")
-	require.NoError(t, err)
-
-	return localVersion.LessThan(checkedVersion)
-}
-
 func TestDockerCommandMissingImage(t *testing.T) {
 	helpers.SkipIntegrationTests(t, "docker", "info")
 
@@ -526,13 +561,7 @@ func TestDockerCommandMissingImage(t *testing.T) {
 	err := build.Run(&common.Config{}, &common.Trace{Writer: os.Stdout})
 	require.Error(t, err)
 	assert.ErrorIs(t, err, &common.BuildError{FailureReason: common.ScriptFailure})
-
-	contains := "repository does not exist"
-	if isDockerOlderThan17_07(t) {
-		contains = "not found"
-	}
-
-	assert.Contains(t, err.Error(), contains)
+	assert.Regexp(t, regexp.MustCompile("not found|repository does not exist|invalid repository name"), err.Error())
 }
 
 func TestDockerCommandMissingTag(t *testing.T) {
@@ -560,13 +589,7 @@ func TestDockerCommandMissingServiceImage(t *testing.T) {
 	err := build.Run(&common.Config{}, &common.Trace{Writer: os.Stdout})
 	require.Error(t, err)
 	assert.ErrorIs(t, err, &common.BuildError{FailureReason: common.ScriptFailure})
-
-	contains := "repository does not exist"
-	if isDockerOlderThan17_07(t) {
-		contains = "not found"
-	}
-
-	assert.Contains(t, err.Error(), contains)
+	assert.Regexp(t, regexp.MustCompile("not found|repository does not exist|invalid repository name"), err.Error())
 }
 
 // TestDockerCommandPullingImageNoHost tests if the DNS resolution failure for the registry host
@@ -596,6 +619,15 @@ func TestBuildMasking(t *testing.T) {
 	buildtest.RunBuildWithMasking(t, getRunnerConfigForOS(t), nil)
 }
 
+func TestBuildExpandedFileVariable(t *testing.T) {
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	shellstest.OnEachShell(t, func(t *testing.T, shell string) {
+		build := getBuildForOS(t, common.GetSuccessfulBuild)
+		buildtest.RunBuildWithExpandedFileVariable(t, build.Runner, nil)
+	})
+}
+
 func TestDockerCommandTwoServicesFromOneImage(t *testing.T) {
 	test.SkipIfGitLabCIOn(t, test.OSWindows)
 	helpers.SkipIntegrationTests(t, "docker", "info")
@@ -632,6 +664,7 @@ func TestDockerCommandTwoServicesFromOneImage(t *testing.T) {
 					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -667,6 +700,7 @@ func TestDockerCommandOutput(t *testing.T) {
 					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -676,7 +710,7 @@ func TestDockerCommandOutput(t *testing.T) {
 	assert.NoError(t, err)
 
 	re, err :=
-		regexp.Compile("(?m)^Initialized empty Git repository in /builds/gitlab-org/ci-cd/tests/gitlab-test/.git/")
+		regexp.Compile("(?m)^Initialized empty Git repository in /builds/gitlab-org/ci-cd/gitlab-runner-pipeline-tests/gitlab-test/.git/")
 	require.NoError(t, err)
 	assert.Regexp(t, re, buffer.String())
 }
@@ -711,6 +745,7 @@ func TestDockerPrivilegedServiceAccessingBuildsFolder(t *testing.T) {
 						Privileged: true,
 					},
 				},
+				SystemIDState: systemIDState,
 			},
 		}
 		build.Steps = common.Steps{
@@ -725,6 +760,12 @@ func TestDockerPrivilegedServiceAccessingBuildsFolder(t *testing.T) {
 		build.Services = common.Services{
 			common.Image{
 				Name: common.TestDockerDindImage,
+				// set bip manually to prevent DinD-ception networking problems
+				// and avoid collision with:
+				// - docker daemon on the host
+				// - dind as a service to the CI job running this test
+				// - dind as a service to this test
+				Command: []string{"--bip", "172.30.0.1/16"},
 			},
 		}
 		build.Variables = append(build.Variables, common.JobVariable{
@@ -755,6 +796,7 @@ func getTestDockerJob(t *testing.T) *common.Build {
 					Privileged: true,
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 	build.Steps = common.Steps{
@@ -834,7 +876,7 @@ func TestCacheInContainer(t *testing.T) {
 	successfulBuild, err := common.GetRemoteSuccessfulBuild()
 	assert.NoError(t, err)
 
-	successfulBuild.JobInfo.ProjectID = int(time.Now().Unix())
+	successfulBuild.JobInfo.ProjectID = time.Now().Unix()
 	successfulBuild.Steps[0].Script = common.StepScript{
 		"(test -d cached/ && ls -lh cached/) || echo \"no cached directory\"",
 		"(test -f cached/date && cat cached/date) || echo \"no cached date\"",
@@ -860,7 +902,9 @@ func TestCacheInContainer(t *testing.T) {
 					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 					Volumes:    []string{"/cache"},
 				},
+				Cache: &common.CacheConfig{},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -950,6 +994,7 @@ func TestDockerImageNameFromVariable(t *testing.T) {
 					AllowedServices: []string{common.TestAlpineImage},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -984,6 +1029,7 @@ func TestDockerServiceNameFromVariable(t *testing.T) {
 					AllowedServices: []string{common.TestAlpineImage},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -1002,6 +1048,8 @@ func TestDockerServiceHealthcheck(t *testing.T) {
 		serviceStarted  bool
 		networkPerBuild string
 		skip            bool
+		port            int
+		variables       common.JobVariables
 	}{
 		"successful service (FF_NETWORK_PER_BUILD=false)": {
 			command:         []string{"server"},
@@ -1015,6 +1063,22 @@ func TestDockerServiceHealthcheck(t *testing.T) {
 			networkPerBuild: "true",
 			skip:            false,
 		},
+		"successful service explicit port (FF_NETWORK_PER_BUILD=false)": {
+			command:         []string{"server", "--addr", ":8888"},
+			serviceStarted:  true,
+			networkPerBuild: "false",
+			skip:            runtime.GOOS == "windows",
+			port:            8888,
+			variables:       []common.JobVariable{{Key: "HEALTHCHECK_TCP_PORT", Value: "8888"}},
+		},
+		"successful service explicit port (FF_NETWORK_PER_BUILD=true)": {
+			command:         []string{"server", "--addr", ":8888"},
+			serviceStarted:  true,
+			networkPerBuild: "true",
+			skip:            false,
+			port:            8888,
+			variables:       []common.JobVariable{{Key: "HEALTHCHECK_TCP_PORT", Value: "8888"}},
+		},
 		"failed service (FF_NETWORK_PER_BUILD=false)": {
 			command:         []string{"server", "--addr", ":8888"},
 			serviceStarted:  false,
@@ -1035,10 +1099,14 @@ func TestDockerServiceHealthcheck(t *testing.T) {
 				t.Skipf("OS %q does not support 'link' networking", runtime.GOOS)
 			}
 
+			if tc.port == 0 {
+				tc.port = 80
+			}
+
 			resp, err := common.GetRemoteBuildResponse(
-				"liveness client db",
-				"liveness client registry.gitlab.com__gitlab-org__ci-cd__tests__liveness",
-				"liveness client registry.gitlab.com-gitlab-org-ci-cd-tests-liveness",
+				fmt.Sprintf("liveness client db:%d", tc.port),
+				fmt.Sprintf("liveness client registry.gitlab.com__gitlab-org__ci-cd__tests__liveness:%d", tc.port),
+				fmt.Sprintf("liveness client registry.gitlab.com-gitlab-org-ci-cd-tests-liveness:%d", tc.port),
 			)
 			require.NoError(t, err)
 
@@ -1047,8 +1115,11 @@ func TestDockerServiceHealthcheck(t *testing.T) {
 				Runner: &common.RunnerConfig{
 					RunnerSettings: common.RunnerSettings{
 						Executor: "docker",
-						Docker:   &common.DockerConfig{},
+						Docker: &common.DockerConfig{
+							WaitForServicesTimeout: 15,
+						},
 					},
+					SystemIDState: systemIDState,
 				},
 			}
 
@@ -1059,14 +1130,14 @@ func TestDockerServiceHealthcheck(t *testing.T) {
 
 			if runtime.GOOS == "windows" {
 				build.Runner.RunnerSettings.Shell = shells.SNPwsh
-				build.Runner.RunnerSettings.Executor = "docker-windows"
 				build.Image.Entrypoint = []string{""}
 			}
 
 			build.Services = append(build.Services, common.Image{
-				Name:    common.TestLivenessImage,
-				Alias:   "db",
-				Command: tc.command,
+				Name:      common.TestLivenessImage,
+				Alias:     "db",
+				Command:   tc.command,
+				Variables: tc.variables,
 			})
 
 			build.Variables = append(build.Variables, common.JobVariable{
@@ -1102,6 +1173,7 @@ func TestDockerServiceHealthcheckOverflow(t *testing.T) {
 				Executor: "docker",
 				Docker:   &common.DockerConfig{},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -1126,6 +1198,46 @@ func TestDockerServiceHealthcheckOverflow(t *testing.T) {
 	assert.NotContains(t, out, ":dataend")
 }
 
+func TestDockerHandlesAliasDuplicates(t *testing.T) {
+	test.SkipIfGitLabCIOn(t, test.OSWindows)
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	resp, err := common.GetRemoteBuildResponse("ping -c 1 alpine && ping -c 1 svc-1")
+	assert.NoError(t, err)
+
+	build := &common.Build{
+		JobResponse: resp,
+		Runner: &common.RunnerConfig{
+			RunnerSettings: common.RunnerSettings{
+				Executor: "docker",
+				Docker:   &common.DockerConfig{},
+			},
+			SystemIDState: systemIDState,
+		},
+	}
+
+	build.Image = common.Image{
+		Name: common.TestAlpineImage,
+	}
+
+	build.Services = append(build.Services, common.Image{
+		Name:    common.TestAlpineImage,
+		Command: []string{"sleep", "5"},
+		Alias:   "alpine alpine svc-1 svc-1",
+	})
+
+	build.Variables = append(build.Variables, common.JobVariable{
+		Key:    "FF_NETWORK_PER_BUILD",
+		Value:  "true",
+		Public: true,
+	})
+
+	out, err := buildtest.RunBuildReturningOutput(t, build)
+	assert.NoError(t, err)
+	assert.Contains(t, out, "PING alpine")
+	assert.Contains(t, out, "PING svc-1")
+}
+
 func runDockerInDocker(version string) (id string, err error) {
 	cmd := exec.Command("docker", "run", "--detach", "--privileged", "-p", "2375", "docker:"+version+"-dind")
 	cmd.Stderr = os.Stderr
@@ -1157,7 +1269,7 @@ func getDockerCredentials(id string) (credentials docker.Credentials, err error)
 }
 
 func waitForDocker(credentials docker.Credentials) error {
-	client, err := docker.New(credentials, "")
+	client, err := docker.New(credentials)
 	if err != nil {
 		return err
 	}
@@ -1214,6 +1326,7 @@ func testDockerVersion(t *testing.T, version string) {
 					CPUS:        "0.1",
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -1282,6 +1395,7 @@ func TestDockerCommandWithBrokenGitSSLCAInfo(t *testing.T) {
 					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -1313,6 +1427,7 @@ func TestDockerCommandWithGitSSLCAInfo(t *testing.T) {
 					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -1344,6 +1459,7 @@ func TestDockerCommandWithHelperImageConfig(t *testing.T) {
 					PullPolicy:  common.StringOrArray{common.PullPolicyIfNotPresent},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -1417,6 +1533,7 @@ func TestDockerCommandWithDoingPruneAndAfterScript(t *testing.T) {
 					},
 				},
 			},
+			SystemIDState: systemIDState,
 		},
 	}
 
@@ -1512,7 +1629,7 @@ func removeBuildContainer(t *testing.T) <-chan string {
 	removedContainer := make(chan string, 1)
 	defer close(removedContainer)
 
-	client, err := docker.New(docker.Credentials{}, "")
+	client, err := docker.New(docker.Credentials{})
 	require.NoError(t, err, "creating docker client")
 	defer client.Close()
 
@@ -1565,7 +1682,7 @@ func TestDockerCommand_WriteToVolumeNonRootImage(t *testing.T) {
 	const volumeBind = "/test"
 	const helperImage = "gitlab/gitlab-runner-helper:x86_64-4c96e5ad"
 
-	client, err := docker.New(docker.Credentials{}, "")
+	client, err := docker.New(docker.Credentials{})
 	require.NoError(t, err, "creating docker client")
 
 	build := getBuildForOS(t, common.GetRemoteSuccessfulBuild)
@@ -1696,6 +1813,7 @@ func TestChownAndUmaskUsage(t *testing.T) {
 							PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
 						},
 					},
+					SystemIDState: systemIDState,
 				},
 			}
 
@@ -1773,3 +1891,279 @@ func TestCleanupProjectGitSubmoduleRecursive(t *testing.T) {
 		untrackedSubSubmoduleFile,
 	)
 }
+
+func TestDockerCommandServiceVariables(t *testing.T) {
+	test.SkipIfGitLabCIOn(t, test.OSWindows)
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	build := getBuildForOS(t, common.GetRemoteSuccessfulBuild)
+	build.Variables = append(build.JobResponse.Variables,
+		common.JobVariable{
+			Key:    "FF_NETWORK_PER_BUILD",
+			Value:  "true",
+			Public: true,
+		},
+		common.JobVariable{
+			Key:    "BUILD_VAR",
+			Value:  "BUILD_VAR_VALUE",
+			Public: true,
+		},
+	)
+
+	shell := "sh"
+	if runtime.GOOS == "windows" {
+		shell = shells.SNPowershell
+	}
+
+	// immediately timeout as triggering an error is the  only way to get a
+	// service to send its output to the log
+	build.Runner.Docker.WaitForServicesTimeout = 1
+
+	build.Services = common.Services{
+		common.Image{
+			Name: common.TestLivenessImage,
+			Variables: []common.JobVariable{
+				{
+					Key:   "SERVICE_VAR",
+					Value: "SERVICE_VAR_VALUE",
+				},
+				{
+					Key:   "SERVICE_VAR_REF_BUILD_VAR",
+					Value: "$BUILD_VAR",
+				},
+			},
+			Entrypoint: append([]string{shell, "-c"}, "echo SERVICE_VAR=$SERVICE_VAR SERVICE_VAR_REF_BUILD_VAR=$SERVICE_VAR_REF_BUILD_VAR"),
+		},
+	}
+
+	var buffer bytes.Buffer
+	err := build.Run(&common.Config{}, &common.Trace{Writer: &buffer})
+	assert.NoError(t, err)
+	out := buffer.String()
+	assert.Contains(t, out, "SERVICE_VAR=SERVICE_VAR_VALUE")
+	assert.Contains(t, out, "SERVICE_VAR_REF_BUILD_VAR=BUILD_VAR_VALUE")
+}
+
+func TestDockerCommandConflictingPullPolicies(t *testing.T) {
+	test.SkipIfGitLabCIOn(t, test.OSWindows)
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	successfulBuild, err := common.GetRemoteSuccessfulBuild()
+	require.NoError(t, err)
+
+	successfulBuild.Image = common.Image{Name: common.TestAlpineImage}
+	build := &common.Build{
+		JobResponse: successfulBuild,
+		Runner: &common.RunnerConfig{
+			RunnerSettings: common.RunnerSettings{
+				Executor: "docker",
+				Docker: &common.DockerConfig{
+					Image: common.TestAlpineImage,
+				},
+			},
+			SystemIDState: systemIDState,
+		},
+	}
+
+	tests := map[string]struct {
+		imagePullPolicies   []common.DockerPullPolicy
+		pullPolicy          common.StringOrArray
+		allowedPullPolicies []common.DockerPullPolicy
+		wantErrMsg          string
+	}{
+		"allowed_pull_policies configured, default pull_policy": {
+			imagePullPolicies:   nil,
+			pullPolicy:          nil,
+			allowedPullPolicies: []common.DockerPullPolicy{common.PullPolicyIfNotPresent},
+			wantErrMsg:          fmt.Sprintf(common.IncompatiblePullPolicy, "[always]", "Runner config (default)", "[if-not-present]"),
+		},
+		"allowed_pull_policies and pull_policy configured": {
+			imagePullPolicies:   nil,
+			pullPolicy:          common.StringOrArray{common.PullPolicyNever},
+			allowedPullPolicies: []common.DockerPullPolicy{common.PullPolicyIfNotPresent},
+			wantErrMsg:          fmt.Sprintf(common.IncompatiblePullPolicy, "[never]", "Runner config", "[if-not-present]"),
+		},
+		"allowed_pull_policies and image pull_policy configured": {
+			imagePullPolicies:   []common.DockerPullPolicy{common.PullPolicyAlways},
+			pullPolicy:          nil,
+			allowedPullPolicies: []common.DockerPullPolicy{common.PullPolicyIfNotPresent},
+			wantErrMsg:          fmt.Sprintf(common.IncompatiblePullPolicy, "[always]", "GitLab pipeline config", "[if-not-present]"),
+		},
+		"all configured": {
+			imagePullPolicies:   []common.DockerPullPolicy{common.PullPolicyAlways},
+			pullPolicy:          common.StringOrArray{common.PullPolicyNever},
+			allowedPullPolicies: []common.DockerPullPolicy{common.PullPolicyIfNotPresent},
+			wantErrMsg:          fmt.Sprintf(common.IncompatiblePullPolicy, "[always]", "GitLab pipeline config", "[if-not-present]"),
+		},
+	}
+
+	for name, test := range tests {
+		t.Run(name, func(t *testing.T) {
+			build.JobResponse.Image.PullPolicies = test.imagePullPolicies
+			build.Runner.RunnerSettings.Docker.PullPolicy = test.pullPolicy
+			build.Runner.RunnerSettings.Docker.AllowedPullPolicies = test.allowedPullPolicies
+
+			gotErr := build.Run(&common.Config{}, &common.Trace{Writer: os.Stdout})
+
+			require.Error(t, gotErr)
+			assert.Contains(t, gotErr.Error(), test.wantErrMsg)
+			assert.Contains(t, gotErr.Error(), "failed to pull image '"+common.TestAlpineImage)
+		})
+	}
+}
+
+func Test_CaptureServiceLogs(t *testing.T) {
+	test.SkipIfGitLabCIOn(t, test.OSWindows)
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	tests := map[string]struct {
+		buildVars []common.JobVariable
+		assert    func(string, error)
+	}{
+		"enabled": {
+			buildVars: []common.JobVariable{
+				{
+					Key:    "CI_DEBUG_SERVICES",
+					Value:  "true",
+					Public: true,
+				}, {
+					Key:    "POSTGRES_PASSWORD",
+					Value:  "password",
+					Public: true,
+				},
+			},
+			assert: func(out string, err error) {
+				assert.NoError(t, err)
+				assert.NotContains(t, out, "WARNING: failed to parse value 'trace' for CI_DEBUG_SERVICES variable")
+				assert.Regexp(t, `\[service:(postgres-db|db-postgres)\] .* The files belonging to this database system will be owned by user "postgres"`, out)
+				assert.Regexp(t, `\[service:(postgres-db|db-postgres)\] .* database system is ready to accept connections`, out)
+				assert.Regexp(t, `\[service:(redis-cache|cache-redis)\] .* oO0OoO0OoO0Oo Redis is starting oO0OoO0OoO0O`, out)
+				assert.Regexp(t, `\[service:(redis-cache|cache-redis)\] .* Ready to accept connections`, out)
+			},
+		},
+		"not enabled": {
+			assert: func(out string, err error) {
+				assert.NoError(t, err)
+				assert.NotRegexp(t, `\[service:(postgres-db|db-postgres)\] .* Error: Database is uninitialized and superuser password is not specified`, out)
+				assert.NotRegexp(t, `\[service:(redis-cache|cache-redis)\] .* oO0OoO0OoO0Oo Redis is starting oO0OoO0OoO0O`, out)
+				assert.NotRegexp(t, `\[service:(redis-cache|cache-redis)\] .* Ready to accept connections`, out)
+			},
+		},
+		"bogus value": {
+			buildVars: []common.JobVariable{{
+				Key:    "CI_DEBUG_SERVICES",
+				Value:  "blammo",
+				Public: true,
+			}},
+			assert: func(out string, err error) {
+				assert.NoError(t, err)
+				assert.Contains(t, out, "WARNING: failed to parse value 'blammo' for CI_DEBUG_SERVICES variable")
+				assert.NotRegexp(t, `\[service:(postgres-db|db-postgres)\] .* Error: Database is uninitialized and superuser password is not specified`, out)
+				assert.NotRegexp(t, `\[service:(redis-cache|cache-redis)\] .* oO0OoO0OoO0Oo Redis is starting oO0OoO0OoO0O`, out)
+				assert.NotRegexp(t, `\[service:(redis-cache|cache-redis)\] .* Ready to accept connections`, out)
+			},
+		},
+	}
+
+	build := getBuildForOS(t, common.GetRemoteSuccessfulBuild)
+	build.Services = append(build.Services, common.Image{Name: "postgres:14.4", Alias: "db"})
+	build.Services = append(build.Services, common.Image{Name: "redis:7.0", Alias: "cache"})
+
+	for name, tt := range tests {
+		t.Run(name, func(t *testing.T) {
+			build.Variables = tt.buildVars
+			out, err := buildtest.RunBuildReturningOutput(t, &build)
+			tt.assert(out, err)
+		})
+	}
+}
+
+func Test_ContainerOptionsExpansion(t *testing.T) {
+	test.SkipIfGitLabCIOn(t, test.OSWindows)
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	successfulBuild, err := common.GetRemoteSuccessfulBuildWithDumpedVariables()
+	assert.NoError(t, err)
+
+	jobVars := common.JobVariables{
+		{Key: "CI_DEBUG_SERVICES", Value: "true", Public: true},
+		{Key: "POSTGRES_PASSWORD", Value: "password", Public: true},
+		{Key: "JOB_IMAGE", Value: "alpine:latest"},
+		{Key: "HELPER_IMAGE", Value: "gitlab/gitlab-runner-helper:x86_64-latest"},
+		{Key: "HELPER_IMAGE_FLAVOR", Value: "alpine"},
+		{Key: "SRVS_IMAGE", Value: "postgres:latest"},
+		{Key: "SRVS_IMAGE_ALIAS", Value: "db"},
+	}
+	successfulBuild.Variables = append(successfulBuild.Variables, jobVars...)
+
+	build := &common.Build{
+		JobResponse: successfulBuild,
+		Runner: &common.RunnerConfig{
+			RunnerSettings: common.RunnerSettings{
+				Executor: "docker",
+				Docker: &common.DockerConfig{
+					Image:             "$JOB_IMAGE",
+					HelperImage:       "$HELPER_IMAGE",
+					HelperImageFlavor: "$HELPER_IMAGE_FLAVOR",
+					Services: []common.Service{
+						{Name: "$SRVS_IMAGE", Alias: "$SRVS_IMAGE_ALIAS"},
+					},
+				},
+			},
+		},
+	}
+
+	out, err := buildtest.RunBuildReturningOutput(t, build)
+	assert.NoError(t, err)
+	assert.Contains(t, out, "Pulling docker image alpine:latest")
+	assert.Contains(t, out, "Pulling docker image postgres:latest")
+	assert.Contains(t, out, "Pulling docker image gitlab/gitlab-runner-helper:x86_64-latest")
+	assert.Regexp(t, `\[service:(postgres-db|db-postgres)\]`, out)
+}
+
+func TestDockerCommandWithRunnerServiceEnvironmentVariables(t *testing.T) {
+	test.SkipIfGitLabCIOn(t, test.OSWindows)
+	helpers.SkipIntegrationTests(t, "docker", "info")
+
+	successfulBuild, err := common.GetRemoteSuccessfulBuild()
+	assert.NoError(t, err)
+
+	// Adding a gitlab-ci.yml variable to test the expansion of the service env variables
+	successfulBuild.Variables = append(successfulBuild.Variables, common.JobVariable{
+		Key:    "MY_GLOBAL_VAR",
+		Value:  "my_global_var_value",
+		Public: true,
+	})
+
+	build := &common.Build{
+		JobResponse: successfulBuild,
+		Runner: &common.RunnerConfig{
+			RunnerSettings: common.RunnerSettings{
+				Executor: "docker",
+				Docker: &common.DockerConfig{
+					Image:      common.TestAlpineImage,
+					PullPolicy: common.StringOrArray{common.PullPolicyIfNotPresent},
+					Services: []common.Service{
+						{
+							Name: common.TestAlpineImage,
+							Environment: []string{
+								// expanded service env var
+								"EXPANDED=$MY_GLOBAL_VAR",
+								"FOO=value from [[runners.docker.services]]",
+							},
+							Entrypoint: []string{"/bin/sh", "-c"},
+							Command:    []string{"echo -e \"FOO = $FOO\nEXPANDED = $EXPANDED\""},
+						},
+					},
+				},
+			},
+		},
+	}
+
+	out := bytes.NewBuffer(nil)
+	err = build.Run(&common.Config{}, &common.Trace{Writer: out})
+	assert.NoError(t, err)
+	assert.Contains(t, out.String(), "FOO = value from [[runners.docker.services]]")
+	assert.Contains(t, out.String(), "EXPANDED = my_global_var_value")
+
+}
