@@ -1,4 +1,4 @@
-// +build integration
+//go:build integration
 
 package networks_test
 
@@ -26,7 +26,7 @@ func TestCreateNetworkLabels(t *testing.T) {
 	successfulJobResponse, err := common.GetRemoteSuccessfulBuild()
 	require.NoError(t, err)
 
-	client, err := docker.New(docker.Credentials{}, "")
+	client, err := docker.New(docker.Credentials{})
 	require.NoError(t, err, "should be able to connect to docker")
 	defer client.Close()
 
@@ -50,18 +50,18 @@ func TestCreateNetworkLabels(t *testing.T) {
 
 	ctx := context.Background()
 
-	networkMode, err := manager.Create(ctx, "")
+	networkMode, err := manager.Create(ctx, "", false)
 	assert.NoError(t, err)
 	assert.Equal(t, container.NetworkMode("runner-test-tok-project-0-concurrent-0-job-0-network"), networkMode)
 
 	network, err := manager.Inspect(ctx)
 	assert.NoError(t, err)
 	assert.Equal(t, map[string]string{
-		"com.gitlab.gitlab-runner.job.before_sha":  "ca50079dac5293292f83a4d454922ba8db44e7a3",
+		"com.gitlab.gitlab-runner.job.before_sha":  "1ea27a9695f80d7816d9e8ce025d9b2df83d0dd7",
 		"com.gitlab.gitlab-runner.job.id":          "0",
 		"com.gitlab.gitlab-runner.job.url":         "https://gitlab.example.com/namespace/project/-/jobs/0",
 		"com.gitlab.gitlab-runner.job.ref":         "main",
-		"com.gitlab.gitlab-runner.job.sha":         "91956efe32fb7bef54f378d90c9bd74c19025872",
+		"com.gitlab.gitlab-runner.job.sha":         "69b18e5ed3610cf646119c3e38f462c64ec462b7",
 		"com.gitlab.gitlab-runner.managed":         "true",
 		"com.gitlab.gitlab-runner.pipeline.id":     "1",
 		"com.gitlab.gitlab-runner.project.id":      "0",
