@@ -1,11 +1,7 @@
-// +build windows
-
 package parser
 
 import (
 	"regexp"
-
-	"gitlab.com/gitlab-org/gitlab-runner/helpers/path"
 )
 
 // The specification of regular expression used for parsing Windows volumes
@@ -74,7 +70,7 @@ type windowsParser struct {
 func NewWindowsParser() Parser {
 	return &windowsParser{
 		baseParser: baseParser{
-			path: path.NewWindowsPath(),
+			path: newWindowsPath(),
 		},
 	}
 }
@@ -87,5 +83,5 @@ func (p *windowsParser) ParseVolume(spec string) (*Volume, error) {
 		return nil, err
 	}
 
-	return newVolume(parts["source"], parts["destination"], parts["mode"], ""), nil
+	return newVolume(parts["source"], parts["destination"], parts["mode"], "", ""), nil
 }
