@@ -26,11 +26,12 @@ type machineProvider struct {
 	stuckRemoveLock sync.Mutex
 
 	// metrics
-	totalActions      *prometheus.CounterVec
-	currentStatesDesc *prometheus.Desc
-	creationHistogram prometheus.Histogram
-	stoppingHistogram prometheus.Histogram
-	removalHistogram  prometheus.Histogram
+	totalActions            *prometheus.CounterVec
+	currentStatesDesc       *prometheus.Desc
+	creationHistogram       prometheus.Histogram
+	stoppingHistogram       prometheus.Histogram
+	removalHistogram        prometheus.Histogram
+	failedCreationHistogram prometheus.Histogram
 }
 
 func (m *machineProvider) machineDetails(name string, acquire bool) *machineDetails {
@@ -85,12 +86,12 @@ func (m *machineProvider) runnerMachinesCoordinator(config *common.RunnerConfig)
 func (m *machineProvider) create(config *common.RunnerConfig, state machineState) (*machineDetails, chan error) {
 	name := newMachineName(config)
 	details := m.machineDetails(name, true)
-	m.lock.Lock()
+	details.Lock()
 	details.State = machineStateCreating
 	details.UsedCount = 0
 	details.RetryCount = 0
 	details.LastSeen = time.Now()
-	m.lock.Unlock()
+	details.Unlock()
 	errCh := make(chan error, 1)
 
 	// Create machine with the required configuration asynchronously
@@ -122,19 +123,22 @@ func (m *machineProvider) createWithGrowthCapacity(
 		logger.WithField("time", time.Since(started)).
 			WithError(err).
 			Errorln("Machine creation failed")
+		m.totalActions.WithLabelValues("creation-failed").Inc()
+		m.failedCreationHistogram.Observe(time.Since(started).Seconds())
 		_ = m.remove(details.Name, "Failed to create")
 	} else {
-		m.lock.Lock()
+		details.Lock()
 		details.State = state
 		details.Used = time.Now()
-		m.lock.Unlock()
+		retryCount := details.RetryCount
+		details.Unlock()
+
 		creationTime := time.Since(started)
-		m.lock.RLock()
 		logger.WithField("duration", creationTime).
 			WithField("now", time.Now()).
-			WithField("retries", details.RetryCount).
+			WithField("retries", retryCount).
 			Infoln("Machine created")
-		m.lock.RUnlock()
+
 		m.totalActions.WithLabelValues("created").Inc()
 		m.creationHistogram.Observe(creationTime.Seconds())
 
@@ -231,8 +235,8 @@ func (m *machineProvider) tryGetFreeExistingMachineFromCoordinator(
 }
 
 func (m *machineProvider) tryAcquireMachineDetails(details *machineDetails) *machineDetails {
-	m.lock.Lock()
-	defer m.lock.Unlock()
+	details.Lock()
+	defer details.Unlock()
 
 	if details.isUsed() {
 		return nil
@@ -255,34 +259,40 @@ func (m *machineProvider) retryUseMachine(config *common.RunnerConfig) (details
 }
 
 func (m *machineProvider) removeMachine(details *machineDetails) (err error) {
+	details.Lock()
+	logger := details.logger()
+	info := details.info()
+	details.Unlock()
+
 	if !m.machine.Exist(details.Name) {
-		details.logger().
-			Warningln("Skipping machine removal, because it doesn't exist")
+		logger.Warningln("Skipping machine removal, because it doesn't exist")
 		return nil
 	}
 
 	// This code limits amount of removal of stuck machines to one machine per interval
-	if details.isStuckOnRemove() {
+	if info.isStuckOnRemove() {
 		m.stuckRemoveLock.Lock()
 		defer m.stuckRemoveLock.Unlock()
 	}
 
-	details.logger().Warningln("Stopping machine")
+	logger.Warningln("Stopping machine")
 	err = runHistogramCountedOperation(m.stoppingHistogram, func() error {
 		return m.machine.Stop(details.Name, machineStopCommandTimeout)
 	})
 	if err != nil {
-		details.logger().
+		logger.
 			WithError(err).
 			Warningln("Error while stopping machine")
 	}
 
-	details.logger().Warningln("Removing machine")
+	logger.Warningln("Removing machine")
 	err = runHistogramCountedOperation(m.removalHistogram, func() error {
 		return m.machine.Remove(details.Name)
 	})
 	if err != nil {
+		details.Lock()
 		details.RetryCount++
+		details.Unlock()
 		time.Sleep(removeRetryInterval)
 		return err
 	}
@@ -310,9 +320,13 @@ func (m *machineProvider) finalizeRemoval(details *machineDetails) {
 	defer m.lock.Unlock()
 	delete(m.details, details.Name)
 
+	details.Lock()
+	retryCount := details.RetryCount
+	details.Unlock()
+
 	details.logger().
 		WithField("now", time.Now()).
-		WithField("retries", details.RetryCount).
+		WithField("retries", retryCount).
 		Infoln("Machine removed")
 
 	m.totalActions.WithLabelValues("removed").Inc()
@@ -327,49 +341,25 @@ func (m *machineProvider) remove(machineName string, reason ...interface{}) erro
 		return errors.New("machine not found")
 	}
 
+	now := time.Now()
+
+	details.Lock()
 	details.Reason = fmt.Sprint(reason...)
 	details.State = machineStateRemoving
 	details.RetryCount = 0
 
 	details.logger().
-		WithField("now", time.Now()).
+		WithField("now", now).
 		Warningln("Requesting machine removal")
 
-	details.Used = time.Now()
+	details.Used = now
 	details.writeDebugInformation()
+	details.Unlock()
 
 	go m.finalizeRemoval(details)
 	return nil
 }
 
-func (m *machineProvider) updateMachine(
-	config *common.RunnerConfig,
-	data *machinesData,
-	details *machineDetails,
-) error {
-	if details.State != machineStateIdle {
-		return nil
-	}
-
-	if config.Machine.MaxBuilds > 0 && details.UsedCount >= config.Machine.MaxBuilds {
-		// Limit number of builds
-		return errors.New("too many builds")
-	}
-
-	if data.Total() >= config.Limit && config.Limit > 0 {
-		// Limit maximum number of machines
-		return errors.New("too many machines")
-	}
-
-	if time.Since(details.Used) > time.Second*time.Duration(config.Machine.GetIdleTime()) {
-		if data.Idle >= config.Machine.GetIdleCount() {
-			// Remove machine that are way over the idle time
-			return errors.New("too many idle machines")
-		}
-	}
-	return nil
-}
-
 func (m *machineProvider) updateMachines(
 	machines []string,
 	config *common.RunnerConfig,
@@ -379,35 +369,38 @@ func (m *machineProvider) updateMachines(
 
 	for _, name := range machines {
 		details := m.machineDetails(name, false)
-		details.LastSeen = time.Now()
 
-		err := m.updateMachine(config, &data, details)
-		if err == nil {
+		details.Lock()
+		details.LastSeen = time.Now()
+		info := details.info()
+		details.Unlock()
+
+		reason := shouldRemoveIdle(config, &data, info)
+		if reason == dontRemoveIdleMachine {
 			validMachines = append(validMachines, name)
 		} else {
-			_ = m.remove(details.Name, err)
+			_ = m.remove(details.Name, reason)
 		}
 
-		data.Add(details)
+		// remove() above can mutate details, so we re-create info:
+		details.Lock()
+		info = details.info()
+		details.Unlock()
+
+		data.Add(info)
 	}
 	return
 }
 
+// createMachines starts goroutines that are creating the new machines.
+// Limiting strategy is used to ensure the autoscaling parameters are respected.
 func (m *machineProvider) createMachines(config *common.RunnerConfig, data *machinesData) {
-	// Create a new machines and mark them as Idle
 	for {
-		if data.Available() >= config.Machine.GetIdleCount() {
-			// Limit maximum number of idle machines
-			break
-		}
-		if data.Total() >= config.Limit && config.Limit > 0 {
-			// Limit maximum number of machines
-			break
-		}
-		if data.Creating >= config.Machine.MaxGrowthRate && config.Machine.MaxGrowthRate > 0 {
-			// Prevent excessive growth in the number of machines
-			break
+		if !canCreateIdle(config, data) {
+			return
 		}
+
+		// Create a new machine and mark it as Idle
 		m.create(config, machineStateIdle)
 		data.Creating++
 	}
@@ -426,7 +419,11 @@ func (m *machineProvider) intermediateMachineList(excludedMachines []string) []s
 	defer m.lock.Unlock()
 
 	for _, details := range m.details {
-		if details.isPersistedOnDisk() {
+		details.Lock()
+		persisted := details.isPersistedOnDisk()
+		details.Unlock()
+
+		if persisted {
 			continue
 		}
 
@@ -479,13 +476,15 @@ func (m *machineProvider) Acquire(config *common.RunnerConfig) (common.ExecutorD
 	// Pre-create machines
 	m.createMachines(config, &machinesData)
 
-	logrus.WithFields(machinesData.Fields()).
+	logger := logrus.WithFields(machinesData.Fields()).
 		WithField("runner", config.ShortDescription()).
-		WithField("minIdleCount", config.Machine.GetIdleCount()).
+		WithField("idleCountMin", config.Machine.GetIdleCountMin()).
+		WithField("idleCount", config.Machine.GetIdleCount()).
+		WithField("idleScaleFactor", config.Machine.GetIdleScaleFactor()).
 		WithField("maxMachines", config.Limit).
-		WithField("maxMachineCreate", config.Machine.MaxGrowthRate).
-		WithField("time", time.Now()).
-		Debugln("Docker Machine Details")
+		WithField("maxMachineCreate", config.Machine.MaxGrowthRate)
+
+	logger.WithField("time", time.Now()).Debugln("Docker Machine Details")
 	machinesData.writeDebugInformation()
 
 	// Try to find a free machine
@@ -494,13 +493,13 @@ func (m *machineProvider) Acquire(config *common.RunnerConfig) (common.ExecutorD
 		return details, nil
 	}
 
-	// If we have a free machines we can process a build
-	if config.Machine.GetIdleCount() != 0 && machinesData.Idle == 0 {
-		err = &common.NoFreeExecutorError{
-			Message: "no free machines that can process builds",
-		}
+	if config.Machine.GetIdleCount() == 0 {
+		logger.Debug("IdleCount is set to 0 so the machine will be created on demand in job context")
+	} else if machinesData.Idle == 0 {
+		return nil, &common.NoFreeExecutorError{Message: "no free machines that can process builds"}
 	}
-	return nil, err
+
+	return nil, nil
 }
 
 //nolint:nakedret
@@ -510,7 +509,15 @@ func (m *machineProvider) Use(
 ) (newConfig common.RunnerConfig, newData common.ExecutorData, err error) {
 	// Find a new machine
 	details, _ := data.(*machineDetails)
-	if details == nil || !details.canBeUsed() || !m.machine.CanConnect(details.Name, true) {
+
+	canBeUsed := false
+	if details != nil {
+		details.Lock()
+		canBeUsed = details.canBeUsed()
+		details.Unlock()
+	}
+
+	if !canBeUsed || !m.machine.CanConnect(details.Name, true) {
 		details, err = m.retryUseMachine(config)
 		if err != nil {
 			return
@@ -539,9 +546,12 @@ func (m *machineProvider) Use(
 	newConfig.Docker.Credentials = dc
 
 	// Mark machine as used
+	details.Lock()
 	details.State = machineStateUsed
 	details.Used = time.Now()
 	details.UsedCount++
+	details.Unlock()
+
 	m.totalActions.WithLabelValues("used").Inc()
 	return
 }
@@ -553,24 +563,25 @@ func (m *machineProvider) Release(config *common.RunnerConfig, data common.Execu
 		return
 	}
 
-	m.lock.Lock()
+	details.Lock()
 	// Mark last used time when is Used
 	if details.State == machineStateUsed {
 		details.Used = time.Now()
 	}
-	m.lock.Unlock()
+	usedCount := details.UsedCount
+	details.Unlock()
 
 	// Remove machine if we already used it
 	if config != nil && config.Machine != nil &&
-		config.Machine.MaxBuilds > 0 && details.UsedCount >= config.Machine.MaxBuilds {
+		config.Machine.MaxBuilds > 0 && usedCount >= config.Machine.MaxBuilds {
 		err := m.remove(details.Name, "Too many builds")
 		if err == nil {
 			return
 		}
 	}
-	m.lock.Lock()
+	details.Lock()
 	details.State = machineStateIdle
-	m.lock.Unlock()
+	details.Unlock()
 
 	// Signal pending builds that a new machine is available.
 	if err := m.signalRelease(config); err != nil {
@@ -613,10 +624,12 @@ func (m *machineProvider) Create() common.Executor {
 	}
 }
 
-func newMachineProvider(name, executor string) *machineProvider {
-	provider := common.GetExecutorProvider(executor)
+//nolint:funlen
+func newMachineProvider() *machineProvider {
+	name := "docker+machine"
+	provider := common.GetExecutorProvider("docker")
 	if provider == nil {
-		logrus.Panicln("Missing", executor)
+		logrus.Panicln("docker executor provider not initialized")
 	}
 
 	return &machineProvider{
@@ -673,5 +686,15 @@ func newMachineProvider(name, executor string) *machineProvider {
 				},
 			},
 		),
+		failedCreationHistogram: prometheus.NewHistogram(
+			prometheus.HistogramOpts{
+				Name:    "gitlab_runner_autoscaling_machine_failed_creation_duration_seconds",
+				Help:    "Histogram of machine failed creation timings",
+				Buckets: []float64{1, 3, 5, 10, 30, 50, 60, 80, 90, 120},
+				ConstLabels: prometheus.Labels{
+					"executor": name,
+				},
+			},
+		),
 	}
 }
