@@ -18,9 +18,15 @@ const (
 	ServiceAccountOverwriteVariableName = "KUBERNETES_SERVICE_ACCOUNT_OVERWRITE"
 	// BearerTokenOverwriteVariableValue is the key for the JobVariable containing user overwritten BearerToken
 	BearerTokenOverwriteVariableValue = "KUBERNETES_BEARER_TOKEN"
+	// PodLabelsOverwriteVariablePrefix is the prefix for all the JobVariable keys containing
+	// user overwritten PodLabels
+	PodLabelsOverwriteVariablePrefix = "KUBERNETES_POD_LABELS_"
 	// PodAnnotationsOverwriteVariablePrefix is the prefix for all the JobVariable keys containing
 	// user overwritten PodAnnotations
 	PodAnnotationsOverwriteVariablePrefix = "KUBERNETES_POD_ANNOTATIONS_"
+	// NodeSelectorOverwriteVariablePrefix is the prefix for all the JobVariable keys containing
+	// user overwritten NodeSelectors
+	NodeSelectorOverwriteVariablePrefix = "KUBERNETES_NODE_SELECTOR_"
 	// CPULimitOverwriteVariableValue is the key for the JobVariable containing user overwritten cpu limit
 	CPULimitOverwriteVariableValue = "KUBERNETES_CPU_LIMIT"
 	// CPURequestOverwriteVariableValue is the key for the JobVariable containing user overwritten cpu limit
@@ -105,7 +111,9 @@ type overwrites struct {
 	namespace      string
 	serviceAccount string
 	bearerToken    string
+	podLabels      map[string]string
 	podAnnotations map[string]string
+	nodeSelector   map[string]string
 
 	buildLimits     api.ResourceList
 	serviceLimits   api.ResourceList
@@ -162,6 +170,18 @@ func createOverwrites(
 		return nil, err
 	}
 
+	o.podLabels, err = o.evaluateMapOverwrite(
+		"PodLabels",
+		config.PodLabels,
+		config.PodLabelsOverwriteAllowed,
+		variables,
+		PodLabelsOverwriteVariablePrefix,
+		logger,
+	)
+	if err != nil {
+		return nil, err
+	}
+
 	o.podAnnotations, err = o.evaluateMapOverwrite(
 		"PodAnnotations",
 		config.PodAnnotations,
@@ -174,6 +194,18 @@ func createOverwrites(
 		return nil, err
 	}
 
+	o.nodeSelector, err = o.evaluateMapOverwrite(
+		"NodeSelector",
+		config.NodeSelector,
+		config.NodeSelectorOverwriteAllowed,
+		variables,
+		NodeSelectorOverwriteVariablePrefix,
+		logger,
+	)
+	if err != nil {
+		return nil, err
+	}
+
 	err = o.evaluateMaxBuildResourcesOverwrite(config, variables, logger)
 	if err != nil {
 		return nil, err
@@ -207,9 +239,9 @@ func (o *overwrites) evaluateMaxBuildResourcesOverwrite(
 		config.CPURequestOverwriteMaxAllowed,
 		config.MemoryRequestOverwriteMaxAllowed,
 		config.EphemeralStorageRequestOverwriteMaxAllowed,
-		variables.Get(CPURequestOverwriteVariableValue),
-		variables.Get(MemoryRequestOverwriteVariableValue),
-		variables.Get(EphemeralStorageRequestOverwriteVariableValue),
+		variables.Value(CPURequestOverwriteVariableValue),
+		variables.Value(MemoryRequestOverwriteVariableValue),
+		variables.Value(EphemeralStorageRequestOverwriteVariableValue),
 		logger,
 	)
 	if err != nil {
@@ -226,9 +258,9 @@ func (o *overwrites) evaluateMaxBuildResourcesOverwrite(
 		config.CPULimitOverwriteMaxAllowed,
 		config.MemoryLimitOverwriteMaxAllowed,
 		config.EphemeralStorageLimitOverwriteMaxAllowed,
-		variables.Get(CPULimitOverwriteVariableValue),
-		variables.Get(MemoryLimitOverwriteVariableValue),
-		variables.Get(EphemeralStorageLimitOverwriteVariableValue),
+		variables.Value(CPULimitOverwriteVariableValue),
+		variables.Value(MemoryLimitOverwriteVariableValue),
+		variables.Value(EphemeralStorageLimitOverwriteVariableValue),
 		logger,
 	)
 	if err != nil {
@@ -253,9 +285,9 @@ func (o *overwrites) evaluateMaxServiceResourcesOverwrite(
 		config.ServiceCPURequestOverwriteMaxAllowed,
 		config.ServiceMemoryRequestOverwriteMaxAllowed,
 		config.ServiceEphemeralStorageRequestOverwriteMaxAllowed,
-		variables.Get(ServiceCPURequestOverwriteVariableValue),
-		variables.Get(ServiceMemoryRequestOverwriteVariableValue),
-		variables.Get(ServiceEphemeralStorageRequestOverwriteVariableValue),
+		variables.Value(ServiceCPURequestOverwriteVariableValue),
+		variables.Value(ServiceMemoryRequestOverwriteVariableValue),
+		variables.Value(ServiceEphemeralStorageRequestOverwriteVariableValue),
 		logger,
 	)
 	if err != nil {
@@ -272,9 +304,9 @@ func (o *overwrites) evaluateMaxServiceResourcesOverwrite(
 		config.ServiceCPULimitOverwriteMaxAllowed,
 		config.ServiceMemoryLimitOverwriteMaxAllowed,
 		config.ServiceEphemeralStorageLimitOverwriteMaxAllowed,
-		variables.Get(ServiceCPULimitOverwriteVariableValue),
-		variables.Get(ServiceMemoryLimitOverwriteVariableValue),
-		variables.Get(ServiceEphemeralStorageLimitOverwriteVariableValue),
+		variables.Value(ServiceCPULimitOverwriteVariableValue),
+		variables.Value(ServiceMemoryLimitOverwriteVariableValue),
+		variables.Value(ServiceEphemeralStorageLimitOverwriteVariableValue),
 		logger,
 	)
 	if err != nil {
@@ -299,9 +331,9 @@ func (o *overwrites) evaluateMaxHelperResourcesOverwrite(
 		config.HelperCPURequestOverwriteMaxAllowed,
 		config.HelperMemoryRequestOverwriteMaxAllowed,
 		config.HelperEphemeralStorageRequestOverwriteMaxAllowed,
-		variables.Get(HelperCPURequestOverwriteVariableValue),
-		variables.Get(HelperMemoryRequestOverwriteVariableValue),
-		variables.Get(HelperEphemeralStorageRequestOverwriteVariableValue),
+		variables.Value(HelperCPURequestOverwriteVariableValue),
+		variables.Value(HelperMemoryRequestOverwriteVariableValue),
+		variables.Value(HelperEphemeralStorageRequestOverwriteVariableValue),
 		logger,
 	)
 	if err != nil {
@@ -318,9 +350,9 @@ func (o *overwrites) evaluateMaxHelperResourcesOverwrite(
 		config.HelperCPULimitOverwriteMaxAllowed,
 		config.HelperMemoryLimitOverwriteMaxAllowed,
 		config.HelperEphemeralStorageLimitOverwriteMaxAllowed,
-		variables.Get(HelperCPULimitOverwriteVariableValue),
-		variables.Get(HelperMemoryLimitOverwriteVariableValue),
-		variables.Get(HelperEphemeralStorageLimitOverwriteVariableValue),
+		variables.Value(HelperCPULimitOverwriteVariableValue),
+		variables.Value(HelperMemoryLimitOverwriteVariableValue),
+		variables.Value(HelperEphemeralStorageLimitOverwriteVariableValue),
 		logger,
 	)
 	if err != nil {
