@@ -42,7 +42,7 @@ func (s *executor) verifyMachine(sshPort string) error {
 		return err
 	}
 	defer sshCommand.Cleanup()
-	err = sshCommand.Run(s.Context, ssh.Command{Command: []string{"exit"}})
+	err = sshCommand.Run(s.Context, ssh.Command{Command: "exit"})
 	if err != nil {
 		return err
 	}
@@ -129,7 +129,7 @@ func (s *executor) createVM(baseImage string) (err error) {
 	s.Debugln("Using local", s.sshPort, "SSH port to connect to VM...")
 
 	s.Debugln("Bootstraping VM...")
-	err = vbox.Start(s.vmName)
+	err = s.startVM()
 	if err != nil {
 		return err
 	}
@@ -242,6 +242,19 @@ func (s *executor) getVMName(baseName string) string {
 	)
 }
 
+func (s *executor) startVM() error {
+	s.Debugln("Starting VM...")
+	startType := s.Config.VirtualBox.StartType
+	if startType == "" {
+		startType = "headless"
+	}
+	err := vbox.Start(s.vmName, startType)
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
 func (s *executor) tryRestoreFromSnapshot() {
 	if !vbox.Exist(s.vmName) {
 		return
@@ -269,8 +282,7 @@ func (s *executor) ensureVMStarted() error {
 	}
 
 	if !vbox.IsStatusOnlineOrTransient(status) {
-		s.Println("Starting VM...")
-		err = vbox.Start(s.vmName)
+		err = s.startVM()
 		if err != nil {
 			return err
 		}
@@ -317,9 +329,8 @@ func (s *executor) sshConnect() error {
 
 func (s *executor) Run(cmd common.ExecutorCommand) error {
 	err := s.sshCommand.Run(cmd.Context, ssh.Command{
-		Environment: s.BuildShell.Environment,
-		Command:     s.BuildShell.GetCommandWithArguments(),
-		Stdin:       cmd.Script,
+		Command: s.BuildShell.CmdLine,
+		Stdin:   cmd.Script,
 	})
 	if exitError, ok := err.(*ssh.ExitError); ok {
 		exitCode := exitError.ExitCode()
