@@ -6,10 +6,10 @@ import (
 	"errors"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"os"
 	"os/exec"
 	"path/filepath"
+	"runtime"
 	"strings"
 	"sync"
 	"time"
@@ -34,7 +34,7 @@ type logWriter struct {
 func (l *logWriter) write(line string) {
 	line = strings.TrimRight(line, "\n")
 
-	if len(line) == 0 {
+	if line == "" {
 		return
 	}
 
@@ -42,19 +42,18 @@ func (l *logWriter) write(line string) {
 }
 
 func (l *logWriter) watch() {
-	for {
-		line, err := l.reader.ReadString('\n')
-		if err == nil || err == io.EOF {
-			l.write(line)
-			if err == io.EOF {
-				return
-			}
-		} else {
+	var err error
+	for err != io.EOF {
+		var line string
+		line, err = l.reader.ReadString('\n')
+		if err != nil && err != io.EOF {
 			if !strings.Contains(err.Error(), "bad file descriptor") {
 				logrus.WithError(err).Warn("Problem while reading command output")
 			}
 			return
 		}
+
+		l.write(line)
 	}
 }
 
@@ -170,11 +169,22 @@ func (m *machineCommand) Remove(name string) error {
 }
 
 func (m *machineCommand) List() (hostNames []string, err error) {
-	dir, err := ioutil.ReadDir(mcndirs.GetMachineDir())
+	dir, err := os.ReadDir(mcndirs.GetMachineDir())
 	if err != nil {
-		if os.IsNotExist(err) {
+		errExist := err
+		// On Windows, ReadDir() on a regular file will satisfy ErrNotExist,
+		// due to this bug: https://github.com/golang/go/issues/46734
+		//
+		// For a workaround, we explicitly check whether the directory
+		// exists or not with a Stat call.
+		//nolint:goconst
+		if runtime.GOOS == "windows" {
+			_, errExist = os.Stat(mcndirs.GetMachineDir())
+		}
+		if os.IsNotExist(errExist) {
 			return nil, nil
 		}
+
 		return nil, err
 	}
 
