@@ -13,7 +13,6 @@ const (
 	UseDirectDownload                    string = "FF_USE_DIRECT_DOWNLOAD"
 	SkipNoOpBuildStages                  string = "FF_SKIP_NOOP_BUILD_STAGES"
 	UseFastzip                           string = "FF_USE_FASTZIP"
-	GitLabRegistryHelperImage            string = "FF_GITLAB_REGISTRY_HELPER_IMAGE"
 	DisableUmaskForDockerExecutor        string = "FF_DISABLE_UMASK_FOR_DOCKER_EXECUTOR"
 	EnableBashExitCodeCheck              string = "FF_ENABLE_BASH_EXIT_CODE_CHECK"
 	UseWindowsLegacyProcessStrategy      string = "FF_USE_WINDOWS_LEGACY_PROCESS_STRATEGY"
@@ -24,6 +23,14 @@ const (
 	UseNewShellEscape                    string = "FF_USE_NEW_SHELL_ESCAPE"
 	EnableJobCleanup                     string = "FF_ENABLE_JOB_CLEANUP"
 	KubernetesHonorEntrypoint            string = "FF_KUBERNETES_HONOR_ENTRYPOINT"
+	PosixlyCorrectEscapes                string = "FF_POSIXLY_CORRECT_ESCAPES"
+	UseImprovedURLMasking                string = "FF_USE_IMPROVED_URL_MASKING"
+	ResolveFullTLSChain                  string = "FF_RESOLVE_FULL_TLS_CHAIN"
+	DisablePowershellStdin               string = "FF_DISABLE_POWERSHELL_STDIN"
+	UsePodActiveDeadlineSeconds          string = "FF_USE_POD_ACTIVE_DEADLINE_SECONDS"
+	UseAdvancedPodSpecConfiguration      string = "FF_USE_ADVANCED_POD_SPEC_CONFIGURATION"
+	SetPermissionsBeforeCleanup          string = "FF_SET_PERMISSIONS_BEFORE_CLEANUP"
+	EnableSecretResolvingFailsIfMissing  string = "FF_SECRET_RESOLVING_FAILS_IF_MISSING"
 )
 
 type FeatureFlag struct {
@@ -53,7 +60,7 @@ var flags = []FeatureFlag{
 		DefaultValue:    false,
 		Deprecated:      false,
 		ToBeRemovedWith: "",
-		Description: "Enables creation of a Docker [network per build](../executors/docker.md#networking) with " +
+		Description: "Enables creation of a Docker [network per build](../executors/docker.md#network-configurations) with " +
 			"the `docker` executor",
 	},
 	{
@@ -89,14 +96,6 @@ var flags = []FeatureFlag{
 		ToBeRemovedWith: "",
 		Description:     "Fastzip is a performant archiver for cache/artifact archiving and extraction",
 	},
-	{
-		Name:            GitLabRegistryHelperImage,
-		DefaultValue:    true,
-		Deprecated:      false,
-		ToBeRemovedWith: "",
-		Description: "Use GitLab Runner helper image for the Docker and " +
-			"Kubernetes executors from `registry.gitlab.com` instead of Docker Hub",
-	},
 	{
 		Name:            DisableUmaskForDockerExecutor,
 		DefaultValue:    false,
@@ -125,7 +124,7 @@ var flags = []FeatureFlag{
 		Description: "When disabled, processes that Runner creates on Windows (shell and custom executor) will be " +
 			"created with additional setup that should improve process termination. This is currently experimental " +
 			"and how we setup these processes may change as we continue to improve this. When set to `true`, legacy " +
-			"process setup is used. To successfully and gracefully drain a Windows Runner, this feature flag should" +
+			"process setup is used. To successfully and gracefully drain a Windows Runner, this feature flag should " +
 			"be set to `false`.",
 	},
 	{
@@ -141,7 +140,7 @@ var flags = []FeatureFlag{
 		DefaultValue:    false,
 		Deprecated:      false,
 		ToBeRemovedWith: "",
-		Description: "When enabled, Powershell resolves pathnames rather than Runner using OS-specific filepath " +
+		Description: "When enabled, PowerShell resolves pathnames rather than Runner using OS-specific filepath " +
 			"functions that are specific to where Runner is hosted.",
 	},
 	{
@@ -149,8 +148,8 @@ var flags = []FeatureFlag{
 		DefaultValue:    false,
 		Deprecated:      false,
 		ToBeRemovedWith: "",
-		Description: "When enabled, the trace force send interval is dynamically adjusted based on the trace " +
-			"update interval.",
+		Description: "When enabled, the trace force send interval for logs is dynamically adjusted based on the " +
+			"trace update interval.",
 	},
 	{
 		Name:            ScriptSections,
@@ -184,6 +183,79 @@ var flags = []FeatureFlag{
 		Description: "When enabled, the Docker entrypoint of an image will be honored if " +
 			"`FF_USE_LEGACY_KUBERNETES_EXECUTION_STRATEGY` is not set to true",
 	},
+	{
+		Name:            PosixlyCorrectEscapes,
+		DefaultValue:    false,
+		Deprecated:      false,
+		ToBeRemovedWith: "",
+		Description: "When enabled, [POSIX shell escapes](https://pubs.opengroup.org/onlinepubs/9699919799/utilities/V3_chap02.html#tag_18_02) " +
+			"are used rather than [`bash`-style ANSI-C quoting](https://www.gnu.org/software/bash/manual/html_node/Quoting.html). " +
+			"This should be enabled if the job environment uses a POSIX-compliant shell.",
+	},
+	{
+		Name:         UseImprovedURLMasking,
+		DefaultValue: false,
+		Deprecated:   false,
+		Description: "When enabled, any sensitive URL parameters are masked no matter where they appear in the trace " +
+			"log output. When this is disabled, sensitive URL parameters are only masked in select places and can " +
+			"[occasionally be revealed](https://gitlab.com/gitlab-org/gitlab-runner/-/issues/4625). This feature " +
+			"flag can only be configured via Runner's config and not from a job.",
+	},
+	{
+		Name:         ResolveFullTLSChain,
+		DefaultValue: true,
+		Deprecated:   false,
+		Description: "When enabled, the runner resolves a full TLS " +
+			"chain all the way down to a self-signed root certificate " +
+			"for `CI_SERVER_TLS_CA_FILE`. This was previously " +
+			"[required to make Git HTTPS clones work](tls-self-signed.md#git-cloning) " +
+			"for a Git client built with libcurl prior to v7.68.0 and OpenSSL. " +
+			"However, the process to resolve certificates may fail on " +
+			"some operating systems, such as macOS, that reject root certificates " +
+			"signed with older signature algorithms. " +
+			"If certificate resolution fails, you may need to disable this feature. " +
+			"This feature flag can only be disabled in the " +
+			"[`[runners.feature_flags]` configuration](#enable-feature-flag-in-runner-configuration).",
+	},
+	{
+		Name:         DisablePowershellStdin,
+		DefaultValue: false,
+		Deprecated:   false,
+		Description: "When enabled, PowerShell scripts for shell and custom executors are passed by " +
+			"file, rather than passed and executed via stdin.",
+	},
+	{
+		Name:            UsePodActiveDeadlineSeconds,
+		DefaultValue:    false,
+		Deprecated:      false,
+		ToBeRemovedWith: "",
+		Description: "When enabled, the [pod `activeDeadlineSeconds`]" +
+			"(https://kubernetes.io/docs/reference/kubernetes-api/workload-resources/pod-v1/#lifecycle)" +
+			" is set to the CI/CD job timeout. This flag affects the " +
+			"[pod's lifecycle](../executors/kubernetes.md#pod-lifecycle).",
+	},
+	{
+		Name:            UseAdvancedPodSpecConfiguration,
+		DefaultValue:    false,
+		Deprecated:      false,
+		ToBeRemovedWith: "",
+		Description: "When enabled, the user can set an entire whole pod specification in the `config.toml` file. " +
+			"For more information, see [Overwrite generated pod specifications (Alpha)]" +
+			"(../executors/kubernetes.md#overwrite-generated-pod-specifications-alpha)..",
+	},
+	{
+		Name:         SetPermissionsBeforeCleanup,
+		DefaultValue: true,
+		Deprecated:   false,
+		Description: "When enabled, permissions on directories and files in the project directory are " +
+			"set first, to ensure that deletions during cleanup are successful.",
+	},
+	{
+		Name:         EnableSecretResolvingFailsIfMissing,
+		DefaultValue: true,
+		Deprecated:   false,
+		Description:  "When enabled, secret resolving fails if the value cannot be found.",
+	},
 }
 
 func GetAll() []FeatureFlag {
