@@ -10,6 +10,7 @@ import (
 	"gitlab.com/gitlab-org/gitlab-runner/common"
 	cli_helpers "gitlab.com/gitlab-org/gitlab-runner/helpers/cli"
 	"gitlab.com/gitlab-org/gitlab-runner/log"
+	"gitlab.com/gitlab-org/labkit/fips"
 
 	_ "gitlab.com/gitlab-org/gitlab-runner/cache/azure"
 	_ "gitlab.com/gitlab-org/gitlab-runner/cache/gcs"
@@ -18,7 +19,9 @@ import (
 	_ "gitlab.com/gitlab-org/gitlab-runner/commands/helpers"
 	_ "gitlab.com/gitlab-org/gitlab-runner/executors/custom"
 	_ "gitlab.com/gitlab-org/gitlab-runner/executors/docker"
+	_ "gitlab.com/gitlab-org/gitlab-runner/executors/docker/autoscaler"
 	_ "gitlab.com/gitlab-org/gitlab-runner/executors/docker/machine"
+	_ "gitlab.com/gitlab-org/gitlab-runner/executors/instance"
 	_ "gitlab.com/gitlab-org/gitlab-runner/executors/kubernetes"
 	_ "gitlab.com/gitlab-org/gitlab-runner/executors/parallels"
 	_ "gitlab.com/gitlab-org/gitlab-runner/executors/shell"
@@ -39,6 +42,8 @@ func main() {
 		}
 	}()
 
+	fips.Check()
+
 	app := cli.NewApp()
 	app.Name = filepath.Base(os.Args[0])
 	app.Usage = "a GitLab Runner"
