@@ -5,95 +5,36 @@ import (
 	"context"
 	"fmt"
 	"io"
-	"io/ioutil"
 	"mime/multipart"
 	"net/http"
 	"net/url"
+	"os"
 	"runtime"
 	"strconv"
+	"strings"
 	"sync"
+	"time"
 
-	"github.com/prometheus/client_golang/prometheus"
 	"github.com/sirupsen/logrus"
 
 	"gitlab.com/gitlab-org/gitlab-runner/common"
 	"gitlab.com/gitlab-org/gitlab-runner/helpers"
+	"gitlab.com/gitlab-org/gitlab-runner/helpers/featureflags"
 )
 
+// createdRunnerTokenPrefix is the token prefix used for GitLab UI-created runner tokens
+const createdRunnerTokenPrefix = "glrt-"
 const clientError = -100
 
-var apiRequestStatuses = prometheus.NewDesc(
-	"gitlab_runner_api_request_statuses_total",
-	"The total number of api requests, partitioned by runner, endpoint and status.",
-	[]string{"runner", "endpoint", "status"},
-	nil,
-)
-
-type APIEndpoint string
-
-const (
-	APIEndpointRequestJob APIEndpoint = "request_job"
-	APIEndpointUpdateJob  APIEndpoint = "update_job"
-	APIEndpointPatchTrace APIEndpoint = "patch_trace"
-)
-
-type apiRequestStatusPermutation struct {
-	runnerID string
-	endpoint APIEndpoint
-	status   int
-}
-
-type APIRequestStatusesMap struct {
-	internal map[apiRequestStatusPermutation]int
-	lock     sync.RWMutex
-}
-
-func (arspm *APIRequestStatusesMap) Append(runnerID string, endpoint APIEndpoint, status int) {
-	arspm.lock.Lock()
-	defer arspm.lock.Unlock()
-
-	permutation := apiRequestStatusPermutation{runnerID: runnerID, endpoint: endpoint, status: status}
-
-	if _, ok := arspm.internal[permutation]; !ok {
-		arspm.internal[permutation] = 0
-	}
-
-	arspm.internal[permutation]++
-}
-
-// Describe implements prometheus.Collector.
-func (arspm *APIRequestStatusesMap) Describe(ch chan<- *prometheus.Desc) {
-	ch <- apiRequestStatuses
-}
-
-// Collect implements prometheus.Collector.
-func (arspm *APIRequestStatusesMap) Collect(ch chan<- prometheus.Metric) {
-	arspm.lock.RLock()
-	defer arspm.lock.RUnlock()
-
-	for permutation, count := range arspm.internal {
-		ch <- prometheus.MustNewConstMetric(
-			apiRequestStatuses,
-			prometheus.CounterValue,
-			float64(count),
-			permutation.runnerID,
-			string(permutation.endpoint),
-			strconv.Itoa(permutation.status),
-		)
-	}
-}
-
-func NewAPIRequestStatusesMap() *APIRequestStatusesMap {
-	return &APIRequestStatusesMap{
-		internal: make(map[apiRequestStatusPermutation]int),
-	}
+func TokenIsCreatedRunnerToken(token string) bool {
+	return strings.HasPrefix(token, createdRunnerTokenPrefix)
 }
 
 type GitLabClient struct {
 	clients map[string]*client
 	lock    sync.Mutex
 
-	requestsStatusesMap *APIRequestStatusesMap
+	apiRequestsCollector *APIRequestsCollector
 }
 
 func (n *GitLabClient) getClient(credentials requestCredentials) (c *client, err error) {
@@ -168,6 +109,60 @@ func (n *GitLabClient) getRunnerVersion(config common.RunnerConfig) common.Versi
 	return info
 }
 
+type doRawParams struct {
+	credentials requestCredentials
+	method      string
+	uri         string
+	request     io.Reader
+	requestType string
+	headers     http.Header
+}
+
+// doMeasuredRaw is a decorator that adds metrics measurements through
+// n.apiRequestsCollector to the doRaw() call
+func (n *GitLabClient) doMeasuredRaw(
+	ctx context.Context,
+	log logrus.FieldLogger,
+	runnerID string,
+	systemID string,
+	endpoint apiEndpoint,
+	params doRawParams,
+) (*http.Response, error) {
+	var response *http.Response
+	var err error
+
+	fn := func() int {
+		// Response body is handled after doMeasuredJSON() decorator call
+		// Linting violation here is a false-positive.
+		// nolint:bodyclose
+		response, err = n.doRaw(
+			ctx,
+			params.credentials,
+			params.method,
+			params.uri,
+			params.request,
+			params.requestType,
+			params.headers,
+		)
+
+		if err != nil {
+			return clientError
+		}
+
+		return response.StatusCode
+	}
+
+	n.apiRequestsCollector.Observe(
+		log,
+		runnerID,
+		systemID,
+		endpoint,
+		fn,
+	)
+
+	return response, err
+}
+
 func (n *GitLabClient) doRaw(
 	ctx context.Context,
 	credentials requestCredentials,
@@ -184,6 +179,57 @@ func (n *GitLabClient) doRaw(
 	return c.do(ctx, uri, method, request, requestType, headers)
 }
 
+type doJSONParams struct {
+	credentials requestCredentials
+	method      string
+	uri         string
+	statusCode  int
+	request     interface{}
+	response    interface{}
+}
+
+// doMeasuredJSON is a decorator that adds metrics measurements through
+// n.apiRequestsCollector to the doJSON() call
+func (n *GitLabClient) doMeasuredJSON(
+	ctx context.Context,
+	log logrus.FieldLogger,
+	runnerID string,
+	systemID string,
+	endpoint apiEndpoint,
+	params doJSONParams,
+) (int, string, *http.Response) {
+	var result int
+	var statusText string
+	var httpResponse *http.Response
+
+	fn := func() int {
+		// Response body is handled after doMeasuredJSON() decorator call
+		// Linting violation here is a false-positive.
+		// nolint:bodyclose
+		result, statusText, httpResponse = n.doJSON(
+			ctx,
+			params.credentials,
+			params.method,
+			params.uri,
+			params.statusCode,
+			params.request,
+			params.response,
+		)
+
+		return result
+	}
+
+	n.apiRequestsCollector.Observe(
+		log,
+		runnerID,
+		systemID,
+		endpoint,
+		fn,
+	)
+
+	return result, statusText, httpResponse
+}
+
 func (n *GitLabClient) doJSON(
 	ctx context.Context,
 	credentials requestCredentials,
@@ -191,17 +237,83 @@ func (n *GitLabClient) doJSON(
 	statusCode int,
 	request interface{},
 	response interface{},
+) (int, string, *http.Response) {
+	return n.doJSONWithPAT(ctx, credentials, method, uri, statusCode, "", request, response)
+}
+
+type doJSONWithPATParams struct {
+	credentials requestCredentials
+	method      string
+	uri         string
+	statusCode  int
+	pat         string
+	request     interface{}
+	response    interface{}
+}
+
+// doMeasuredJSONWithPAT is a decorator that adds metrics measurements through
+// n.apiRequestsCollector to the doJSONWithPAT() call
+func (n *GitLabClient) doMeasuredJSONWithPAT(
+	ctx context.Context,
+	log logrus.FieldLogger,
+	runnerID string,
+	systemID string,
+	endpoint apiEndpoint,
+	params doJSONWithPATParams,
+) (int, string, *http.Response) {
+	var result int
+	var statusText string
+	var httpResponse *http.Response
+
+	fn := func() int {
+		// Response body is handled after doJSONWithPATParams() decorator call
+		// Linting violation here is a false-positive.
+		// nolint:bodyclose
+		result, statusText, httpResponse = n.doJSONWithPAT(
+			ctx,
+			params.credentials,
+			params.method,
+			params.uri,
+			params.statusCode,
+			params.pat,
+			params.request,
+			params.response,
+		)
+
+		return result
+	}
+
+	n.apiRequestsCollector.Observe(
+		log,
+		runnerID,
+		systemID,
+		endpoint,
+		fn,
+	)
+
+	return result, statusText, httpResponse
+}
+
+func (n *GitLabClient) doJSONWithPAT(
+	ctx context.Context,
+	credentials requestCredentials,
+	method, uri string,
+	statusCode int,
+	pat string,
+	request interface{},
+	response interface{},
 ) (int, string, *http.Response) {
 	c, err := n.getClient(credentials)
 	if err != nil {
 		return clientError, err.Error(), nil
 	}
 
-	return c.doJSON(ctx, uri, method, statusCode, request, response)
+	return c.doJSONWithPAT(ctx, uri, method, statusCode, pat, request, response)
 }
 
 func (n *GitLabClient) getResponseTLSData(
 	credentials requestCredentials,
+	resolveFullChain bool,
 	response *http.Response,
 ) (ResponseTLSData, error) {
 	c, err := n.getClient(credentials)
@@ -209,7 +321,7 @@ func (n *GitLabClient) getResponseTLSData(
 		return ResponseTLSData{}, fmt.Errorf("couldn't get client: %w", err)
 	}
 
-	return c.getResponseTLSData(response.TLS)
+	return c.getResponseTLSData(response.TLS, resolveFullChain)
 }
 
 func (n *GitLabClient) RegisterRunner(
@@ -253,11 +365,13 @@ func (n *GitLabClient) RegisterRunner(
 	}
 }
 
-func (n *GitLabClient) VerifyRunner(runner common.RunnerCredentials) bool {
+func (n *GitLabClient) VerifyRunner(runner common.RunnerCredentials, systemID string) *common.VerifyRunnerResponse {
 	request := common.VerifyRunnerRequest{
-		Token: runner.Token,
+		Token:    runner.Token,
+		SystemID: systemID,
 	}
 
+	var response common.VerifyRunnerResponse
 	result, statusText, resp := n.doJSON(
 		context.Background(),
 		&runner,
@@ -265,8 +379,20 @@ func (n *GitLabClient) VerifyRunner(runner common.RunnerCredentials) bool {
 		"runners/verify",
 		http.StatusOK,
 		&request,
-		nil,
+		&response,
 	)
+	if result == -1 {
+		// if server is not able to return JSON, let's try plain text (the legacy response format)
+		result, statusText, resp = n.doJSON(
+			context.Background(),
+			&runner,
+			http.MethodPost,
+			"runners/verify",
+			http.StatusOK,
+			&request,
+			nil,
+		)
+	}
 	if resp != nil {
 		defer func() { _ = resp.Body.Close() }()
 	}
@@ -274,17 +400,25 @@ func (n *GitLabClient) VerifyRunner(runner common.RunnerCredentials) bool {
 	switch result {
 	case http.StatusOK:
 		// this is expected due to fact that we ask for non-existing job
-		runner.Log().Println("Verifying runner...", "is alive")
-		return true
+		if TokenIsCreatedRunnerToken(runner.Token) {
+			runner.Log().Println("Verifying runner...", "is valid")
+		} else {
+			runner.Log().Println("Verifying runner...", "is alive")
+		}
+		return &response
 	case http.StatusForbidden:
-		runner.Log().Errorln("Verifying runner...", "is removed")
-		return false
+		if TokenIsCreatedRunnerToken(runner.Token) {
+			runner.Log().Println("Verifying runner...", "is not valid")
+		} else {
+			runner.Log().Errorln("Verifying runner...", "is removed")
+		}
+		return nil
 	case clientError:
 		runner.Log().WithField("status", statusText).Errorln("Verifying runner...", "error")
-		return true
+		return &response
 	default:
 		runner.Log().WithField("status", statusText).Errorln("Verifying runner...", "failed")
-		return true
+		return &response
 	}
 }
 
@@ -323,17 +457,119 @@ func (n *GitLabClient) UnregisterRunner(runner common.RunnerCredentials) bool {
 	}
 }
 
+func (n *GitLabClient) UnregisterRunnerManager(runner common.RunnerCredentials, systemID string) bool {
+	request := common.UnregisterRunnerManagerRequest{
+		Token:    runner.Token,
+		SystemID: systemID,
+	}
+
+	result, statusText, resp := n.doJSON(
+		context.Background(),
+		&runner,
+		http.MethodDelete,
+		"runners/managers",
+		http.StatusNoContent,
+		&request,
+		nil,
+	)
+	if resp != nil {
+		defer func() { _ = resp.Body.Close() }()
+	}
+
+	const baseLogText = "Unregistering runner from GitLab"
+	switch result {
+	case http.StatusNoContent:
+		runner.Log().Println(baseLogText, "succeeded")
+		return true
+	case http.StatusForbidden:
+		runner.Log().Errorln(baseLogText, "forbidden")
+		return false
+	case clientError:
+		runner.Log().WithField("status", statusText).Errorln(baseLogText, "error")
+		return false
+	default:
+		runner.Log().WithField("status", statusText).Errorln(baseLogText, "failed")
+		return false
+	}
+}
+
+func (n *GitLabClient) ResetToken(runner common.RunnerCredentials, systemID string) *common.ResetTokenResponse {
+	return n.resetToken(runner, systemID, "runners/reset_authentication_token", "")
+}
+
+func (n *GitLabClient) ResetTokenWithPAT(
+	runner common.RunnerCredentials,
+	systemID string,
+	pat string,
+) *common.ResetTokenResponse {
+	return n.resetToken(runner, systemID, fmt.Sprintf("runners/%d/reset_authentication_token", runner.ID), pat)
+}
+
+func (n *GitLabClient) resetToken(
+	runner common.RunnerCredentials,
+	systemID string,
+	uri string,
+	pat string,
+) *common.ResetTokenResponse {
+	var request *common.ResetTokenRequest
+	if pat == "" {
+		request = &common.ResetTokenRequest{
+			Token: runner.Token,
+		}
+	}
+
+	var response common.ResetTokenResponse
+
+	result, statusText, resp := n.doMeasuredJSONWithPAT(
+		context.Background(),
+		runner.Log(),
+		runner.ShortDescription(),
+		systemID,
+		apiEndpointResetToken,
+		doJSONWithPATParams{
+			credentials: &runner,
+			method:      http.MethodPost,
+			uri:         uri,
+			statusCode:  http.StatusCreated,
+			pat:         pat,
+			request:     request,
+			response:    &response,
+		},
+	)
+
+	if resp != nil {
+		defer func() { _ = resp.Body.Close() }()
+	}
+
+	const baseLogText = "Resetting runner token..."
+	switch result {
+	case http.StatusCreated:
+		runner.Log().Println(baseLogText, "succeeded")
+		response.TokenObtainedAt = time.Now().UTC()
+		return &response
+	case http.StatusForbidden:
+		runner.Log().Errorln(baseLogText, "failed (check used token)")
+		return nil
+	case clientError:
+		runner.Log().WithField("status", statusText).Errorln(baseLogText, "error")
+		return nil
+	default:
+		runner.Log().WithField("status", statusText).Errorln(baseLogText, "failed")
+		return nil
+	}
+}
+
 func addTLSData(response *common.JobResponse, tlsData ResponseTLSData) {
 	if tlsData.CAChain != "" {
 		response.TLSCAChain = tlsData.CAChain
 	}
 
 	if tlsData.CertFile != "" && tlsData.KeyFile != "" {
-		data, err := ioutil.ReadFile(tlsData.CertFile)
+		data, err := os.ReadFile(tlsData.CertFile)
 		if err == nil {
 			response.TLSAuthCert = string(data)
 		}
-		data, err = ioutil.ReadFile(tlsData.KeyFile)
+		data, err = os.ReadFile(tlsData.KeyFile)
 		if err == nil {
 			response.TLSAuthKey = string(data)
 		}
@@ -348,22 +584,28 @@ func (n *GitLabClient) RequestJob(
 	request := common.JobRequest{
 		Info:       n.getRunnerVersion(config),
 		Token:      config.Token,
+		SystemID:   config.SystemIDState.GetSystemID(),
 		LastUpdate: n.getLastUpdate(&config.RunnerCredentials),
 		Session:    sessionInfo,
 	}
 
 	var response common.JobResponse
-	result, statusText, httpResponse := n.doJSON(
-		ctx,
-		&config.RunnerCredentials,
-		http.MethodPost,
-		"jobs/request",
-		http.StatusCreated,
-		&request,
-		&response,
-	)
 
-	n.requestsStatusesMap.Append(config.RunnerCredentials.ShortDescription(), APIEndpointRequestJob, result)
+	//nolint:bodyclose
+	result, statusText, httpResponse := n.doMeasuredJSON(
+		ctx,
+		config.Log(),
+		config.RunnerCredentials.ShortDescription(),
+		config.SystemIDState.GetSystemID(),
+		apiEndpointRequestJob,
+		doJSONParams{
+			credentials: &config.RunnerCredentials,
+			method:      http.MethodPost,
+			uri:         "jobs/request",
+			statusCode:  http.StatusCreated,
+			request:     &request, response: &response,
+		},
+	)
 
 	switch result {
 	case http.StatusCreated:
@@ -372,7 +614,8 @@ func (n *GitLabClient) RequestJob(
 			"repo_url": response.RepoCleanURL(),
 		}).Println("Checking for jobs...", "received")
 
-		tlsData, err := n.getResponseTLSData(&config.RunnerCredentials, httpResponse)
+		resolveFullChain := config.IsFeatureFlagOn(featureflags.ResolveFullTLSChain)
+		tlsData, err := n.getResponseTLSData(&config.RunnerCredentials, resolveFullChain, httpResponse)
 		if err != nil {
 			config.Log().
 				WithError(err).Errorln("Error on fetching TLS Data from API response...", "error")
@@ -384,7 +627,7 @@ func (n *GitLabClient) RequestJob(
 		config.Log().Errorln("Checking for jobs...", "forbidden")
 		return nil, false
 	case http.StatusNoContent:
-		config.Log().Debugln("Checking for jobs...", "nothing")
+		config.Log().Debug("Checking for jobs...", "nothing")
 		return nil, true
 	case clientError:
 		config.Log().WithField("status", statusText).Errorln("Checking for jobs...", "error")
@@ -410,18 +653,29 @@ func (n *GitLabClient) UpdateJob(
 		ExitCode:      jobInfo.ExitCode,
 	}
 
-	statusCode, statusText, response := n.doJSON(
-		context.Background(),
-		&config.RunnerCredentials,
-		http.MethodPut,
-		fmt.Sprintf("jobs/%d", jobInfo.ID),
-		http.StatusOK,
-		&request,
-		nil,
-	)
-	n.requestsStatusesMap.Append(config.RunnerCredentials.ShortDescription(), APIEndpointUpdateJob, statusCode)
+	log := config.Log().
+		WithField("job", jobInfo.ID).
+		WithField("checksum", request.Output.Checksum).
+		WithField("bytesize", request.Output.Bytesize)
 
-	log := config.Log().WithField("job", jobInfo.ID)
+	log.Info("Updating job...")
+
+	//nolint:bodyclose
+	statusCode, statusText, response := n.doMeasuredJSON(
+		context.Background(),
+		config.Log(),
+		config.RunnerCredentials.ShortDescription(),
+		config.SystemIDState.GetSystemID(),
+		apiEndpointUpdateJob,
+		doJSONParams{
+			credentials: &config.RunnerCredentials,
+			method:      http.MethodPut,
+			uri:         fmt.Sprintf("jobs/%d", jobInfo.ID),
+			statusCode:  http.StatusOK,
+			request:     &request,
+			response:    nil,
+		},
+	)
 
 	return n.createUpdateJobResult(log, statusCode, statusText, response)
 }
@@ -450,13 +704,13 @@ func (n *GitLabClient) createUpdateJobResult(
 		log.Warningln("Submitting job to coordinator...", "job failed")
 		result.State = common.UpdateAbort
 	case statusCode == http.StatusOK:
-		log.Debugln("Submitting job to coordinator...", "ok")
+		log.Info("Submitting job to coordinator...", "ok")
 		result.State = common.UpdateSucceeded
 	case statusCode == http.StatusAccepted:
-		log.Debugln("Submitting job to coordinator...", "accepted, but not yet completed")
+		log.Info("Submitting job to coordinator...", "accepted, but not yet completed")
 		result.State = common.UpdateAcceptedButNotCompleted
 	case statusCode == http.StatusPreconditionFailed:
-		log.Debugln("Submitting job to coordinator...", "trace validation failed")
+		log.Info("Submitting job to coordinator...", "trace validation failed")
 		result.State = common.UpdateTraceValidationFailed
 	case statusCode == http.StatusNotFound:
 		log.Warningln("Submitting job to coordinator...", "not found")
@@ -480,12 +734,13 @@ func (n *GitLabClient) PatchTrace(
 	jobCredentials *common.JobCredentials,
 	content []byte,
 	startOffset int,
+	debugTraceEnabled bool,
 ) common.PatchTraceResult {
 	id := jobCredentials.ID
 
 	baseLog := config.Log().WithField("job", id)
 	if len(content) == 0 {
-		baseLog.Debugln("Appending trace to coordinator...", "skipped due to empty patch")
+		baseLog.Info("Appending trace to coordinator...", "skipped due to empty patch")
 		return common.NewPatchTraceResult(startOffset, common.PatchSucceeded, 0)
 	}
 
@@ -496,31 +751,28 @@ func (n *GitLabClient) PatchTrace(
 	headers.Set("Content-Range", contentRange)
 	headers.Set("JOB-TOKEN", jobCredentials.Token)
 
-	uri := fmt.Sprintf("jobs/%d/trace", id)
-	request := bytes.NewReader(content)
-
-	response, err := n.doRaw(
+	response, err := n.doMeasuredRaw(
 		context.Background(),
-		&config.RunnerCredentials,
-		"PATCH",
-		uri,
-		request,
-		"text/plain",
-		headers,
+		config.Log(),
+		config.RunnerCredentials.ShortDescription(),
+		config.SystemIDState.GetSystemID(),
+		apiEndpointPatchTrace,
+		doRawParams{
+			credentials: &config.RunnerCredentials,
+			method:      "PATCH",
+			uri:         fmt.Sprintf("jobs/%d/trace?%s", id, patchTraceQuery(debugTraceEnabled)),
+			request:     bytes.NewReader(content),
+			requestType: "text/plain",
+			headers:     headers,
+		},
 	)
 	if err != nil {
 		config.Log().Errorln("Appending trace to coordinator...", "error", err.Error())
 		return common.NewPatchTraceResult(startOffset, common.PatchFailed, 0)
 	}
 
-	n.requestsStatusesMap.Append(
-		config.RunnerCredentials.ShortDescription(),
-		APIEndpointPatchTrace,
-		response.StatusCode,
-	)
-
 	defer func() {
-		_, _ = io.Copy(ioutil.Discard, response.Body)
+		_, _ = io.Copy(io.Discard, response.Body)
 		_ = response.Body.Close()
 	}()
 
@@ -537,6 +789,13 @@ func (n *GitLabClient) PatchTrace(
 	return n.createPatchTraceResult(startOffset, tracePatchResponse, response, endOffset, log)
 }
 
+func patchTraceQuery(debugTraceEnabled bool) string {
+	query := url.Values{}
+	query.Set("debug_trace", strconv.FormatBool(debugTraceEnabled))
+
+	return query.Encode()
+}
+
 func (n *GitLabClient) createPatchTraceResult(
 	startOffset int,
 	tracePatchResponse *TracePatchResponse,
@@ -558,7 +817,7 @@ func (n *GitLabClient) createPatchTraceResult(
 		return result
 
 	case response.StatusCode == http.StatusAccepted:
-		log.Debugln("Appending trace to coordinator...", "ok")
+		log.Info("Appending trace to coordinator...", "ok")
 		result.SentOffset = endOffset
 		result.State = common.PatchSucceeded
 
@@ -639,7 +898,7 @@ func (n *GitLabClient) UploadRawArtifacts(
 	config common.JobCredentials,
 	reader io.ReadCloser,
 	options common.ArtifactsOptions,
-) common.UploadState {
+) (common.UploadState, string) {
 	defer func() {
 		_ = reader.Close()
 	}()
@@ -682,15 +941,15 @@ func (n *GitLabClient) UploadRawArtifacts(
 
 	if err != nil {
 		log.WithError(err).Errorln(messagePrefix, "error")
-		return common.UploadFailed
+		return common.UploadFailed, ""
 	}
 
 	defer func() {
-		_, _ = io.Copy(ioutil.Discard, res.Body)
+		_, _ = io.Copy(io.Discard, res.Body)
 		_ = res.Body.Close()
 	}()
 
-	return n.determineUploadState(res.StatusCode, log, messagePrefix)
+	return n.determineUploadState(res, log, messagePrefix)
 }
 
 func closeWithLogging(log logrus.FieldLogger, c io.Closer, name string) {
@@ -701,29 +960,48 @@ func closeWithLogging(log logrus.FieldLogger, c io.Closer, name string) {
 }
 
 func (n *GitLabClient) determineUploadState(
-	statusCode int,
+	resp *http.Response,
 	log *logrus.Entry,
 	messagePrefix string,
-) common.UploadState {
-	switch statusCode {
+) (common.UploadState, string) {
+	statusText := getMessageFromJSONResponse(resp)
+
+	switch resp.StatusCode {
 	case http.StatusCreated:
-		log.Println(messagePrefix, "ok")
-		return common.UploadSucceeded
+		log.Println(messagePrefix, statusText)
+		return common.UploadSucceeded, ""
+	case http.StatusTemporaryRedirect:
+		return handleUploadRedirectionState(resp, log, messagePrefix, statusText)
 	case http.StatusForbidden:
-		log.WithField("status", statusCode).Errorln(messagePrefix, "forbidden")
-		return common.UploadForbidden
+		log.WithField("status", resp.StatusCode).Errorln(messagePrefix, statusText)
+		return common.UploadForbidden, ""
 	case http.StatusRequestEntityTooLarge:
-		log.WithField("status", statusCode).Errorln(messagePrefix, "too large archive")
-		return common.UploadTooLarge
+		log.WithField("status", resp.StatusCode).Errorln(messagePrefix, statusText)
+		return common.UploadTooLarge, ""
 	case http.StatusServiceUnavailable:
-		log.WithField("status", statusCode).Errorln(messagePrefix, "service unavailable")
-		return common.UploadServiceUnavailable
+		log.WithField("status", resp.StatusCode).Errorln(messagePrefix, statusText)
+		return common.UploadServiceUnavailable, ""
 	default:
-		log.WithField("status", statusCode).Warningln(messagePrefix, "failed")
-		return common.UploadFailed
+		log.WithField("status", resp.StatusCode).Warningln(messagePrefix, statusText)
+		return common.UploadFailed, ""
 	}
 }
 
+func handleUploadRedirectionState(
+	resp *http.Response,
+	log *logrus.Entry,
+	messagePrefix string,
+	statusText string,
+) (common.UploadState, string) {
+	location := resp.Header.Get("Location")
+	if location == "" {
+		log.WithField("status", resp.StatusCode).Errorln(messagePrefix, statusText, "empty location")
+		return common.UploadFailed, ""
+	}
+
+	return common.UploadRedirected, location
+}
+
 func (n *GitLabClient) DownloadArtifacts(
 	config common.JobCredentials,
 	artifactsFile io.WriteCloser,
@@ -748,6 +1026,10 @@ func (n *GitLabClient) DownloadArtifacts(
 
 	if res != nil {
 		log = log.WithField("responseStatus", res.Status)
+
+		if res.Request != nil && res.Request.URL != nil {
+			log = log.WithField("host", res.Request.URL.Host)
+		}
 	}
 
 	if err != nil {
@@ -755,7 +1037,7 @@ func (n *GitLabClient) DownloadArtifacts(
 		return common.DownloadFailed
 	}
 	defer func() {
-		_, _ = io.Copy(ioutil.Discard, res.Body)
+		_, _ = io.Copy(io.Discard, res.Body)
 		_ = res.Body.Close()
 	}()
 
@@ -763,8 +1045,16 @@ func (n *GitLabClient) DownloadArtifacts(
 	case http.StatusOK:
 		return n.downloadArtifactFile(log, artifactsFile, res)
 	case http.StatusForbidden:
-		log.WithField("status", res.Status).Errorln("Downloading artifacts from coordinator...", "forbidden")
+		// We generally expect JSON responses from the GitLab API, but a
+		// 302 redirection to object storage may result in an XML
+		// response that might include important details why the request
+		// was rejected (e.g. Google VPC Service Controls).
+		statusText := getMessageFromJSONOrXMLResponse(res)
+		log.WithField("status", statusText).Errorln("Downloading artifacts from coordinator...", "forbidden")
 		return common.DownloadForbidden
+	case http.StatusUnauthorized:
+		log.WithField("status", res.Status).Errorln("Downloading artifacts from coordinator...", "unauthorized")
+		return common.DownloadUnauthorized
 	case http.StatusNotFound:
 		log.Errorln("Downloading artifacts from coordinator...", "not found")
 		return common.DownloadNotFound
@@ -806,12 +1096,12 @@ func (n *GitLabClient) ProcessJob(
 	return trace, nil
 }
 
-func NewGitLabClientWithRequestStatusesMap(rsMap *APIRequestStatusesMap) *GitLabClient {
+func NewGitLabClientWithAPIRequestsCollector(c *APIRequestsCollector) *GitLabClient {
 	return &GitLabClient{
-		requestsStatusesMap: rsMap,
+		apiRequestsCollector: c,
 	}
 }
 
 func NewGitLabClient() *GitLabClient {
-	return NewGitLabClientWithRequestStatusesMap(NewAPIRequestStatusesMap())
+	return NewGitLabClientWithAPIRequestsCollector(NewAPIRequestsCollector())
 }
