@@ -1,4 +1,4 @@
-// +build !integration
+//go:build !integration
 
 package shells
 
@@ -7,6 +7,7 @@ import (
 	"path/filepath"
 	"strings"
 	"testing"
+	"time"
 
 	"github.com/stretchr/testify/assert"
 	"github.com/stretchr/testify/mock"
@@ -40,21 +41,21 @@ func TestWriteGitSSLConfig(t *testing.T) {
 	mockWriter.On("EnvVariableKey", tls.VariableKeyFile).Return("VariableKeyFile").Once()
 
 	mockWriter.On(
-		"Command",
+		"CommandArgExpand",
 		"git",
 		"config",
 		fmt.Sprintf("http.%s.%s", expectedURL, "sslCAInfo"),
 		"VariableCAFile",
 	).Once()
 	mockWriter.On(
-		"Command",
+		"CommandArgExpand",
 		"git",
 		"config",
 		fmt.Sprintf("http.%s.%s", expectedURL, "sslCert"),
 		"VariableCertFile",
 	).Once()
 	mockWriter.On(
-		"Command",
+		"CommandArgExpand",
 		"git",
 		"config",
 		fmt.Sprintf("http.%s.%s", expectedURL, "sslKey"),
@@ -716,7 +717,7 @@ func TestGitCleanFlags(t *testing.T) {
 			mockWriter := new(MockShellWriter)
 			defer mockWriter.AssertExpectations(t)
 
-			mockWriter.On("Noticef", "Checking out %s as %s...", dummySha[0:8], dummyRef).Once()
+			mockWriter.On("Noticef", "Checking out %s as detached HEAD (ref is %s)...", dummySha[0:8], dummyRef).Once()
 			mockWriter.On("Command", "git", "checkout", "-f", "-q", dummySha).Once()
 
 			if test.expectedGitClean {
@@ -797,6 +798,12 @@ func TestGitFetchFlags(t *testing.T) {
 }
 
 func TestAbstractShell_writeSubmoduleUpdateCmd(t *testing.T) {
+	const (
+		exampleBaseURL  = "http://test.remote"
+		exampleJobToken = "job-token"
+		insteadOf       = "url.http://gitlab-ci-token:job-token@test.remote.insteadOf=http://test.remote"
+	)
+
 	tests := map[string]struct {
 		Recursive               bool
 		Depth                   int
@@ -847,6 +854,11 @@ func TestAbstractShell_writeSubmoduleUpdateCmd(t *testing.T) {
 				)
 			}
 			mockWriter.On("Noticef", tc.ExpectedNoticeArgs...).Once()
+			mockWriter.
+				On(
+					"Command",
+					[]interface{}{"git", "submodule", "init"}...,
+				).Once()
 			mockWriter.
 				On(
 					"Command",
@@ -855,7 +867,8 @@ func TestAbstractShell_writeSubmoduleUpdateCmd(t *testing.T) {
 			mockWriter.
 				On(
 					"Command",
-					append([]interface{}{"git", "submodule", "update", "--init"}, tc.ExpectedGitUpdateFlags...)...,
+					append([]interface{}{
+						"git", "-c", insteadOf, "submodule", "update", "--init"}, tc.ExpectedGitUpdateFlags...)...,
 				).Once()
 			cleanCmd := mockWriter.
 				On("Command", append(expectedGitForEachArgsFn(), "git clean -ffxd")...).
@@ -869,24 +882,41 @@ func TestAbstractShell_writeSubmoduleUpdateCmd(t *testing.T) {
 			mockWriter.On("Command", append(expectedGitForEachArgsFn(), "git lfs pull")...).Once()
 			mockWriter.On("EndIf").Once()
 
-			shell.writeSubmoduleUpdateCmd(
+			err := shell.writeSubmoduleUpdateCmd(
 				mockWriter,
 				&common.Build{
 					JobResponse: common.JobResponse{
 						GitInfo: common.GitInfo{Depth: tc.Depth},
+						Token:   exampleJobToken,
+					},
+					Runner: &common.RunnerConfig{
+						RunnerCredentials: common.RunnerCredentials{URL: exampleBaseURL},
 					},
 				},
 				tc.Recursive,
 			)
+			assert.NoError(t, err)
 		})
 	}
 }
 
-func TestAbstractShell_extractCacheWithFallbackKey(t *testing.T) {
+func TestAbstractShell_extractCacheWithDefaultFallbackKey(t *testing.T) {
 	testCacheKey := "test-cache-key"
-	testFallbackCacheKey := "test-fallback-cache-key"
 
-	shell := AbstractShell{}
+	tests := map[string]struct {
+		cacheFallbackKeyVarValue string
+		expectedCacheKey         string
+	}{
+		"using allowed key value": {
+			cacheFallbackKeyVarValue: "test-fallback-cache-key",
+			expectedCacheKey:         "test-fallback-cache-key",
+		},
+		"using reserved suffix": {
+			cacheFallbackKeyVarValue: "main-protected",
+			expectedCacheKey:         "test-cache-key",
+		},
+	}
+
 	runnerConfig := &common.RunnerConfig{
 		RunnerSettings: common.RunnerSettings{
 			Cache: &common.CacheConfig{
@@ -895,77 +925,211 @@ func TestAbstractShell_extractCacheWithFallbackKey(t *testing.T) {
 			},
 		},
 	}
-	build := &common.Build{
-		BuildDir: "/builds",
-		CacheDir: "/cache",
-		Runner:   runnerConfig,
-		JobResponse: common.JobResponse{
-			ID: 1000,
-			JobInfo: common.JobInfo{
-				ProjectID: 1000,
-			},
-			Cache: common.Caches{
-				{
-					Key:    testCacheKey,
-					Policy: common.CachePolicyPullPush,
-					Paths:  []string{"path1", "path2"},
+	shell := AbstractShell{}
+
+	for tn, tc := range tests {
+		t.Run(tn, func(t *testing.T) {
+			build := &common.Build{
+				BuildDir: "/builds",
+				CacheDir: "/cache",
+				Runner:   runnerConfig,
+				JobResponse: common.JobResponse{
+					ID: 1000,
+					JobInfo: common.JobInfo{
+						ProjectID: 1000,
+					},
+					Cache: common.Caches{
+						{
+							Key:    testCacheKey,
+							Policy: common.CachePolicyPullPush,
+							Paths:  []string{"path1", "path2"},
+						},
+					},
+					Variables: common.JobVariables{
+						{
+							Key:   "CACHE_FALLBACK_KEY",
+							Value: tc.cacheFallbackKeyVarValue,
+						},
+					},
 				},
+			}
+			info := common.ShellScriptInfo{
+				RunnerCommand: "runner-command",
+				Build:         build,
+			}
+
+			mockWriter := new(MockShellWriter)
+			defer mockWriter.AssertExpectations(t)
+
+			mockWriter.On("IfCmd", "runner-command", "--version").Once()
+			mockWriter.On("Noticef", "Checking cache for %s...", testCacheKey).Once()
+			mockWriter.On(
+				"IfCmdWithOutput",
+				"runner-command",
+				"cache-extractor",
+				"--file",
+				filepath.Join("..", build.CacheDir, testCacheKey, "cache.zip"),
+				"--timeout",
+				"10",
+				"--url",
+				fmt.Sprintf("test://download/project/1000/%s", testCacheKey),
+			).Once()
+			mockWriter.On("Noticef", "Successfully extracted cache").Once()
+			mockWriter.On("Else").Once()
+			mockWriter.On("Warningf", "Failed to extract cache").Once()
+			if tc.cacheFallbackKeyVarValue == tc.expectedCacheKey {
+				mockWriter.On("Noticef", "Checking cache for %s...", tc.expectedCacheKey).Once()
+				mockWriter.On(
+					"IfCmdWithOutput",
+					"runner-command",
+					"cache-extractor",
+					"--file",
+					filepath.Join("..", build.CacheDir, testCacheKey, "cache.zip"),
+					"--timeout",
+					"10",
+					"--url",
+					fmt.Sprintf("test://download/project/1000/%s", tc.expectedCacheKey),
+				).Once()
+				mockWriter.On("Noticef", "Successfully extracted cache").Once()
+				mockWriter.On("Else").Once()
+				mockWriter.On("Warningf", "Failed to extract cache").Once()
+				mockWriter.On("EndIf").Once()
+			}
+			mockWriter.On("EndIf").Once()
+			mockWriter.On("Else").Once()
+			mockWriter.On("Warningf", "Missing %s. %s is disabled.", "runner-command", "Extracting cache").Once()
+			mockWriter.On("EndIf").Once()
+
+			err := shell.cacheExtractor(mockWriter, info)
+			assert.NoError(t, err)
+		})
+	}
+}
+
+func TestAbstractShell_extractCacheWithMultipleFallbackKeys(t *testing.T) {
+	testCacheKey := "test-cache-key"
+	tests := map[string]struct {
+		cacheFallbackKeyVarValue string
+		cacheFallbackKeysValues  []string
+		allowedCacheKeys         []string
+	}{
+		"multiple fallback keys": {
+			cacheFallbackKeyVarValue: "test-var-fallback-cache-key",
+			cacheFallbackKeysValues: []string{
+				"test-fallback-cache-key-1",
+				"test-fallback-cache-key-2",
 			},
-			Variables: common.JobVariables{
-				{
-					Key:   "CACHE_FALLBACK_KEY",
-					Value: testFallbackCacheKey,
-				},
+			allowedCacheKeys: []string{
+				"test-cache-key",
+				"test-fallback-cache-key-1",
+				"test-fallback-cache-key-2",
+				"test-var-fallback-cache-key",
+			},
+		},
+		"protected fallback keys": {
+			cacheFallbackKeyVarValue: "test-var-fallback-cache-key",
+			cacheFallbackKeysValues: []string{
+				"test-fallback-protected-1",
+				"test-fallback-protected-2",
+			},
+			allowedCacheKeys: []string{
+				"test-cache-key",
+				"test-fallback-protected-1",
+				"test-fallback-protected-2",
+				"test-var-fallback-cache-key",
+			},
+		},
+		"invalid global protected fallback key": {
+			cacheFallbackKeyVarValue: "test-var-fallback-key-protected",
+			cacheFallbackKeysValues: []string{
+				"test-fallback-cache-key-1",
+				"test-fallback-cache-key-2",
+			},
+			allowedCacheKeys: []string{
+				"test-cache-key",
+				"test-fallback-cache-key-1",
+				"test-fallback-cache-key-2",
 			},
 		},
 	}
-	info := common.ShellScriptInfo{
-		RunnerCommand: "runner-command",
-		Build:         build,
+
+	runnerConfig := &common.RunnerConfig{
+		RunnerSettings: common.RunnerSettings{
+			Cache: &common.CacheConfig{
+				Type:   "test",
+				Shared: true,
+			},
+		},
 	}
+	shell := AbstractShell{}
 
-	mockWriter := new(MockShellWriter)
-	defer mockWriter.AssertExpectations(t)
+	for tn, tc := range tests {
+		t.Run(tn, func(t *testing.T) {
+			build := &common.Build{
+				BuildDir: "/builds",
+				CacheDir: "/cache",
+				Runner:   runnerConfig,
+				JobResponse: common.JobResponse{
+					ID: 1000,
+					JobInfo: common.JobInfo{
+						ProjectID: 1000,
+					},
+					Cache: common.Caches{
+						{
+							Key:          testCacheKey,
+							Policy:       common.CachePolicyPullPush,
+							Paths:        []string{"path1", "path2"},
+							FallbackKeys: tc.cacheFallbackKeysValues,
+						},
+					},
+					Variables: common.JobVariables{
+						{
+							Key:   "CACHE_FALLBACK_KEY",
+							Value: tc.cacheFallbackKeyVarValue,
+						},
+					},
+				},
+			}
+			info := common.ShellScriptInfo{
+				RunnerCommand: "runner-command",
+				Build:         build,
+			}
 
-	mockWriter.On("IfCmd", "runner-command", "--version").Once()
-	mockWriter.On("Noticef", "Checking cache for %s...", testCacheKey).Once()
-	mockWriter.On(
-		"IfCmdWithOutput",
-		"runner-command",
-		"cache-extractor",
-		"--file",
-		filepath.Join("..", build.CacheDir, testCacheKey, "cache.zip"),
-		"--timeout",
-		"10",
-		"--url",
-		fmt.Sprintf("test://download/project/1000/%s", testCacheKey),
-	).Once()
-	mockWriter.On("Noticef", "Successfully extracted cache").Once()
-	mockWriter.On("Else").Once()
-	mockWriter.On("Warningf", "Failed to extract cache").Once()
-	mockWriter.On("Noticef", "Checking cache for %s...", testFallbackCacheKey).Once()
-	mockWriter.On(
-		"IfCmdWithOutput",
-		"runner-command",
-		"cache-extractor",
-		"--file",
-		filepath.Join("..", build.CacheDir, testCacheKey, "cache.zip"),
-		"--timeout",
-		"10",
-		"--url",
-		fmt.Sprintf("test://download/project/1000/%s", testFallbackCacheKey),
-	).Once()
-	mockWriter.On("Noticef", "Successfully extracted cache").Once()
-	mockWriter.On("Else").Once()
-	mockWriter.On("Warningf", "Failed to extract cache").Once()
-	mockWriter.On("EndIf").Once()
-	mockWriter.On("EndIf").Once()
-	mockWriter.On("Else").Once()
-	mockWriter.On("Warningf", "Missing %s. %s is disabled.", "runner-command", "Extracting cache").Once()
-	mockWriter.On("EndIf").Once()
+			mockWriter := new(MockShellWriter)
+			defer mockWriter.AssertExpectations(t)
 
-	err := shell.cacheExtractor(mockWriter, info)
-	assert.NoError(t, err)
+			mockWriter.On("IfCmd", "runner-command", "--version").Once()
+
+			for _, cacheKey := range tc.allowedCacheKeys {
+				mockWriter.On("Noticef", "Checking cache for %s...", cacheKey).Once()
+				mockWriter.On(
+					"IfCmdWithOutput",
+					"runner-command",
+					"cache-extractor",
+					"--file",
+					filepath.Join("..", build.CacheDir, testCacheKey, "cache.zip"),
+					"--timeout",
+					"10",
+					"--url",
+					fmt.Sprintf("test://download/project/1000/%s", cacheKey),
+				).Once()
+				mockWriter.On("Noticef", "Successfully extracted cache").Once()
+				mockWriter.On("Else").Once()
+				mockWriter.On("Warningf", "Failed to extract cache").Once()
+			}
+
+			for range tc.allowedCacheKeys {
+				mockWriter.On("EndIf").Once()
+			}
+
+			mockWriter.On("Else").Once()
+			mockWriter.On("Warningf", "Missing %s. %s is disabled.", "runner-command", "Extracting cache").Once()
+			mockWriter.On("EndIf").Once()
+
+			err := shell.cacheExtractor(mockWriter, info)
+			assert.NoError(t, err)
+		})
+	}
 }
 
 func TestAbstractShell_writeSubmoduleUpdateCmdPath(t *testing.T) {
@@ -999,23 +1163,25 @@ func TestAbstractShell_writeSubmoduleUpdateCmdPath(t *testing.T) {
 	}
 
 	submoduleCommand := func(paths string, args ...string) []interface{} {
-		command := []interface{}{
-			"git",
-			"submodule",
-		}
+		command := []interface{}{"git"}
 
 		for _, a := range args {
 			command = append(command, a)
 		}
 
-		paths = strings.TrimSpace(paths)
-		if paths != "" {
-			command = append(command, "--", paths)
+		subpaths := strings.Fields(paths)
+		if len(subpaths) != 0 {
+			command = append(command, "--")
+			for i := 0; i < len(subpaths); i++ {
+				command = append(command, subpaths[i])
+			}
 		}
 
 		return command
 	}
 
+	insteadOf := "url.https://gitlab-ci-token:xxx@example.com.insteadOf=https://example.com"
+
 	for name, test := range tests {
 		t.Run(name, func(t *testing.T) {
 			shell := AbstractShell{}
@@ -1023,8 +1189,12 @@ func TestAbstractShell_writeSubmoduleUpdateCmdPath(t *testing.T) {
 			defer mockWriter.AssertExpectations(t)
 
 			mockWriter.On("Noticef", mock.Anything).Once()
-			mockWriter.On("Command", submoduleCommand(test.paths, "sync")...).Once()
-			mockWriter.On("Command", submoduleCommand(test.paths, "update", "--init")...).Once()
+			mockWriter.On("Command", "git", "submodule", "init").Once()
+			mockWriter.On("Command", submoduleCommand(test.paths, "submodule", "sync")...).Once()
+			mockWriter.On("Command", submoduleCommand(
+				test.paths, "-c", insteadOf, "submodule", "update", "--init",
+			)...,
+			).Once()
 			cleanCmd := mockWriter.On("Command", "git", "submodule", "foreach", "git clean -ffxd").Once()
 			mockWriter.On("Command", "git", "submodule", "foreach", "git reset --hard").
 				Run(func(args mock.Arguments) {
@@ -1035,9 +1205,15 @@ func TestAbstractShell_writeSubmoduleUpdateCmdPath(t *testing.T) {
 			mockWriter.On("Command", "git", "submodule", "foreach", "git lfs pull").Once()
 			mockWriter.On("EndIf").Once()
 
-			build := &common.Build{}
+			build := &common.Build{
+				JobResponse: common.JobResponse{Token: "xxx"},
+				Runner: &common.RunnerConfig{
+					RunnerCredentials: common.RunnerCredentials{URL: "https://example.com"},
+				},
+			}
 			build.Variables = append(build.Variables, common.JobVariable{Key: "GIT_SUBMODULE_PATHS", Value: test.paths})
-			shell.writeSubmoduleUpdateCmd(mockWriter, build, false)
+			err := shell.writeSubmoduleUpdateCmd(mockWriter, build, false)
+			assert.NoError(t, err)
 		})
 	}
 }
@@ -1557,6 +1733,125 @@ func TestAbstractShell_writeCleanupFileVariablesScript(t *testing.T) {
 	assert.NoError(t, err)
 }
 
+func testGenerateArtifactsMetadataData() (common.ShellScriptInfo, []interface{}) {
+	info := common.ShellScriptInfo{
+		Build: &common.Build{
+			JobResponse: common.JobResponse{
+				Variables: common.JobVariables{
+					{Key: "CI_RUNNER_ID", Value: "1000"},
+					{Key: "TEST_VARIABLE", Value: ""},
+				},
+				GitInfo: common.GitInfo{
+					RepoURL: "https://gitlab.com/my/repo.git",
+					Sha:     "testsha",
+				},
+				JobInfo: common.JobInfo{
+					Name: "testjob",
+				},
+			},
+			Runner: &common.RunnerConfig{
+				Name: "testrunner",
+				RunnerSettings: common.RunnerSettings{
+					Executor: "testexecutor",
+				},
+			},
+		},
+	}
+
+	parseRFC3339Mock := func(t string) bool {
+		_, err := time.Parse(time.RFC3339, t)
+		return err == nil
+	}
+
+	expected := []interface{}{
+		"--generate-artifacts-metadata",
+		"--runner-id",
+		"1000",
+		"--repo-url",
+		"https://gitlab.com/my/repo",
+		"--repo-digest",
+		"testsha",
+		"--job-name",
+		"testjob",
+		"--executor-name",
+		"testexecutor",
+		"--runner-name",
+		"testrunner",
+		"--started-at",
+		mock.MatchedBy(parseRFC3339Mock),
+		"--ended-at",
+		mock.MatchedBy(parseRFC3339Mock),
+		"--metadata-parameter",
+		"CI_RUNNER_ID",
+		"--metadata-parameter",
+		"TEST_VARIABLE",
+		"--metadata-parameter",
+		"RUNNER_GENERATE_ARTIFACTS_METADATA",
+	}
+
+	return info, expected
+}
+
+func TestWriteUploadArtifactIncludesGenerateArtifactsMetadataArgs(t *testing.T) {
+	info, expectedMetadataArgs := testGenerateArtifactsMetadataData()
+
+	info.Build.Runner.URL = "testurl"
+	info.Build.Token = "testtoken"
+	info.Build.ID = 1000
+	info.RunnerCommand = "testcommand"
+	info.Build.Variables = append(
+		info.Build.Variables,
+		common.JobVariable{Key: common.GenerateArtifactsMetadataVariable, Value: "true"},
+	)
+
+	uploaderArgs := []interface{}{
+		"artifacts-uploader",
+		"--url",
+		"testurl",
+		"--token",
+		"testtoken",
+		"--id",
+		"1000",
+	}
+
+	for _, f := range []common.ArtifactFormat{
+		common.ArtifactFormatZip,
+		common.ArtifactFormatGzip,
+		common.ArtifactFormatRaw,
+		common.ArtifactFormatDefault,
+	} {
+		t.Run(string(f), func(t *testing.T) {
+			args := []interface{}{"testcommand"}
+			args = append(args, uploaderArgs...)
+
+			if f == common.ArtifactFormatZip {
+				args = append(args, expectedMetadataArgs...)
+			}
+
+			args = append(args, "--path", "testpath")
+
+			if f != common.ArtifactFormatDefault {
+				args = append(args, "--artifact-format", string(f))
+			}
+
+			shellWriter := &MockShellWriter{}
+			shellWriter.On("IfCmd", mock.Anything, mock.Anything).Once()
+			shellWriter.On("Noticef", "Uploading artifacts...").Once()
+			shellWriter.On("Command", args...).Once()
+			shellWriter.On("Else").Once()
+			shellWriter.On("Warningf", mock.Anything, mock.Anything, mock.Anything).Once()
+			shellWriter.On("EndIf").Once()
+			defer shellWriter.AssertExpectations(t)
+
+			shell := &AbstractShell{}
+			shell.writeUploadArtifact(shellWriter, info, common.Artifact{
+				Paths:  []string{"testpath"},
+				Format: f,
+			})
+		})
+	}
+}
+
 func BenchmarkScriptStage(b *testing.B) {
 	stages := []common.BuildStage{
 		common.BuildStagePrepare,
@@ -1690,3 +1985,78 @@ func benchmarkScriptStage(b *testing.B, shell common.Shell, stage common.BuildSt
 		assert.NoError(b, err, stage)
 	}
 }
+
+func TestAbstractShell_writeGetSourcesScript_scriptHooks(t *testing.T) {
+	info := common.ShellScriptInfo{
+		Build: &common.Build{
+			JobResponse: common.JobResponse{
+				Variables: common.JobVariables{
+					{Key: "GIT_STRATEGY", Value: "fetch"},
+					{Key: "GIT_CHECKOUT", Value: "false"},
+				},
+				GitInfo: common.GitInfo{
+					RepoURL: "repo-url",
+				},
+				Hooks: common.Hooks{
+					{
+						Name:   common.HookPreGetSourcesScript,
+						Script: common.StepScript{"job payload", "pre_get_sources"},
+					},
+					{
+						Name:   common.HookPostGetSourcesScript,
+						Script: common.StepScript{"job payload", "post_get_sources"},
+					},
+				},
+			},
+			Runner:   &common.RunnerConfig{},
+			BuildDir: "build-dir",
+		},
+		PreGetSourcesScript:  "config pre_get_sources",
+		PostGetSourcesScript: "config post_get_sources",
+	}
+
+	m := &MockShellWriter{}
+	defer m.AssertExpectations(t)
+
+	m.On("Variable", mock.Anything)
+
+	// Pre get sources from configuration file
+	m.On("Noticef", "$ %s", "config pre_get_sources")
+	m.On("Line", "config pre_get_sources")
+	// Pre get sources from job payload
+	m.On("Noticef", "$ %s", "job payload")
+	m.On("Line", "job payload")
+	m.On("Noticef", "$ %s", "pre_get_sources")
+	m.On("Line", "pre_get_sources")
+
+	m.On("CheckForErrors")
+	m.On("Noticef", "Fetching changes...")
+	m.On("MkTmpDir", "git-template").Return("git-template-dir").Once()
+	m.On("Join", "git-template-dir", "config").Return("git-template-dir-config").Once()
+	m.On("Command", "git", "config", "-f", "git-template-dir-config", mock.Anything, mock.Anything)
+	m.On("RmFile", mock.Anything)
+	m.On("Command", "git", "init", "build-dir", "--template", "git-template-dir")
+	m.On("Cd", "build-dir")
+	m.On("IfCmd", "git", "remote", "add", "origin", "repo-url")
+	m.On("Noticef", "Created fresh repository.")
+	m.On("Else")
+	m.On("Command", "git", "remote", "set-url", "origin", "repo-url")
+	m.On("EndIf")
+	m.On("Command", "git", "-c", mock.Anything, "fetch", "origin", "--prune", "--quiet")
+	m.On("Noticef", "Skipping Git checkout")
+	m.On("Noticef", "Skipping Git submodules setup")
+
+	// Post get sources from job payload
+	m.On("Noticef", "$ %s", "job payload")
+	m.On("Line", "job payload")
+	m.On("Noticef", "$ %s", "post_get_sources")
+	m.On("Line", "post_get_sources")
+	// Post get sources from configuration file
+	m.On("Noticef", "$ %s", "config post_get_sources")
+	m.On("Line", "config post_get_sources")
+
+	shell := new(AbstractShell)
+
+	err := shell.writeGetSourcesScript(m, info)
+	assert.NoError(t, err)
+}
